/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <stdio.h>
#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"
#include "color.h"

/*
@BeginDoc

@BeginModule

@Name      = Outputgmt
@Title     = Formatted GMT output
@Section   = Formatted I/O
@Arguments = ifiles
@Operators = outputcenter outputbounds

@EndModule


@BeginOperator_outputcenter

@Title     = Output grid center

@BeginDesciption
@EndDesciption

@EndOperator


@BeginOperator_outputbounds

@Title     = Output grid bounds

@BeginDesciption
@EndDesciption

@EndOperator

@EndDoc
*/


void *Outputgmt(void *argument)
{
  static char func[] = "Outputgmt";
  int i;
  const char *operators[] = {"outputcenter", "outputbounds", "outputboundscpt"};
  enum                     {f_outputcenter, f_outputbounds, f_outputboundscpt};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int varID, recID;
  int gridsize = 0;
  int gridID, gridID0 = -1, code;
  int nrecs;
  int levelID;
  int tsID;
  int streamID = 0;
  int vlistID;
  int nmiss;
  int nlon, nlat, nalloc;
  int gridcorners = 0, ic;
  int lgrid_gen_bounds = FALSE, luse_grid_corner = FALSE;
  double missval;
  double *array = NULL;
  double *grid_center_lat = NULL, *grid_center_lon = NULL;
  double *grid_corner_lat = NULL, *grid_corner_lon = NULL;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  if ( operatorID == f_outputbounds || operatorID == f_outputboundscpt )
    luse_grid_corner = TRUE;

  if ( operatorID == f_outputboundscpt )
    {
      char *cpt_file;

      cpt_file = operatorArgv()[0];

      read_cpt(cpt_file);
    }

  streamID = streamOpenRead(cdoStreamName(0));
  if ( streamID < 0 ) cdiError(streamID, "Open failed on %s", cdoStreamName(0));

  vlistID = streamInqVlist(streamID);

  gridsize = vlistGridsizeMax(vlistID);

  array = (double *) malloc(gridsize*sizeof(double));

  tsID = 0;
  nrecs = streamInqTimestep(streamID, tsID);
  /*  while ( (nrecs = streamInqTimestep(streamID, tsID)) ) */
    {
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID, &varID, &levelID);

	  if ( recID > 0 ) continue;

	  code     = vlistInqVarCode(vlistID, varID);
	  gridID   = vlistInqVarGrid(vlistID, varID);
	  missval  = vlistInqVarMissval(vlistID, varID);

	  if ( gridID != gridID0 )
	    {
	      gridID0 = gridID;

	      if ( gridInqType(gridID) != GRID_LONLAT      &&
		   gridInqType(gridID) != GRID_GAUSSIAN    &&
		   gridInqType(gridID) != GRID_TRIANGULAR  &&
		   gridInqType(gridID) != GRID_CURVILINEAR &&
		   gridInqType(gridID) != GRID_CELL )
		cdoAbort("Output of %s data failed!", gridNamePtr(gridInqType(gridID)));

	      if ( gridInqType(gridID) != GRID_CELL && gridInqType(gridID) != GRID_CURVILINEAR )
		{
		  if ( gridInqType(gridID) != GRID_CELL ) gridID = gridToCell(gridID);
		  else
		    {
		      gridID = gridToCurvilinear(gridID);
		      lgrid_gen_bounds = TRUE;
		    }
		}

	      gridsize = gridInqSize(gridID);
	      nlon     = gridInqXsize(gridID);
	      nlat     = gridInqYsize(gridID);

	      if ( gridInqType(gridID) == GRID_CELL )
		gridcorners = gridInqNvertex(gridID);
	      else
		gridcorners = 4;

	      grid_center_lat = (double *) realloc(grid_center_lat, gridsize*sizeof(double));
	      grid_center_lon = (double *) realloc(grid_center_lon, gridsize*sizeof(double));

	      gridInqYvals(gridID, grid_center_lat);
	      gridInqXvals(gridID, grid_center_lon);

	      if ( luse_grid_corner )
		{
		  nalloc = gridcorners*gridsize;
		  grid_corner_lat = (double *) realloc(grid_corner_lat, nalloc*sizeof(double));
		  grid_corner_lon = (double *) realloc(grid_corner_lon, nalloc*sizeof(double));

		  if ( gridInqYbounds(gridID, NULL) && gridInqXbounds(gridID, NULL) )
		    {
		      gridInqYbounds(gridID, grid_corner_lat);
		      gridInqXbounds(gridID, grid_corner_lon);
		    }
		  else
		    {
		      if ( lgrid_gen_bounds )
			{
			  genXbounds(nlon, nlat, grid_center_lon, grid_corner_lon);
			  genYbounds(nlon, nlat, grid_center_lat, grid_corner_lat);
			}
		      else
			cdoAbort("grid corner missing!");
		    }
		}
	    }

	  streamReadRecord(streamID, array, &nmiss);

	  if ( operatorID == f_outputcenter )
	    {
	      for ( i = 0; i < gridsize; i++ )
		{
		  if ( array[i] != missval )
		    fprintf(stdout, " %g  %g  %g\n", grid_center_lon[i], grid_center_lat[i], array[i]);
		}
	      fprintf(stdout, "\n");
	    }
	  else if ( operatorID == f_outputbounds || operatorID == f_outputboundscpt )
	    {
	      for ( i = 0; i < gridsize; i++ )
		{
		  if ( array[i] != missval )
		    fprintf(stdout, "> -Z%g", array[i]);
		  else
		    fprintf(stdout, ">");

		  if ( operatorID == f_outputboundscpt )
		    {
		      int r = 0, g = 0, b = 0, n;

		      if ( array[i] != missval )
			{
			  for ( n = 0; n < n_colors; n++ )
			    if ( array[i] > lut[n].z_low && array[i] <= lut[n].z_high ) break;

			  if ( n == n_colors )
			    {
			      r = bfn[0].rgb[0];  g = bfn[0].rgb[1];  b = bfn[0].rgb[2];
			    }
			  else
			    {
			      r = lut[n].rgb_high[0];  g = lut[n].rgb_high[1];  b = lut[n].rgb_high[2];
			    }
			}
		      else
			{
			  r = bfn[2].rgb[0];  g = bfn[2].rgb[1];  b = bfn[2].rgb[2]; 
			}

		      fprintf(stdout, " -G%d/%d/%d", r, g, b);
		    }

		  fprintf(stdout, "\n");

		  for ( ic = 0; ic < gridcorners; ic++ )
		    fprintf(stdout, "   %g  %g\n",
			    grid_corner_lon[i*gridcorners+ic], grid_corner_lat[i*gridcorners+ic]);
		  fprintf(stdout, "   %g  %g\n",
			  grid_corner_lon[i*gridcorners+0], grid_corner_lat[i*gridcorners+0]);
		}
	      fprintf(stdout, "\n");
	    }
	}
      tsID++;
    }

  streamClose(streamID);

  if ( array ) free(array);
  if ( grid_center_lon ) free(grid_center_lon);
  if ( grid_center_lat ) free(grid_center_lat);
  if ( grid_corner_lon ) free(grid_corner_lon);
  if ( grid_corner_lat ) free(grid_corner_lat);

  cdoFinish();

  return (0);
}
