/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>
#include <math.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"
#include "remap.h"

/*
@BeginDoc

@BeginModule

@Name      = Remap
@Title     = SCRIP interpolation
@Section   = Interpolation
@Arguments = ifile ofile
@Operators = remapcon remapbil remapbic remapdis

@EndModule


@BeginOperator_remapcon

@Title     = Conservative remapping
@Parameter = grid

@BeginDesciption
SCRIP first order conservative remapping.
@EndDesciption

@BeginParameter
@Item = grid
STRING  Grid description file or name of the target grid
@EndParameter

@BeginEnvironment
@Item = NORMALIZE_OPT
This variable is used to choose the normalization
of the remapping. By default, NORMALIZE_OPT is set to be
'fracarea' and will include the destination area fraction in
the output weights; other options are 'none' and 'destarea'
(for more information see \cite{SCRIP}).
@EndEnvironment

@EndOperator


@BeginOperator_remapbil

@Title     = Bilinear interpolation
@Parameter = grid

@BeginDesciption
SCRIP bilinear interpolation (only rectangular grids).
@EndDesciption

@BeginParameter
@Item = grid
STRING  Grid description file or name of the target grid
@EndParameter

@EndOperator


@BeginOperator_remapbic

@Title     = Bicubic interpolation
@Parameter = grid

@BeginDesciption
SCRIP bicubic interpolation (only rectangular grids).
@EndDesciption

@BeginParameter
@Item = grid
STRING  Grid description file or name of the target grid
@EndParameter

@EndOperator


@BeginOperator_remapdis

@Title     = Distance-weighted averaging
@Parameter = grid

@BeginDesciption
SCRIP distance-weighted average of the four nearest neighbor values.
@EndDesciption

@BeginParameter
@Item = grid
STRING  Grid description file or name of the target grid
@EndParameter

@EndOperator

@EndDoc
*/


void *Remap(void *argument)
{
  static char func[] = "Remap";
  const char *operators[] = {"remapcon", "remapbil", "remapbic", "remapdis", "remapdis1", "remapcon1"};
  enum                     {f_remapcon, f_remapbil, f_remapbic, f_remapdis, f_remapdis1, f_remapcon1};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2;
  int nrecs, ngrids;
  int index;
  int tsID, recID, varID, levelID;
  int gridsize, gridsize2;
  int vlistID1, vlistID2;
  int taxisID1, taxisID2;
  int vdate, vtime;
  int gridID1 = -1, gridID2;
  int nmiss1, nmiss2, i;
  int *imask = NULL, *imasknew = NULL, newmask;
  double missval;
  double *array1 = NULL, *array2 = NULL;
  double *grad1_lat = NULL, *grad1_lon = NULL, *grad1_latlon = NULL;
  FIELD field1, field2;
  REMAPGRID remapgrid;
  REMAPVARS remapvars;
  int gridIDremap = -1;
  int map_type = -1;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, operators);

  switch (operatorID)
    {
    case f_remapcon:
    case f_remapcon1:
      map_type = map_type_conserv;
      remapgrid.luse_grid_centers = FALSE;
      remapvars.norm_opt = norm_opt_fracarea;
      break;
    case f_remapbil:
      map_type = map_type_bilinear;
      remapgrid.luse_grid_centers = TRUE;
      remapvars.norm_opt = norm_opt_none;
      break;
    case f_remapbic:
      map_type = map_type_bicubic;
      remapgrid.luse_grid_centers = TRUE;
      remapvars.norm_opt = norm_opt_none;
      break;
    case f_remapdis:
      map_type = map_type_distwgt;
      remapgrid.luse_grid_centers = TRUE;
      remapvars.norm_opt = norm_opt_none;
      break;
    case f_remapdis1:
      map_type = map_type_distwgt1;
      remapgrid.luse_grid_centers = TRUE;
      remapvars.norm_opt = norm_opt_none;
      break;
    default:
      cdoAbort("unknown mapping method");
    }

  if ( operatorID == f_remapcon )
    {
      char *envstring;

      envstring = getenv("NORMALIZE_OPT");

      if ( envstring )
        {
	  if ( strcmp(envstring, "fracarea") == 0 )
	    remapvars.norm_opt = norm_opt_fracarea;
	  else if ( strcmp(envstring, "destarea") == 0 )
	    remapvars.norm_opt = norm_opt_destarea;
	  else if ( strcmp(envstring, "none") == 0 )
	    remapvars.norm_opt = norm_opt_none;
	  else
	    cdoWarning("NORMALIZE_OPT=%s unsupported!\n", envstring);
	}

      if ( cdoVerbose )
        {
	  if ( remapvars.norm_opt == norm_opt_fracarea )
	    cdoPrint("Normalization option: fracarea");
	  else if ( remapvars.norm_opt == norm_opt_destarea )
	    cdoPrint("Normalization option: destarea");
	  else
	    cdoPrint("Normalization option: none");
	}
    }

  /*
  remapgrid.luse_grid1_area = FALSE;
  remapgrid.luse_grid2_area = FALSE;
  */
  remapgrid.restrict_type = restrict_latitude;
  remapgrid.num_srch_bins = 180;
  remapgrid.pinit = FALSE;
  remapvars.pinit = FALSE;

  operatorInputArg("grid description file or name");
  gridID2 = cdoDefineGrid(operatorArgv()[0]);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  taxisID1 = vlistInqTaxis(vlistID1);
  taxisID2 = taxisDuplicate(taxisID1);

  vlistDefTaxis(vlistID2, taxisID2);

  ngrids = vlistNgrids(vlistID1);
  for ( index = 0; index < ngrids; index++ )
    {
      gridID1 = vlistGrid(vlistID1, index);

      if ( gridInqType(gridID1) != GRID_LONLAT &&
	   gridInqType(gridID1) != GRID_GAUSSIAN &&
	   gridInqType(gridID1) != GRID_TRIANGULAR &&
	   gridInqType(gridID1) != GRID_CURVILINEAR &&
	   gridInqType(gridID1) != GRID_CELL )
	cdoAbort("Remapping of %s data failed!", gridNamePtr(gridInqType(gridID1)) );

      vlistChangeGridIndex(vlistID2, index, gridID2);
    }

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID1);

  if ( map_type == map_type_bicubic )
    {
      grad1_lat    = (double *) malloc(gridsize*sizeof(double));
      grad1_lon    = (double *) malloc(gridsize*sizeof(double));
      grad1_latlon = (double *) malloc(gridsize*sizeof(double));
    }

  array1   = (double *) malloc(gridsize*sizeof(double));
  imask    = (int *) malloc(gridsize*sizeof(int));
  imasknew = (int *) malloc(gridsize*sizeof(int));
  for ( i = 0; i < gridsize; i++ ) imask[i] = TRUE;

  gridsize = gridInqSize(gridID2);
  array2   = (double *) malloc(gridsize*sizeof(double));

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      vdate = taxisInqVdate(taxisID1);
      vtime = taxisInqVtime(taxisID1);

      taxisDefVdate(taxisID2, vdate);
      taxisDefVtime(taxisID2, vtime);

      streamDefTimestep(streamID2, tsID);
	       
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, array1, &nmiss1);

	  gridID1 = vlistInqVarGrid(vlistID1, varID);
	  missval = vlistInqVarMissval(vlistID1, varID);
	  gridsize = gridInqSize(gridID1);

	  field1.grid    = gridID1;
	  field1.nmiss   = nmiss1;
	  field1.missval = missval;
	  field1.ptr     = array1;
	  field2.grid    = gridID2;
	  field2.ptr     = array2;
	  field2.nmiss   = 0;

	  for ( i = 0; i < gridsize; i++ )
	    if ( array1[i] == missval )
	      imasknew[i] = FALSE;
	    else
	      imasknew[i] = TRUE;

	  newmask = FALSE;
	  for ( i = 0; i < gridsize; i++ )
	    {
	      if ( imask[i] != imasknew[i] )
		{
		  imask[i] = imasknew[i];
		  newmask = TRUE;
		}
	    }

	  if ( gridID1 != gridIDremap || newmask )
	    {
	      /* initialize grid information for both grids */
	      if ( gridID1 != gridIDremap )
		remapGridInit(map_type, gridID1, gridID2, &remapgrid);

	      if ( newmask )
		{
		  for ( i = 0; i < remapgrid.grid1_size; i++ )
		    remapgrid.grid1_mask[i] = imask[i];

		  if ( map_type == map_type_conserv )
		    {
		      for ( i = 0; i < remapgrid.grid1_size; i++ )
			{
			  remapgrid.grid1_area[i] = 0;
			  remapgrid.grid1_frac[i] = 0;
			}
		      for ( i = 0; i < remapgrid.grid2_size; i++ )
			{
			  remapgrid.grid2_area[i] = 0;
			  remapgrid.grid2_frac[i] = 0;
			}
		    }
		}

	      /* initialize some remapping variables */
	      remapVarsInit(map_type, &remapgrid, &remapvars);

	      if      ( map_type == map_type_conserv  ) remap_conserv(&remapgrid, &remapvars);
	      else if ( map_type == map_type_bilinear ) remap_bilin(&remapgrid, &remapvars);
	      else if ( map_type == map_type_bicubic  ) remap_bicub(&remapgrid, &remapvars);
	      else if ( map_type == map_type_distwgt  ) remap_distwgt(&remapgrid, &remapvars);
	      else if ( map_type == map_type_distwgt1 ) remap_distwgt1(&remapgrid, &remapvars);

	      if ( remapvars.num_links != remapvars.max_links )
		resize_remap_vars(&remapvars, remapvars.num_links-remapvars.max_links);

	      gridIDremap = gridID1;
	    }

	  if ( map_type == map_type_bicubic )
	    remap_gradients(remapgrid, array1, grad1_lat, grad1_lon, grad1_latlon);

	  if ( operatorID == f_remapcon1 )
	    remap_con1(array2, missval, gridInqSize(gridID2), remapvars.num_links, remapvars.wts,
		  remapvars.num_wts, remapvars.grid2_add, remapvars.grid1_add, array1);
	  else
	    remap(array2, missval, gridInqSize(gridID2), remapvars.num_links, remapvars.wts,
		  remapvars.num_wts, remapvars.grid2_add, remapvars.grid1_add,
		  array1, grad1_lat, grad1_lon, grad1_latlon);

	  gridsize2 = gridInqSize(gridID2);

	  /*
	  if ( operatorID == f_remapcon )
	    {
	      double grid2_err;

	      if ( remapvars.norm_opt == norm_opt_none )
		{
		  for ( i = 0; i < gridsize2; i++ )
		    {
		      grid2_err = remapgrid.grid2_frac[i]*remapgrid.grid2_area[i];
		      if ( grid2_err != zero )
			array2[i] = array2[i]/grid2_err;
		      else
			array2[i] = missval;
		    }
		}
	      else if ( remapvars.norm_opt == norm_opt_destarea )
		{
		  for ( i = 0; i < gridsize2; i++ )
		    {
		      if ( remapgrid.grid2_frac[i] != zero )
			array2[i] = array2[i]/remapgrid.grid2_frac[i];
		      else
			array2[i] = missval;
		    }
		}
	    }
	  */

	  nmiss2 = 0;
	  for ( i = 0; i < gridsize2; i++ )
	    if ( array2[i] == missval ) nmiss2++;

	  /* calculate some statistics */
	  if ( cdoVerbose ) remap_stat(remapgrid, remapvars, array1, array2, missval);

	  streamDefRecord(streamID2, varID, levelID);
	  streamWriteRecord(streamID2, array2, nmiss2);
	}
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);

  if ( imask ) free(imask);
  if ( imasknew ) free(imasknew);
  if ( array2 ) free(array2);
  if ( array1 ) free(array1);

  if ( grad1_latlon ) free(grad1_latlon);
  if ( grad1_lon ) free(grad1_lon);
  if ( grad1_lat ) free(grad1_lat);

  remapVarsFree(&remapvars);
  remapGridFree(&remapgrid);

  cdoFinish();

  return (0);
}
