/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>
#include <math.h>   /* fabs */

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"  /* processSelf */
#include "pstream.h"
#include "error.h"
#include "util.h"
#include "functs.h"

/*
@BeginDoc

@BeginModule

@Name      = Select
@Title     = Select
@Section   = Selection
@Arguments = ifile ofile
@Operators = selcode delcode selvar delvar sellevel selgrid selzaxis

@EndModule


@BeginOperator_selcode

@Title     = Select codes
@Parameter = codes

@BeginDesciption
Selects all fields with a code in a user given list.
@EndDesciption

@BeginParameter
@Item = codes
INTEGER  Comma separated list of codes
@EndParameter

@EndOperator


@BeginOperator_delcode

@Title     = Delete codes
@Parameter = codes

@BeginDesciption
Deletes all fields with a code in a user given list.
@EndDesciption

@BeginParameter
@Item = codes
INTEGER  Comma separated list of codes
@EndParameter

@EndOperator


@BeginOperator_selvar

@Title     = Select variables
@Parameter = vars

@BeginDesciption
Selects all fields with a variable name in a user given list.
@EndDesciption

@BeginParameter
@Item = vars
STRING  Comma separated list of variable names
@EndParameter

@EndOperator


@BeginOperator_delvar

@Title     = Delete variables
@Parameter = vars

@BeginDesciption
Deletes all fields with a variable name in a user given list.
@EndDesciption

@BeginParameter
@Item = vars
STRING  Comma separated list of variable names
@EndParameter

@EndOperator


@BeginOperator_sellevel

@Title     = Select levels
@Parameter = levels

@BeginDesciption
Selects all fields with a level in a user given list.
@EndDesciption

@BeginParameter
@Item = levels
FLOAT  Comma separated list of levels
@EndParameter

@EndOperator


@BeginOperator_selgrid

@Title     = Select grids
@Parameter = grids

@BeginDesciption
Selects all fields with a grid in a user given list.
@EndDesciption

@BeginParameter
@Item = grids
INTEGER  Comma separated list of grids
@EndParameter

@EndOperator


@BeginOperator_selzaxis

@Title     = Select zaxis
@Parameter = zaxis

@BeginDesciption
Selects all fields with a zaxis in a user given list.
@EndDesciption

@BeginParameter
@Item = zaxis
INTEGER  Comma separated list of zaxis
@EndParameter

@EndOperator

@EndDoc
*/

#define  NOPERATORS     7
#define  MAXARG     16384

void *Select(void *argument)
{
  const char func[] = "Select";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  char *ostring[NOPERATORS];
  int function[NOPERATORS];
  int streamID1, streamID2;
  int tsID, nrecs;
  int nvars, nlevs;
  int code, gridID, zaxisID, levID;
  double level;
  int varID2, levelID2;
  int recID, varID, levelID;
  int intarr[MAXARG], nsel = 0;
  int sfound[MAXARG];
  double fltarr[MAXARG];
  char varname[128];
  char **varnames = NULL;
  int vlistID1 = -1, vlistID2 = -1;
  int isel;
  int i;
  int lcopy = FALSE;
  int gridsize;
  int nmiss;
  double *array = NULL;

  i = 0;
  operators[i] = "selcode";   ostring[i] = "codes";     function[i++] = func_code;
  operators[i] = "selvar";    ostring[i] = "vars";      function[i++] = func_name;
  operators[i] = "sellevel";  ostring[i] = "levels";    function[i++] = func_level;
  operators[i] = "selgrid";   ostring[i] = "gridIDs";   function[i++] = func_grid;
  operators[i] = "selzaxis";  ostring[i] = "zaxisIDs";  function[i++] = func_zaxis;
  operators[i] = "delcode";   ostring[i] = "codes";     function[i++] = 777;
  operators[i] = "delvar";    ostring[i] = "vars";      function[i++] = 888;

  cdoInitialize(argument);

  if ( UNCHANGED_RECORD ) lcopy = TRUE;

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  operatorInputArg(ostring[operatorID]);

  if ( function[operatorID] == func_name || function[operatorID] == 888 )
    {
      nsel     = operatorArgc();
      varnames = operatorArgv();
    }
  else if ( function[operatorID] == func_level )
    {
      nsel     = operatorArgc();
      for ( i = 0; i < nsel; i++ )
	fltarr[i] = atof(operatorArgv()[i]);
    }
  else
    {
      nsel = args2intarr(operatorArgc(), operatorArgv(), intarr, MAXARG);
    }

  for ( i = 0; i < nsel; i++ ) sfound[i] = FALSE;
  /*
  printf("nsel = %d\n", nsel);
  for ( i = 0; i < nsel; i++ )
    if ( function[operatorID] == func_name )
      printf("name %d = %s\n", i+1, varnames[i]);
    else
      printf("int %d = %d\n", i+1, intarr[i]);
  */
  /*
  if ( nsel == 0 )
    cdoAbort("missing code argument!");
  */
  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);

  vlistClearFlag(vlistID1);
  nvars = vlistNvars(vlistID1);
  for ( varID = 0; varID < nvars; varID++ )
    {
      vlistInqVarName(vlistID1, varID, varname);
      code    = vlistInqVarCode(vlistID1, varID);
      gridID  = vlistInqVarGrid(vlistID1, varID);
      zaxisID = vlistInqVarZaxis(vlistID1, varID);
      nlevs   = zaxisInqSize(zaxisID);

      for ( levID = 0; levID < nlevs; levID++ )
	{
	  level = zaxisInqLevel(zaxisID, levID);

	  if ( function[operatorID] == 777 || function[operatorID] == 888 )
	    vlistDefFlag(vlistID1, varID, levID, TRUE);

	  for ( isel = 0; isel < nsel; isel++ )
	    {
	      switch (function[operatorID])
		{
		case func_code:
		  {
		    if ( intarr[isel] == code )
		      {
			vlistDefFlag(vlistID1, varID, levID, TRUE);
			sfound[isel] = TRUE;
		      }
		    break;
		  }
		case func_name:
		  {
		    if ( strcmp(varnames[isel], varname) == 0 )
		      {
			vlistDefFlag(vlistID1, varID, levID, TRUE);
			sfound[isel] = TRUE;
		      }
		    break;
		  }
		case func_level:
		  {
		    if ( fabs(fltarr[isel] - level) < 0.0001 )
		      {
			vlistDefFlag(vlistID1, varID, levID, TRUE);
			sfound[isel] = TRUE;
		      }
		    break;
		  }
		case func_grid:
		  {
		    if ( intarr[isel] == (gridID+1) )
		      {
			vlistDefFlag(vlistID1, varID, levID, TRUE);
			sfound[isel] = TRUE;
		      }
		    break;
		  }
		case func_zaxis:
		  {
		    if ( intarr[isel] == (zaxisID+1) )
		      {
			vlistDefFlag(vlistID1, varID, levID, TRUE);
			sfound[isel] = TRUE;
		      }
		    break;
		  }
		case 777:
		  {
		    if ( intarr[isel] == code )
		      {
			vlistDefFlag(vlistID1, varID, levID, FALSE);
			sfound[isel] = TRUE;
		      }
		    break;
		  }
		case 888:
		  {
		    if ( strcmp(varnames[isel], varname) == 0 )
		      {
			vlistDefFlag(vlistID1, varID, levID, FALSE);
			sfound[isel] = TRUE;
		      }
		    break;
		  }
		}
	    }
	}
    }

  for ( isel = 0; isel < nsel; isel++ )
    {
      if ( sfound[isel] == FALSE )
	switch (function[operatorID])
	  {
	  case func_code:
	  case 777:
	    {
	      cdoWarning("Code %d not found!", intarr[isel]);
	      break;
	    }
	  case func_name:
	  case 888:
	    {
	      cdoWarning("Variable %s not found!", varnames[isel]);
	      break;
	    }
	  case func_level:
	    {
	      cdoWarning("Level %g not found!", fltarr[isel]);
	      break;
	    }
	  case func_grid:
	    {
	      cdoWarning("Grid %d not found!", intarr[isel]);
	      break;
	    }
	  case func_zaxis:
	    {
	      cdoWarning("Zaxis %d not found!", intarr[isel]);
	      break;
	    }
	  }
    }

  vlistID2 = vlistNew();
  vlistCopyFlag(vlistID2, vlistID1);

  nrecs = vlistNrecs(vlistID2);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  if ( ! lcopy )
    {
      gridsize = vlistGridsizeMax(vlistID1);
      array = (double *) malloc(gridsize*sizeof(double));
    }

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      streamDefTimestep(streamID2, tsID);
     
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  if ( vlistInqFlag(vlistID1, varID, levelID) == TRUE )
	    {
	      varID2   = vlistFindVar(vlistID2, varID);
	      levelID2 = vlistFindLevel(vlistID2, varID, levelID);

	      streamDefRecord(streamID2, varID2, levelID2);
	      if ( lcopy )
		{
		  streamCopyRecord(streamID2, streamID1);
		}
	      else
		{
		  streamReadRecord(streamID1, array, &nmiss);
		  streamWriteRecord(streamID2, array, nmiss);
		}
     	    }
	}
       
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);
 
  if ( ! lcopy )
    if ( array ) free(array);

  cdoFinish();

  return (NULL);
}
