/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>
#include <ctype.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"  /* processSelf */
#include "pstream.h"
#include "error.h"
#include "util.h"
#include "functs.h"

/*
@BeginDoc

@BeginModule

@Name      = Seltime
@Title     = Select time
@Section   = Selection
@Arguments = ifile ofile
@Operators = seltimestep seltime selhour selday selmon selseas selyear seldate

@EndModule


@BeginOperator_seltimestep

@Title     = Select timesteps
@Parameter = timesteps

@BeginDesciption
Selects all timesteps with a timestep in a user given list.
@EndDesciption

@BeginParameter
@Item = timesteps
INTEGER  Comma separated list of timesteps
@EndParameter

@EndOperator


@BeginOperator_seltime

@Title     = Select times
@Parameter = times

@BeginDesciption
Selects all timesteps with a time in a user given list.
@EndDesciption

@BeginParameter
@Item = times
INTEGER  Comma separated list of times
@EndParameter

@EndOperator


@BeginOperator_selhour

@Title     = Select hours
@Parameter = hours

@BeginDesciption
Selects all timesteps with a hour in a user given list.
@EndDesciption

@BeginParameter
@Item = hours
INTEGER  Comma separated list of hours
@EndParameter

@EndOperator


@BeginOperator_selday

@Title     = Select days
@Parameter = days

@BeginDesciption
Selects all timesteps with a day in a user given list.
@EndDesciption

@BeginParameter
@Item = days
INTEGER  Comma separated list of days
@EndParameter

@EndOperator


@BeginOperator_selmon

@Title     = Select months
@Parameter = months

@BeginDesciption
Selects all timesteps with a month in a user given list.
@EndDesciption

@BeginParameter
@Item = months
INTEGER  Comma separated list of months
@EndParameter

@EndOperator


@BeginOperator_selseas

@Title     = Select seasons
@Parameter = seasons

@BeginDesciption
Selects all timesteps with a month of a season in a user given list.
@EndDesciption

@BeginParameter
@Item = seasons
STRING   Comma separated list of seasons (DJF, MAM, JJA, SON)
@EndParameter

@EndOperator


@BeginOperator_selyear

@Title     = Select years
@Parameter = years

@BeginDesciption
Selects all timesteps with a year in a user given list.
@EndDesciption

@BeginParameter
@Item = years
INTEGER  Comma separated list of years
@EndParameter

@EndOperator


@BeginOperator_seldate

@Title     = Select dates
@Parameter = date1 [date2]

@BeginDesciption
Selects all timesteps with a date in a given range.
@EndDesciption

@BeginParameter
@Item = date1
INTEGER  Start date
@Item = date2
INTEGER  End date
@EndParameter

@EndOperator

@EndDoc
*/

#define  NOPERATORS   8
#define  MAXARG       439200 /* 12 ts/day * 366 * 100 */
#define  EE2          100
#define  EE4          10000
#define  EE9          1000000000

void *Seltime(void *argument)
{
  const char func[] = "Seltime";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  char *ostring[NOPERATORS];
  int ttype[NOPERATORS];
  int intdat[NOPERATORS];
  int moddat[NOPERATORS];
  int streamID1, streamID2;
  int tsID, tsID2, nrecs;
  int recID, varID, levelID;
  int intarr[MAXARG], nsel = 0, selval;
  int vlistID1 = -1, vlistID2 = -1;
  int vdate, vtime;
  int taxisID;
  int copytimestep;
  int i;
  int lcopy = FALSE;
  int gridsize;
  int nmiss;
  double *array = NULL;

  i = 0;
  operators[i] = "seltimestep"; intdat[i] =   1; moddat[i] =   1; ttype[i] = func_step; ostring[i] = "timesteps"; i++;
  operators[i] = "seldate";     intdat[i] =   1; moddat[i] = EE9; ttype[i] = func_date; ostring[i] = "start date and end date"; i++;
  operators[i] = "seltime";     intdat[i] =   1; moddat[i] = EE4; ttype[i] = func_time; ostring[i] = "times";     i++;
  operators[i] = "selhour";     intdat[i] = EE2; moddat[i] = EE2; ttype[i] = func_time; ostring[i] = "hours";     i++;
  operators[i] = "selday";      intdat[i] =   1; moddat[i] = EE2; ttype[i] = func_date; ostring[i] = "days";      i++;
  operators[i] = "selmon";      intdat[i] = EE2; moddat[i] = EE2; ttype[i] = func_date; ostring[i] = "months";    i++;
  operators[i] = "selyear";     intdat[i] = EE4; moddat[i] = EE4; ttype[i] = func_date; ostring[i] = "years";     i++;
  operators[i] = "selseas";     intdat[i] = EE2; moddat[i] = EE2; ttype[i] = func_date; ostring[i] = "seasons";   i++;

  cdoInitialize(argument);

  if ( UNCHANGED_RECORD ) lcopy = TRUE;

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  operatorInputArg(ostring[operatorID]);

  if ( strcmp(operators[operatorID], "selseas") == 0 )
    {
      char Seas[3];
      int seas[4] = {FALSE, FALSE, FALSE, FALSE};
      size_t len;

      nsel = operatorArgc();
      if ( isdigit(*operatorArgv()[0]))
	for ( i = 0; i < nsel; i++ )
	  {
	    if      ( atoi(operatorArgv()[i]) == 1 || atoi(operatorArgv()[i]) == 13 ) seas[0] = TRUE;
	    else if ( atoi(operatorArgv()[i]) == 2 || atoi(operatorArgv()[i]) == 14 ) seas[1] = TRUE;
	    else if ( atoi(operatorArgv()[i]) == 3 || atoi(operatorArgv()[i]) == 15 ) seas[2] = TRUE;
	    else if ( atoi(operatorArgv()[i]) == 4 || atoi(operatorArgv()[i]) == 16 ) seas[3] = TRUE;
	  }
      else
	for ( i = 0; i < nsel; i++ )
	  {
	    len = strlen(operatorArgv()[i]);
	    if ( len > 3 ) len = 3;
	    while ( len-- > 0 ) Seas[len] = toupper(operatorArgv()[i][len]);
	    if      ( Seas[0] == 'D' && Seas[1] == 'J'  && Seas[2] == 'F' ) seas[0] = TRUE;
	    else if ( Seas[0] == 'M' && Seas[1] == 'A'  && Seas[2] == 'M' ) seas[1] = TRUE;
	    else if ( Seas[0] == 'J' && Seas[1] == 'J'  && Seas[2] == 'A' ) seas[2] = TRUE;
	    else if ( Seas[0] == 'S' && Seas[1] == 'O'  && Seas[2] == 'N' ) seas[3] = TRUE;
	  }

      nsel = 0;
      if ( seas[0] )  { intarr[nsel++] = 12; intarr[nsel++] =  1; intarr[nsel++] =  2; intarr[nsel++] = 13; }
      if ( seas[1] )  { intarr[nsel++] =  3; intarr[nsel++] =  4; intarr[nsel++] =  5; intarr[nsel++] = 14; }
      if ( seas[2] )  { intarr[nsel++] =  6; intarr[nsel++] =  7; intarr[nsel++] =  8; intarr[nsel++] = 15; }
      if ( seas[3] )  { intarr[nsel++] =  9; intarr[nsel++] = 10; intarr[nsel++] = 11; intarr[nsel++] = 16; }
    }
  else
    nsel = args2intarr(operatorArgc(), operatorArgv(), intarr, MAXARG);

  /*
  printf("nsel = %d\n", nsel);
  for ( i = 0; i < nsel; i++ )
    if ( ttype[operatorID] == func_name )
      printf("name %d = %s\n", i+1, operatorArgv()[i]);
    else
      printf("int %d = %d\n", i+1, intarr[i]);
  */
  /*
  if ( nsel == 0 )
    cdoAbort("missing code argument!");
  */
  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);
  if ( nsel == 1 && ttype[operatorID] == func_step )  vlistDefNtsteps(vlistID2, 1);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  taxisID = vlistInqTaxis(vlistID1);

  if ( ! lcopy )
    {
      gridsize = vlistGridsizeMax(vlistID1);
      array = (double *) malloc(gridsize*sizeof(double));
    }

  tsID  = 0;
  tsID2 = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      vdate = taxisInqVdate(taxisID);
      vtime = taxisInqVtime(taxisID);

      copytimestep = FALSE;
      selval = -1;

      if ( ttype[operatorID] == func_step )
	{
	  selval = tsID + 1;
	  if ( selval > intarr[nsel-1] ) break;
	}
      else if ( ttype[operatorID] == func_date )
	{
	  selval = (vdate/intdat[operatorID])%moddat[operatorID];
	}
      else if ( ttype[operatorID] == func_time )
	{
	  selval = (vtime/intdat[operatorID])%moddat[operatorID];
	}

      if ( strcmp(operators[operatorID], "seldate") == 0 )
	{
	  if ( selval >= intarr[0] && selval <= intarr[nsel-1] ) copytimestep = TRUE;
	}
      else
	{
	  for ( i = 0; i < nsel; i++ )
	    if ( selval == intarr[i] )
	      {
		copytimestep = TRUE;
		break;
	      }
	}

      if ( copytimestep )
	{
	  streamDefTimestep(streamID2, tsID2++);

	  for ( recID = 0; recID < nrecs; recID++ )
	    {
	      streamInqRecord(streamID1, &varID, &levelID);
	      streamDefRecord(streamID2, varID, levelID);
	      if ( lcopy )
		{
		  streamCopyRecord(streamID2, streamID1);
		}
	      else
		{
		  streamReadRecord(streamID1, array, &nmiss);
		  streamWriteRecord(streamID2, array, nmiss);
		}
	    }
	}
       
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);
 
  if ( ! lcopy )
    if ( array ) free(array);

  cdoFinish();

  return (NULL);
}
