/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"
#include "specspace.h"

/*
@BeginDoc

@BeginModule

@Name      = Spectral
@Title     = Spectral transformation
@Section   = Spectral transformation
@Arguments = ifile ofile
@Operators = sp2gp gp2sp sp2sp spcut

@EndModule


@BeginOperator_sp2gp

@Title     = Spectral to gridpoint

@BeginDesciption
Convert all spectral fields to gaussian gridpoint.
@EndDesciption

@EndOperator


@BeginOperator_gp2sp

@Title     = Gridpoint to spectral

@BeginDesciption
Convert all gaussian gridpoint fields to spectral.
@EndDesciption

@EndOperator


@BeginOperator_sp2sp

@Title     = Spectral to spectral
@Parameter = trunc

@BeginDesciption
Change truncation of all spectral fields.
The operator performs downward conversion by cutting the resolution.
Upward conversions are achieved by filling in zeros.
@EndDesciption

@BeginParameter
@Item = trunc
INTEGER  New spectral resolution
@EndParameter

@EndOperator


@BeginOperator_spcut

@Title     = Cut spectral wave number
@Parameter = wnums

@BeginDesciption
Set the user defined wave numbers to zero.
@EndDesciption

@BeginParameter
@Item = wnums
INTEGER  Comma separated list of wave numbers
@EndParameter

@EndOperator

@EndDoc
*/

#define  MAX_TRUNC  9999

void *Spectral(void *argument)
{
  static char func[] = "Spectral";
  const char *operators[] = {"gp2sp", "sp2gp", "sp2sp", "spcut"};
  enum                     {f_gp2sp, f_sp2gp, f_sp2sp, f_spcut};
  int noperators = sizeof(operators)/sizeof(operators[0]);
  int operatorID;
  int streamID1, streamID2;
  int nrecs, nvars;
  int tsID, recID, varID, levelID;
  int gridsize;
  int index, ngrids;
  int vlistID1, vlistID2;
  int gridIDsp = -1, gridIDgp = -1;
  int gridID1 = -1, gridID2 = -1;
  int gridID;
  int nmiss;
  int ncut = 0;
  int wnums[MAX_TRUNC], waves[MAX_TRUNC];
  int *vars;
  int lcopy = FALSE;
  double *array1 = NULL, *array2 = NULL;

  cdoInitialize(argument);

  if ( UNCHANGED_RECORD ) lcopy = TRUE;

  operatorID = cdoOperatorID(noperators, operators);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  ngrids = vlistNgrids(vlistID1);
  /* find first spectral grid */
  for ( index = 0; index < ngrids; index++ )
    {
      gridID = vlistGrid(vlistID1, index);
      if ( gridInqType(gridID) == GRID_SPECTRAL )
	{
	  gridIDsp = gridID;
	  break;
	}
    }
  /* find first gaussian grid */
  for ( index = 0; index < ngrids; index++ )
    {
      gridID = vlistGrid(vlistID1, index);
      if ( gridInqType(gridID) == GRID_GAUSSIAN )
	{
	  gridIDgp = gridID;
	  break;
	}
    }

  /* define output grid */
  if ( operatorID == f_gp2sp )
    {
      gridID1  = gridIDgp;  

      if ( gridID1 != -1 )
	{
	  if ( gridIDsp != -1 )
	    if ( gridInqTrunc(gridIDgp) != gridInqTrunc(gridIDsp) ) gridIDsp = -1;

	  if ( gridIDsp == -1 )
	    {
	      int trunc = gridInqTrunc(gridIDgp);
	      int nsp = (trunc+1)*(trunc+2);
	      gridIDsp = gridNew(GRID_SPECTRAL, nsp);
	      gridDefTrunc(gridIDsp, trunc);
	    }
	}

      if ( gridIDsp == -1 && gridInqType(vlistGrid(vlistID1, 0)) == GRID_GAUSSIAN_REDUCED )
	cdoAbort("Gaussian reduced grid found. Use option -R to convert it to a regular grid!");

      if ( gridIDsp == -1 )
	cdoAbort("No gaussian grid data found!");

      gridID2 = gridIDsp;
    }
  else if ( operatorID == f_sp2gp )
    {
      gridID1  = gridIDsp;

      if ( gridID1 != -1 )
	{
	  if ( gridIDgp != -1 )
	    if ( gridInqTrunc(gridIDsp) != gridInqTrunc(gridIDgp) ) gridIDgp = -1;

	  if ( gridIDgp == -1 )
	    {
	      char gridname[20];
	      sprintf(gridname, "t%dgrid", gridInqTrunc(gridIDsp));
	      gridIDgp = gridFromName(gridname);
	    }
	}
      
      if ( gridIDgp == -1 )
	cdoAbort("No spectral data found!");

      gridID2 = gridIDgp;
    }
  else if ( operatorID == f_sp2sp )
    {
      gridID1 = gridIDsp;

      operatorInputArg("truncation");
      if ( gridID1 != -1 )
	{
	  int trunc = atoi(operatorArgv()[0]);
	  int nsp = (trunc+1)*(trunc+2);
	  gridIDsp = gridNew(GRID_SPECTRAL, nsp);
	  gridDefTrunc(gridIDsp, trunc);
	}
      else
	cdoAbort("No spectral data found!");

      gridID2 = gridIDsp;
    }
  else if ( operatorID == f_spcut )
    {
      int i, j;
      gridID1 = gridIDsp;

      operatorInputArg("wave numbers");
      if ( gridID1 != -1 )
	{
	  ncut = args2intarr(operatorArgc(), operatorArgv(), wnums, MAX_TRUNC);
	  for ( i = 0; i < MAX_TRUNC; i++ ) waves[i] = 1;
	  for ( i = 0; i < ncut; i++ )
	    {
	      j = wnums[i] - 1;
	      if ( j < 0 || j >= MAX_TRUNC )
		cdoAbort("wave number %d out of range!", wnums[i]);
	      waves[j] = 0;
	    }
	  /*
	  for ( i = 0; i < 200; i++ )
	    printf("%d %d\n", i+1, waves[i]);
	  */
	}
      else
	cdoAbort("No spectral data found!");

      gridID2 = gridIDsp;
    }

  nvars = vlistNvars(vlistID2);
  vars  = (int *) malloc(nvars*sizeof(int));
  for ( varID = 0; varID < nvars; varID++ )
    {
      if ( gridID1 == vlistInqVarGrid(vlistID1, varID) )
	vars[varID] = TRUE;
      else
	vars[varID] = FALSE;
    }

  vlistChangeGrid(vlistID2, gridID1, gridID2);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID1);
  array1 = (double *) malloc(gridsize*sizeof(double));

  gridsize = gridInqSize(gridID2);
  array2 = (double *) malloc(gridsize*sizeof(double));

  tsID = 0;
  while ( TRUE )
    {
      nrecs = streamInqTimestep(streamID1, tsID);
      if ( nrecs == 0 ) break;

      /*     if ( ntimesteps != 0 ) */
	{
	  streamDefTimestep(streamID2, tsID);
	}
	       
      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);

	  if ( vars[varID] )
	    {
	      streamReadRecord(streamID1, array1, &nmiss);
	      if ( nmiss ) cdoAbort("missing values unsupported for spectral data!");

	      gridID1 = vlistInqVarGrid(vlistID1, varID);
	      if ( operatorID == f_gp2sp )
		grid2spec(gridID1, array1, gridID2, array2);	      
	      else if ( operatorID == f_sp2gp )
		spec2grid(gridID1, array1, gridID2, array2);
	      else if ( operatorID == f_sp2sp )
		spec2spec(gridID1, array1, gridID2, array2);
	      else if ( operatorID == f_spcut )
		speccut(gridID1, array1, array2, waves);

	      streamDefRecord(streamID2, varID, levelID);
	      streamWriteRecord(streamID2, array2, nmiss);  
	    }   
	  else
	    {
	      streamDefRecord(streamID2, varID, levelID);
	      if ( lcopy )
		{
		  streamCopyRecord(streamID2, streamID1);
		}
	      else
		{
		  streamReadRecord(streamID1, array1, &nmiss);
		  streamWriteRecord(streamID2, array1, nmiss);
		}
	    }    
	}
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);

  if ( array2 ) free(array2);
  if ( array1 ) free(array1);

  cdoFinish();

  return (0);
}
