/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <string.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"

/*
@BeginDoc

@BeginModule

@Name      = Splittime
@Title     = 
@Section   = File operations
@Arguments = ifile oprefix
@Operators = splithour splitday splitmon splitseas

@EndModule


@BeginOperator_splithour

@Title     = Split hours

@BeginDesciption
Splits a file into pieces, one for each hour.
@EndDesciption

@EndOperator


@BeginOperator_splitday

@Title     = Split days

@BeginDesciption
Splits a file into pieces, one for each day.
@EndDesciption

@EndOperator


@BeginOperator_splitmon

@Title     = Split months

@BeginDesciption
Splits a file into pieces, one for each month.
@EndDesciption

@EndOperator


@BeginOperator_splitseas

@Title     = Split seasons

@BeginDesciption
Splits a file into pieces, one for each season.
@EndDesciption

@EndOperator

@EndDoc
*/

#define  MAX_STREAMS 32
#define  NOPERATORS  4

void *Splittime(void *argument)
{
  static char func[] = "Splittime";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  int ttype[NOPERATORS];
  int intdat[NOPERATORS];
  int nchars;
  int streamID1, streamID2;
  int varID;
  int nrecs;
  int tsID, recID, levelID;
  int vlistID1, vlistID2;
  int  streamIDs[MAX_STREAMS], tsIDs[MAX_STREAMS];
  char *filesuffix;
  char filename[1024];
  char *Seas[4] = {"DJF", "MAM", "JJA", "SON"};
  int index = 0;
  int i;
  int vdate, vtime;
  int taxisID;
  int lcopy = FALSE;
  int gridsize;
  int nmiss;
  double *array = NULL;

  i = 0;
  operators[i] = "splithour";   intdat[i] = 100;  ttype[i] = func_time; i++;
  operators[i] = "splitday";    intdat[i] =   1;  ttype[i] = func_date; i++;
  operators[i] = "splitmon";    intdat[i] = 100;  ttype[i] = func_date; i++;
  operators[i] = "splitseas";   intdat[i] = 100;  ttype[i] = func_date; i++;

  cdoInitialize(argument);

  if ( UNCHANGED_RECORD ) lcopy = TRUE;

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  for ( i = 0; i < MAX_STREAMS; i++ ) streamIDs[i] = -1;
  for ( i = 0; i < MAX_STREAMS; i++ ) tsIDs[i] = 0;

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  strcpy(filename, cdoStreamName(1));
  nchars = strlen(filename);
  filesuffix = streamFilesuffix(cdoDefaultFileType);

  if ( ! lcopy )
    {
      gridsize = vlistGridsizeMax(vlistID1);
      array = (double *) malloc(gridsize*sizeof(double));
    }

  taxisID = vlistInqTaxis(vlistID1);

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      vdate = taxisInqVdate(taxisID);
      vtime = taxisInqVtime(taxisID);

      if ( ttype[operatorID] == func_date )
	{
	  index = (vdate/intdat[operatorID])%100;

	  if ( strcmp(operators[operatorID], "splitseas") == 0 )
	    {
	      if ( index < 0 || index > 16 )
		cdoAbort("Month %d out of range!", index);

	      if ( index <= 12 )
		index = (index % 12) / 3;
	      else
		index = index - 13;

	      if ( index < 0 || index > 3 )
		cdoAbort("Season %d out of range!", index+1);
	    }
	}
      else if ( ttype[operatorID] == func_time )
	{
	  index = (vtime/intdat[operatorID])%100;
	}

      if ( index < 0 || index >= MAX_STREAMS )
	cdoAbort("Index out of range!");

      streamID2 = streamIDs[index];
      if ( streamID2 < 0 )
	{
	  if ( strcmp(operators[operatorID], "splitseas") == 0 )
	    {
	      sprintf(filename+nchars, "%3s", Seas[index]);
	      sprintf(filename+nchars+3, "%s", filesuffix);
	    }
	  else
	    {
	      sprintf(filename+nchars, "%02d", index);
	      sprintf(filename+nchars+2, "%s", filesuffix);
	    }

	  if ( cdoVerbose ) cdoPrint("create file %s", filename);

	  streamID2 = streamOpenWrite(filename, cdoFiletype());
	  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", filename);

	  streamDefVlist(streamID2, vlistID2);

	  streamIDs[index] = streamID2;
	}

      streamDefTimestep(streamID2, tsIDs[index]++);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamDefRecord(streamID2,  varID,  levelID);
	  if ( lcopy )
	    {
	      streamCopyRecord(streamID2, streamID1);
	    }
	  else
	    {
	      streamReadRecord(streamID1, array, &nmiss);
	      streamWriteRecord(streamID2, array, nmiss);
	    }
	}

      tsID++;
    }

  streamClose(streamID1);

  for ( index = 0; index < MAX_STREAMS; index++ )
    {
      streamID2 = streamIDs[index];
      if ( streamID2 >= 0 )  streamClose(streamID2);
    }
 
  if ( ! lcopy )
    if ( array ) free(array);

  cdoFinish();

  return (0);
}
