/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <stdio.h>
#include <math.h>

#include "cdi.h"
#include "cdo.h"
#include "pstream.h"
#include "dtypes.h"
#include "functs.h"
#include "field.h"
#include "dmemory.h"

/*
@BeginDoc

@BeginModule

@Name      = Timstat
@Title     = 
%@Section   = Statistical description of a time period
@Section   = Statistical description of the data
@Arguments = ifile ofile
@Operators = timmin timmax timsum timmean timavg timstd hourmin hourmax hoursum hourmean houravg hourstd daymin daymax daysum daymean dayavg daystd monmin monmax monsum monmean monavg monstd yearmin yearmax yearsum yearmean yearavg yearstd

@EndModule


@BeginOperator_timmin

@Title     = Time minimum

@BeginDesciption
@IfMan
o(1,x) = min{i(t',x'), x'=x}
@EndifMan
@IfDoc
@BeginMath
o(1,x) = \mbox{\bf min}\{i(t',x'), x' = x\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_timmax

@Title     = Time maximum

@BeginDesciption
@IfMan
o(1,x) = max{i(t',x'), x'=x}
@EndifMan
@IfDoc
@BeginMath
o(1,x) = \mbox{\bf max}\{i(t',x'), x' = x\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_timsum

@Title     = Time sum

@BeginDesciption
@IfMan
o(1,x) = sum{i(t,x)}
@EndifMan
@IfDoc
@BeginMath
o(1,x) = \sum\limits_{t=1}^{n} i(t',x)
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_timmean

@Title     = Time mean

@BeginDesciption
@IfMan
o(1,x) = mean{i(t',x'), x'=x}
@EndifMan
@IfDoc
@BeginMath
o(1,x) = \mbox{\bf mean}\{i(t',x'), x' = x\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_timavg

@Title     = Time average

@BeginDesciption
@IfMan
o(1,x) = avg{i(t',x'), x'=x}
@EndifMan
@IfDoc
@BeginMath
o(1,x) = \mbox{\bf avg}\{i(t',x'), x' = x\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_timstd

@Title     = Time standard deviation

@BeginDesciption
@IfMan
o(1,x) = sqrt{var{i(t',x'), x'=x}}
@EndifMan
@IfDoc
@BeginMath
o(1,x) = \sqrt{\mbox{\bf var}\{i(t',x'), x' = x\}}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_yearmin

@NewTable

@Title     = Yearly minimum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same year, it is

o(t,x) = min{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same year, it is
@BeginMath
o(t,x) = \mbox{\bf min}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_yearmax

@Title     = Yearly maximum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same year, it is

o(t,x) = max{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same year, it is
@BeginMath
o(t,x) = \mbox{\bf max}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_yearsum

@Title     = Yearly sum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same year, it is

o(t,x) = sum{i(t',x)}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same year, it is
@BeginMath
o(t,x) = \sum\limits_{t=1}^{n} i(t',x)
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_yearmean

@Title     = Yearly mean

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same year, it is

o(t,x) = mean{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same year, it is
@BeginMath
o(t,x) = \mbox{\bf mean}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_yearavg

@Title     = Yearly average

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same year, it is

o(t,x) = avg{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same year, it is
@BeginMath
o(t,x) = \mbox{\bf avg}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_yearstd

@Title     = Yearly standard deviation

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same year, it is

o(t,x) = sqrt{var{i(t',x), t_1 < t' <= t_n}}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same year, it is
@BeginMath
o(t,x) = \sqrt{\mbox{\bf var}\{i(t',x), t_1 < t' \leq t_n\}}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_monmin

@NewTable

@Title     = Monthly minimum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same month, it is

o(t,x) = min{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same month, it is
@BeginMath
o(t,x) = \mbox{\bf min}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_monmax

@Title     = Monthly maximum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same month, it is

o(t,x) = max{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same month, it is
@BeginMath
o(t,x) = \mbox{\bf max}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_monsum

@Title     = Monthly sum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same month, it is

o(t,x) = sum{i(t',x)}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same month, it is
@BeginMath
o(t,x) = \sum\limits_{t=1}^{n} i(t',x)
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_monmean

@Title     = Monthly mean

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same month, it is

o(t,x) = mean{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same month, it is
@BeginMath
o(t,x) = \mbox{\bf mean}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_monavg

@Title     = Monthly average

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same month, it is

o(t,x) = avg{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same month, it is
@BeginMath
o(t,x) = \mbox{\bf avg}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_monstd

@Title     = Monthly standard deviation

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same month, it is

o(t,x) = sqrt{var{i(t',x), t_1 < t' <= t_n}}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same month, it is
@BeginMath
o(t,x) = \sqrt{\mbox{\bf var}\{i(t',x), t_1 < t' \leq t_n\}}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_daymin

@NewTable

@Title     = Daily minimum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same day, it is

o(t,x) = min{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same day, it is
@BeginMath
o(t,x) = \mbox{\bf min}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_daymax

@Title     = Daily maximum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same day, it is

o(t,x) = max{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same day, it is
@BeginMath
o(t,x) = \mbox{\bf max}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_daysum

@Title     = Daily sum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same month, it is

o(t,x) = sum{i(t',x)}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same month, it is
@BeginMath
o(t,x) = \sum\limits_{t=1}^{n} i(t',x)
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_daymean

@Title     = Daily mean

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same day, it is

o(t,x) = mean{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same day, it is
@BeginMath
o(t,x) = \mbox{\bf mean}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_dayavg

@Title     = Daily average

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same day, it is

o(t,x) = avg{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same day, it is
@BeginMath
o(t,x) = \mbox{\bf avg}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_daystd

@Title     = Daily standard deviation

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same day, it is

o(t,x) = sqrt{var{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same day, it is
@BeginMath
o(t,x) = \sqrt{\mbox{\bf var}\{i(t',x), t_1 < t' \leq t_n\}}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_hourmin

@NewTable

@Title     = Hourly minimum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same hour, it is

o(t,x) = min{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same hour, it is
@BeginMath
o(t,x) = \mbox{\bf min}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_hourmax

@Title     = Hourly maximum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same hour, it is

o(t,x) = max{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same hour, it is
@BeginMath
o(t,x) = \mbox{\bf max}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_hoursum

@Title     = Hourly sum

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same hour, it is

o(t,x) = sum{i(t',x)}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same hour, it is
@BeginMath
o(t,x) = \sum\limits_{t=1}^{n} i(t',x)
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_hourmean

@Title     = Hourly mean

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same day, it is

o(t,x) = mean{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same day, it is
@BeginMath
o(t,x) = \mbox{\bf mean}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_houravg

@Title     = Hourly average

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same day, it is

o(t,x) = avg{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same day, it is
@BeginMath
o(t,x) = \mbox{\bf avg}\{i(t',x), t_1 < t' \leq t_n\}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator


@BeginOperator_hourstd

@Title     = Hourly standard deviation

@BeginDesciption
@IfMan
For every adjacent sequence t_1, ...,t_n of field of the same hour, it is

o(t,x) = sqrt{var{i(t',x), t_1<t'<=t_n}
@EndifMan
@IfDoc
For every adjacent sequence \begin{math}t_1, ...,t_n\end{math} of field of the same hour, it is
@BeginMath
o(t,x) = \sqrt{\mbox{\bf var}\{i(t',x), t_1 < t' \leq t_n\}}
@EndMath
@EndifDoc
@EndDesciption

@EndOperator

@EndDoc
*/

#define  NOPERATORS  30

void *Timstat(void *argument)
{
  static char func[] = "Timstat";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  int function[NOPERATORS];
  int intdat[NOPERATORS];
  INT64 intvdat;
  INT64 indate1, indate2 = 0;
  int gridsize;
  int vdate = 0, vtime = 0;
  int vdate0 = 0, vtime0 = 0;
  int nrecs, nrecords;
  int gridID, varID, levelID, recID;
  int tsID;
  int otsID;
  long nsets;
  int i;
  int streamID1, streamID2;
  int vlistID1, vlistID2, taxisID1, taxisID2;
  int nmiss;
  int nvars, nlevel;
  int *recVarID, *recLevelID;
  double missval;
  FIELD **vars1 = NULL, **vars2 = NULL, **samp1 = NULL;
  FIELD field;

  i = 0;
  operators[i] = "timmin";   function[i] = func_min;  intdat[i] = 17; i++;
  operators[i] = "timmax";   function[i] = func_max;  intdat[i] = 17; i++;
  operators[i] = "timsum";   function[i] = func_sum;  intdat[i] = 17; i++;
  operators[i] = "timmean";  function[i] = func_mean; intdat[i] = 17; i++;
  operators[i] = "timavg";   function[i] = func_avg;  intdat[i] = 17; i++;
  operators[i] = "timstd";   function[i] = func_std;  intdat[i] = 17; i++;
  operators[i] = "yearmin";  function[i] = func_min;  intdat[i] =  8; i++;
  operators[i] = "yearmax";  function[i] = func_max;  intdat[i] =  8; i++;
  operators[i] = "yearsum";  function[i] = func_sum;  intdat[i] =  8; i++;
  operators[i] = "yearmean"; function[i] = func_mean; intdat[i] =  8; i++;
  operators[i] = "yearavg";  function[i] = func_avg;  intdat[i] =  8; i++;
  operators[i] = "yearstd";  function[i] = func_std;  intdat[i] =  8; i++;
  operators[i] = "monmin";   function[i] = func_min;  intdat[i] =  6; i++;
  operators[i] = "monmax";   function[i] = func_max;  intdat[i] =  6; i++;
  operators[i] = "monsum";   function[i] = func_sum;  intdat[i] =  6; i++;
  operators[i] = "monmean";  function[i] = func_mean; intdat[i] =  6; i++;
  operators[i] = "monavg";   function[i] = func_avg;  intdat[i] =  6; i++;
  operators[i] = "monstd";   function[i] = func_std;  intdat[i] =  6; i++;
  operators[i] = "daymin";   function[i] = func_min;  intdat[i] =  4; i++;
  operators[i] = "daymax";   function[i] = func_max;  intdat[i] =  4; i++;
  operators[i] = "daysum";   function[i] = func_sum;  intdat[i] =  4; i++;
  operators[i] = "daymean";  function[i] = func_mean; intdat[i] =  4; i++;
  operators[i] = "dayavg";   function[i] = func_avg;  intdat[i] =  4; i++;
  operators[i] = "daystd";   function[i] = func_std;  intdat[i] =  4; i++;
  operators[i] = "hourmin";  function[i] = func_min;  intdat[i] =  2; i++;
  operators[i] = "hourmax";  function[i] = func_max;  intdat[i] =  2; i++;
  operators[i] = "hoursum";  function[i] = func_sum;  intdat[i] =  2; i++;
  operators[i] = "hourmean"; function[i] = func_mean; intdat[i] =  2; i++;
  operators[i] = "houravg";  function[i] = func_avg;  intdat[i] =  2; i++;
  operators[i] = "hourstd";  function[i] = func_std;  intdat[i] =  2; i++;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  intvdat = (INT64) pow(10.0, (double) intdat[operatorID]);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);
  vlistID2 = vlistDuplicate(vlistID1);

  if ( intdat[operatorID] == 17 ) vlistDefNtsteps(vlistID2, 1);

  taxisID1 = vlistInqTaxis(vlistID1);
  taxisID2 = taxisNew(TAXIS_ABSOLUTE);
  vlistDefTaxis(vlistID2, taxisID2);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  nvars    = vlistNvars(vlistID1);
  nrecords = vlistNrecs(vlistID1);

  recVarID   = (int *) malloc(nrecords*sizeof(int));
  recLevelID = (int *) malloc(nrecords*sizeof(int));

  gridsize = vlistGridsizeMax(vlistID1);

  field.ptr = (double *) malloc(gridsize*sizeof(double));

  vars1 = (FIELD **) malloc(nvars*sizeof(FIELD *));
  samp1 = (FIELD **) malloc(nvars*sizeof(FIELD *));
  if ( function[operatorID] == func_std )
    vars2 = (FIELD **) malloc(nvars*sizeof(FIELD *));

  for ( varID = 0; varID < nvars; varID++ )
    {
      gridID   = vlistInqVarGrid(vlistID1, varID);
      gridsize = gridInqSize(gridID);
      nlevel   = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
      missval  = vlistInqVarMissval(vlistID1, varID);

      vars1[varID] = (FIELD *)  malloc(nlevel*sizeof(FIELD));
      samp1[varID] = (FIELD *)  malloc(nlevel*sizeof(FIELD));
      if ( function[operatorID] == func_std )
	vars2[varID] = (FIELD *)  malloc(nlevel*sizeof(FIELD));

      for ( levelID = 0; levelID < nlevel; levelID++ )
	{
	  vars1[varID][levelID].grid    = gridID;
	  vars1[varID][levelID].nmiss   = 0;
	  vars1[varID][levelID].missval = missval;
	  vars1[varID][levelID].ptr     = (double *) malloc(gridsize*sizeof(double));
	  samp1[varID][levelID].grid    = gridID;
	  samp1[varID][levelID].nmiss   = 0;
	  samp1[varID][levelID].missval = missval;
	  samp1[varID][levelID].ptr     = NULL;
	  if ( function[operatorID] == func_std )
	    {
	      vars2[varID][levelID].grid    = gridID;
	      vars2[varID][levelID].nmiss   = 0;
	      vars2[varID][levelID].missval = missval;
	      vars2[varID][levelID].ptr     = (double *) malloc(gridsize*sizeof(double));
	    }
	}
    }

  indate1 = 0;
  tsID    = 0;
  otsID   = 0;
  while ( TRUE )
    {
      nsets = 0;
      while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
	{
	  vdate = taxisInqVdate(taxisID1);
	  vtime = taxisInqVtime(taxisID1);

	  if ( nsets == 0 ) indate2 = (INT64)vdate*10000 + vtime;
	  indate1 = (INT64)vdate*10000 + vtime;

	  if ( indate1/intvdat != indate2/intvdat ) break;

	  for ( recID = 0; recID < nrecs; recID++ )
	    {
	      streamInqRecord(streamID1, &varID, &levelID);

	      if ( tsID == 0 )
		{
		  recVarID[recID]   = varID;
		  recLevelID[recID] = levelID;
		}

	      gridsize = gridInqSize(vlistInqVarGrid(vlistID1, varID));

	      if ( nsets == 0 )
		{
		  streamReadRecord(streamID1, vars1[varID][levelID].ptr, &nmiss);
		  vars1[varID][levelID].nmiss = nmiss;

		  if ( nmiss > 0 || samp1[varID][levelID].ptr )
		    {
		      if ( samp1[varID][levelID].ptr == NULL )
			samp1[varID][levelID].ptr = (double *) malloc(gridsize*sizeof(double));

		      for ( i = 0; i < gridsize; i++ )
			if ( vars1[varID][levelID].ptr[i] == vars1[varID][levelID].missval )
			  samp1[varID][levelID].ptr[i] = 0;
			else
			  samp1[varID][levelID].ptr[i] = 1;
		    }
		}
	      else
		{
		  streamReadRecord(streamID1, field.ptr, &field.nmiss);
		  field.grid    = vars1[varID][levelID].grid;
		  field.missval = vars1[varID][levelID].missval;

		  if ( field.nmiss > 0 || samp1[varID][levelID].ptr )
		    {
		      if ( samp1[varID][levelID].ptr == NULL )
			{
			  samp1[varID][levelID].ptr = (double *) malloc(gridsize*sizeof(double));
			  for ( i = 0; i < gridsize; i++ )
			    samp1[varID][levelID].ptr[i] = nsets;
			}

		      for ( i = 0; i < gridsize; i++ )
			if ( field.ptr[i] != vars1[varID][levelID].missval )
			  samp1[varID][levelID].ptr[i]++;
		    }

		  if ( function[operatorID] == func_std )
		    {
		      farsumq(&vars2[varID][levelID], field);
		      farsum(&vars1[varID][levelID], field);
		    }
		  else
		    {
		      farfun(&vars1[varID][levelID], field, function[operatorID]);
		    }
		}
	    }

	  if ( nsets == 0 && function[operatorID] == func_std )
	    for ( varID = 0; varID < nvars; varID++ )
	      {
		if ( vlistInqVarTime(vlistID1, varID) == TIME_CONSTANT ) continue;
		nlevel   = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
		for ( levelID = 0; levelID < nlevel; levelID++ )
		  farmoq(&vars2[varID][levelID], vars1[varID][levelID]);
	      }

	  vdate0 = vdate;
	  vtime0 = vtime;
	  nsets++;
	  tsID++;
	}

      if ( nrecs == 0 && nsets == 0 ) break;

      if      ( function[operatorID] == func_mean || function[operatorID] == func_avg )
	for ( varID = 0; varID < nvars; varID++ )
	  {
	    if ( vlistInqVarTime(vlistID1, varID) == TIME_CONSTANT ) continue;
	    nlevel   = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
	    for ( levelID = 0; levelID < nlevel; levelID++ )
	      {
		if ( samp1[varID][levelID].ptr == NULL )
		  farcmul(&vars1[varID][levelID], 1.0/nsets);
		else
		  fardiv(&vars1[varID][levelID], samp1[varID][levelID]);
	      }
	  }
      else if ( function[operatorID] == func_std )
	for ( varID = 0; varID < nvars; varID++ )
	  {
	    if ( vlistInqVarTime(vlistID1, varID) == TIME_CONSTANT ) continue;
	    nlevel   = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
	    for ( levelID = 0; levelID < nlevel; levelID++ )
	      {
		if ( samp1[varID][levelID].ptr == NULL )
		  farcstd(&vars1[varID][levelID], vars2[varID][levelID], 1.0/nsets);
		else
		  {
		    farinv(&samp1[varID][levelID]);
		    farstd(&vars1[varID][levelID], vars2[varID][levelID], samp1[varID][levelID]);
		  }
	      }
	  }

      taxisDefVdate(taxisID2, vdate0);
      taxisDefVtime(taxisID2, vtime0);
      streamDefTimestep(streamID2, otsID++);

      for ( recID = 0; recID < nrecords; recID++ )
	{
	  varID   = recVarID[recID];
	  levelID = recLevelID[recID];

	  if ( otsID == 1 || vlistInqVarTime(vlistID1, varID) == TIME_VARIABLE )
	    {
	      streamDefRecord(streamID2, varID, levelID);
	      streamWriteRecord(streamID2, vars1[varID][levelID].ptr,  vars1[varID][levelID].nmiss);
	    }
	}

      if ( nrecs == 0 ) break;
    }

  for ( varID = 0; varID < nvars; varID++ )
    {
      nlevel = zaxisInqSize(vlistInqVarZaxis(vlistID1, varID));
      for ( levelID = 0; levelID < nlevel; levelID++ )
	{
	  free(vars1[varID][levelID].ptr);
	  if ( samp1[varID][levelID].ptr ) free(samp1[varID][levelID].ptr);
	  if ( function[operatorID] == func_std ) free(vars2[varID][levelID].ptr);
	}

      free(vars1[varID]);
      free(samp1[varID]);
      if ( function[operatorID] == func_std ) free(vars2[varID]);
    }

  free(vars1);
  free(samp1);
  if ( function[operatorID] == func_std ) free(vars2);

  if ( field.ptr ) free(field.ptr);

  if ( recVarID   ) free(recVarID);
  if ( recLevelID ) free(recLevelID);

  streamClose(streamID2);
  streamClose(streamID1);

  cdoFinish();

  return (0);
}
