/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"
#include "functs.h"

/*
@BeginDoc

@BeginModule

@Name      = Vertstat
@Title     = Vertical statistic
%@Section   = Statistical description of the fields
@Section   = Statistical description of the data
@Arguments = ifile ofile
@Operators = vertmin vertmax vertsum vertmean vertavg vertstd

@EndModule


@BeginOperator_vertmin

@Title     = Vertical minimum

@BeginDesciption
For every gridpoint the minimum over all levels is computed.
@EndDesciption

@EndOperator


@BeginOperator_vertmax

@Title     = Vertical maximum

@BeginDesciption
For every gridpoint the maximum over all levels is computed.
@EndDesciption

@EndOperator


@BeginOperator_vertsum

@Title     = Vertical sum

@BeginDesciption
For every gridpoint the sum over all levels is computed.
@EndDesciption

@EndOperator


@BeginOperator_vertmean

@Title     = Vertical mean

@BeginDesciption
For every gridpoint the mean over all levels is computed.
@EndDesciption

@EndOperator


@BeginOperator_vertavg

@Title     = Vertical average

@BeginDesciption
For every gridpoint the average over all levels is computed.
@EndDesciption

@EndOperator


@BeginOperator_vertstd

@Title     = Vertical standard deviation

@BeginDesciption
For every gridpoint the standard deviation over all levels is computed.
@EndDesciption

@EndOperator

@EndDoc
*/


#define  NOPERATORS  6

void *Vertstat(void *argument)
{
  static char func[] = "Vertstat";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  int function[NOPERATORS];
  int streamID1, streamID2;
  int vlistID1, vlistID2;
  int gridsize;
  int recID, nrecs;
  int gridID;
  int i;
  int tsID, varID, levelID;
  int nmiss, nvars;
  int zaxisID, nzaxis;
  double missval;
  FIELD *vars1 = NULL, *vars2 = NULL, *samp1 = NULL;
  FIELD field;

  i = 0;
  operators[i] = "vertmin";   function[i] = func_min;   i++;
  operators[i] = "vertmax";   function[i] = func_max;   i++;
  operators[i] = "vertsum";   function[i] = func_sum;   i++;
  operators[i] = "vertmean";  function[i] = func_mean;  i++;
  operators[i] = "vertavg";   function[i] = func_avg;   i++;
  operators[i] = "vertstd";   function[i] = func_std;   i++;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);

  vlistClearFlag(vlistID1);
  nvars = vlistNvars(vlistID1);
  for ( varID = 0; varID < nvars; varID++ )
    vlistDefFlag(vlistID1, varID, 0, TRUE);

  vlistID2 = vlistNew();
  vlistCopyFlag(vlistID2, vlistID1);

  zaxisID = zaxisNew(ZAXIS_SURFACE, 1);
  nzaxis  = vlistNzaxis(vlistID1);
  for ( i = 0; i < nzaxis; i++ )
    if ( zaxisInqSize(vlistZaxis(vlistID1, i)) > 1 )
      vlistChangeZaxisIndex(vlistID2, i, zaxisID);

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridsize = vlistGridsizeMax(vlistID1);

  field.ptr = (double *) malloc(gridsize*sizeof(double));

  vars1 = (FIELD *) malloc(nvars*sizeof(FIELD));
  samp1 = (FIELD *) malloc(nvars*sizeof(FIELD));
  if ( function[operatorID] == func_std )
    vars2 = (FIELD *) malloc(nvars*sizeof(FIELD));

  for ( varID = 0; varID < nvars; varID++ )
    {
      gridID   = vlistInqVarGrid(vlistID1, varID);
      gridsize = gridInqSize(gridID);
      missval  = vlistInqVarMissval(vlistID1, varID);

      vars1[varID].grid    = gridID;
      vars1[varID].nsamp   = 0;
      vars1[varID].nmiss   = 0;
      vars1[varID].missval = missval;
      vars1[varID].ptr     = (double *) malloc(gridsize*sizeof(double));
      samp1[varID].grid    = gridID;
      samp1[varID].nmiss   = 0;
      samp1[varID].missval = missval;
      samp1[varID].ptr     = NULL;
      if ( function[operatorID] == func_std )
	{
	  vars2[varID].grid    = gridID;
	  vars2[varID].nmiss   = 0;
	  vars2[varID].missval = missval;
	  vars2[varID].ptr     = (double *) malloc(gridsize*sizeof(double));
	}
    }

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      streamDefTimestep(streamID2, tsID);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
          vars1[varID].nsamp++;
	  gridsize = gridInqSize(vlistInqVarGrid(vlistID1, varID));
	  if ( levelID == 0 )
	    {
	      streamReadRecord(streamID1, vars1[varID].ptr, &nmiss);
	      vars1[varID].nmiss = nmiss;

	      if ( function[operatorID] == func_std )
		farmoq(&vars2[varID], vars1[varID]);

	      if ( nmiss > 0 || samp1[varID].ptr )
		{
		  if ( samp1[varID].ptr == NULL )
		    samp1[varID].ptr = (double *) malloc(gridsize*sizeof(double));

		  for ( i = 0; i < gridsize; i++ )
		    if ( vars1[varID].ptr[i] == vars1[varID].missval )
		      samp1[varID].ptr[i] = 0;
		    else
		      samp1[varID].ptr[i] = 1;
		}
	    }
	  else
	    {
	      streamReadRecord(streamID1, field.ptr, &field.nmiss);
	      field.grid    = vars1[varID].grid;
	      field.missval = vars1[varID].missval;

	      if ( field.nmiss > 0 || samp1[varID].ptr )
		{
		  if ( samp1[varID].ptr == NULL )
		    {
		      samp1[varID].ptr = (double *) malloc(gridsize*sizeof(double));
		      for ( i = 0; i < gridsize; i++ )
			samp1[varID].ptr[i] = vars1[varID].nsamp;
		    }

		  for ( i = 0; i < gridsize; i++ )
		    if ( field.ptr[i] != vars1[varID].missval )
		      samp1[varID].ptr[i]++;
		}

	      if ( function[operatorID] == func_std )
		{
		  farsumq(&vars2[varID], field);
		  farsum(&vars1[varID], field);
		}
	      else
		{
		  farfun(&vars1[varID], field, function[operatorID]);
		}
	    }
	}

      if      ( function[operatorID] == func_mean || function[operatorID] == func_avg )
	for ( varID = 0; varID < nvars; varID++ )
	  {
	    if ( vlistInqVarTime(vlistID1, varID) == TIME_CONSTANT ) continue;
	    if ( samp1[varID].ptr == NULL )
	      farcmul(&vars1[varID], 1.0/vars1[varID].nsamp);
	    else
	      fardiv(&vars1[varID], samp1[varID]);
	  }
      else if ( function[operatorID] == func_std )
	for ( varID = 0; varID < nvars; varID++ )
	  {
	    if ( vlistInqVarTime(vlistID1, varID) == TIME_CONSTANT ) continue;
	    if ( samp1[varID].ptr == NULL )
	      farcstd(&vars1[varID], vars2[varID], 1.0/vars1[varID].nsamp);
	    else
	      {
		farinv(&samp1[varID]);
		farstd(&vars1[varID], vars2[varID], samp1[varID]);
	      }
	  }

      for ( varID = 0; varID < nvars; varID++ )
	{
	  streamDefRecord(streamID2, varID, 0);
	  streamWriteRecord(streamID2, vars1[varID].ptr, vars1[varID].nmiss);
	  vars1[varID].nsamp = 0;
	}

      tsID++;
    }

  for ( varID = 0; varID < nvars; varID++ )
    {
      free(vars1[varID].ptr);
      if ( samp1[varID].ptr ) free(samp1[varID].ptr);
      if ( function[operatorID] == func_std ) free(vars2[varID].ptr);
    }

  free(vars1);
  free(samp1);
  if ( function[operatorID] == func_std ) free(vars2);

  if ( field.ptr ) free(field.ptr);

  streamClose(streamID2);
  streamClose(streamID1);

  cdoFinish();

  return (0);
}
