/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "pstream.h"
#include "functs.h"

/*
@BeginDoc

@BeginModule

@Name      = Zonstat
@Title     = Zonal statistic
%@Section   = Statistical description of the fields
@Section   = Statistical description of the data
@Arguments = ifile ofile
@Operators = zonmin zonmax zonsum zonmean zonavg zonstd zonvar

@EndModule


@BeginOperator_zonmin

@Title     = Zonal minimum

@BeginDesciption
For every latitude the minimum over all longitudes is computed.
@EndDesciption

@EndOperator

@BeginOperator_zonmax

@Title     = Zonal maximum

@BeginDesciption
For every latitude the maximum over all longitudes is computed.
@EndDesciption

@EndOperator

@BeginOperator_zonsum

@Title     = Zonal sum

@BeginDesciption
For every latitude the sum over all longitudes is computed.
@EndDesciption

@EndOperator

@BeginOperator_zonmean

@Title     = Zonal mean

@BeginDesciption
For every latitude the mean over all longitudes is computed.
@EndDesciption

@EndOperator

@BeginOperator_zonavg

@Title     = Zonal average

@BeginDesciption
For every latitude the average over all longitudes is computed.
@EndDesciption

@EndOperator

@BeginOperator_zonvar

@Title     = Zonal variance

@BeginDesciption
For every latitude the variance over all longitudes is computed.
@EndDesciption

@EndOperator

@BeginOperator_zonstd

@Title     = Zonal standard deviation

@BeginDesciption
For every latitude the standard deviation over all longitudes is computed.
@EndDesciption

@EndOperator

@EndDoc
*/


#define  NOPERATORS  7

void *Zonstat(void *argument)
{
  static char func[] = "Zonstat";
  const int noperators = NOPERATORS;
  char *operators[NOPERATORS];
  int operatorID;
  int function[NOPERATORS];
  int streamID1, streamID2;
  int vlistID1, vlistID2;
  int gridID1, gridID2;
  int nlatmax;
  int index, ngrids;
  int recID, nrecs;
  int i;
  int tsID, varID, levelID;
  int lim;
  FIELD field1, field2;

  i = 0;
  operators[i] = "zonmin";   function[i++] = func_min;
  operators[i] = "zonmax";   function[i++] = func_max;
  operators[i] = "zonsum";   function[i++] = func_sum;
  operators[i] = "zonmean";  function[i++] = func_mean;
  operators[i] = "zonavg";   function[i++] = func_avg;
  operators[i] = "zonvar";   function[i++] = func_var;
  operators[i] = "zonstd";   function[i++] = func_std;

  cdoInitialize(argument);

  operatorID = cdoOperatorID(noperators, (const char **) operators);

  streamID1 = streamOpenRead(cdoStreamName(0));
  if ( streamID1 < 0 ) cdiError(streamID1, "Open failed on %s", cdoStreamName(0));

  vlistID1 = streamInqVlist(streamID1);

  vlistID2 = vlistDuplicate(vlistID1);

  ngrids = vlistNgrids(vlistID1);
  index = 0;
  gridID1 = vlistGrid(vlistID1, index);
  if ( gridInqType(gridID1) != GRID_LONLAT &&
       gridInqType(gridID1) != GRID_GAUSSIAN )
    cdoAbort("Unsupported gridtype: %s", gridNamePtr(gridInqType(gridID1)));

  gridID2 = gridToZonal(gridID1);
  vlistChangeGridIndex(vlistID2, index, gridID2);
  if ( ngrids > 1 ) cdoAbort("Too many different grids!");

  streamID2 = streamOpenWrite(cdoStreamName(1), cdoFiletype());
  if ( streamID2 < 0 ) cdiError(streamID2, "Open failed on %s", cdoStreamName(1));

  streamDefVlist(streamID2, vlistID2);

  gridID1 = vlistInqVarGrid(vlistID1, 0);
  nlatmax = gridInqYsize(gridID1); /* max nlat ? */

  lim = vlistGridsizeMax(vlistID1);
  field1.ptr  = (double *) malloc(lim*sizeof(double));
  field2.ptr  = (double *) malloc(nlatmax*sizeof(double));
  field2.grid = gridID2;

  tsID = 0;
  while ( (nrecs = streamInqTimestep(streamID1, tsID)) )
    {
      streamDefTimestep(streamID2, tsID);

      for ( recID = 0; recID < nrecs; recID++ )
	{
	  streamInqRecord(streamID1, &varID, &levelID);
	  streamReadRecord(streamID1, field1.ptr, &field1.nmiss);

	  field1.grid    = vlistInqVarGrid(vlistID1, varID);
	  field1.missval = vlistInqVarMissval(vlistID1, varID);
	  field2.missval = vlistInqVarMissval(vlistID1, varID);

	  zonfun(field1, &field2, function[operatorID]);

	  streamDefRecord(streamID2, varID,  levelID);
	  streamWriteRecord(streamID2, field2.ptr, field2.nmiss);
	}
      tsID++;
    }

  streamClose(streamID2);
  streamClose(streamID1);

  if ( field1.ptr )    free(field1.ptr);
  if ( field2.ptr )    free(field2.ptr);

  cdoFinish();

  return (0);
}
