/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <float.h>
#include <math.h>

#include "cdo.h"
#include "cdi.h"
#include "functs.h"
#include "field.h"

void fldfun(FIELD field1, FIELD *field2, int function)
{
  if      ( function == func_min )  fldmin(field1, field2);
  else if ( function == func_max )  fldmax(field1, field2);  
  else if ( function == func_sum )  fldsum(field1, field2);  
  else if ( function == func_mean ) fldmean(field1, field2);  
  else if ( function == func_avg )  fldavg(field1, field2);  
  else if ( function == func_std )  fldstd(field1, field2);  
  else if ( function == func_var )  fldvar(field1, field2);
  else cdoAbort("function %d not implemented!", function);
}

void fldmin(FIELD field1, FIELD *field2)
{
  int i, len, rnmiss = 0;
  int    grid    = field1.grid;
  int    nmiss   = field1.nmiss;
  double missval = field1.missval;
  double *array  = field1.ptr;
  double rmin = 0;

  len    = gridInqSize(grid);

  if ( nmiss > 0 )
    {
      rmin = DBL_MAX;
      for ( i = 0; i < len; i++ ) 
	if ( array[i] != missval )
	  if ( array[i] < rmin ) rmin = array[i];

      if ( rmin == DBL_MAX )
	{
	  rnmiss++;
	  rmin = missval;
	}
    }
  else
    {
      rmin = array[0];
      for ( i = 1; i < len; i++ ) 
	if ( array[i] < rmin )  rmin = array[i];
    }

  field2->ptr[0] = rmin;
  field2->nmiss  = rnmiss;
}

void fldmax(FIELD field1, FIELD *field2)
{
  int i, len, rnmiss = 0;
  int    grid    = field1.grid;
  int    nmiss   = field1.nmiss;
  double missval = field1.missval;
  double *array  = field1.ptr;
  double rmax = 0;

  len    = gridInqSize(grid);

  if ( nmiss > 0 )
    {
      rmax = DBL_MIN;
      for ( i = 0; i < len; i++ ) 
	if ( array[i] != missval )
	  if ( array[i] > rmax ) rmax = array[i];

      if ( rmax == DBL_MIN )
	{
	  rnmiss++;
	  rmax = missval;
	}
    }
  else
    {
      rmax = array[0];
      for ( i = 1; i < len; i++ ) 
	if ( array[i] > rmax )  rmax = array[i];
    }

  field2->ptr[0] = rmax;
  field2->nmiss  = rnmiss;
}

void fldsum(FIELD field1, FIELD *field2)
{
  int i, len, rnmiss = 0;
  int    grid    = field1.grid;
  int    nmiss   = field1.nmiss;
  double missval = field1.missval;
  double *array  = field1.ptr;
  double rsum = 0;

  len    = gridInqSize(grid);

  if ( nmiss > 0 )
    {
      for ( i = 0; i < len; i++ ) 
	if ( array[i] != missval )
	  rsum += array[i];
    }
  else
    {
      for ( i = 0; i < len; i++ ) 
	rsum += array[i];
    }

  field2->ptr[0] = rsum;
  field2->nmiss  = rnmiss;
}

void fldmean(FIELD field1, FIELD *field2)
{
  int i, len, rnmiss = 0;
  int    grid     = field1.grid;
  int    nmiss    = field1.nmiss;
  double missval1 = field1.missval;
  double missval2 = field1.missval;
  double *array   = field1.ptr;
  double *w       = field1.weight;
  double rsum = 0, rsumw = 0, ravg = 0;

  len    = gridInqSize(grid);

  if ( nmiss > 0 )
    {
      for ( i = 0; i < len; i++ ) 
	if ( array[i] != missval1 && w[i] != missval1 )
	  {
	    rsum  += w[i] * array[i];
	    rsumw += w[i];
	  }
    }
  else
    {
      for ( i = 0; i < len; i++ ) 
	{
	  rsum  += w[i] * array[i];
	  rsumw += w[i];
	}
    }

  ravg = DIV(rsum, rsumw);

  if ( ravg == missval1 ) rnmiss++;

  field2->ptr[0] = ravg;
  field2->nmiss  = rnmiss;
}

void fldavg(FIELD field1, FIELD *field2)
{
  int i, len, rnmiss = 0;
  int    grid     = field1.grid;
  int    nmiss    = field1.nmiss;
  double missval1 = field1.missval;
  double missval2 = field1.missval;
  double *array   = field1.ptr;
  double *w       = field1.weight;
  double rsum = 0, rsumw = 0, ravg = 0;

  len    = gridInqSize(grid);

  if ( nmiss > 0 )
    {
      for ( i = 0; i < len; i++ ) 
	if ( w[i] != missval1 )
	  {
	    rsum  = ADD(rsum, MUL(w[i], array[i]));
	    rsumw = ADD(rsumw, w[i]);
	  }
    }
  else
    {
      for ( i = 0; i < len; i++ ) 
	{
	  rsum  += w[i] * array[i];
	  rsumw += w[i];
	}
    }

  ravg = DIV(rsum, rsumw);

  if ( ravg == missval1 ) rnmiss++;

  field2->ptr[0] = ravg;
  field2->nmiss  = rnmiss;
}

void fldvar(FIELD field1, FIELD *field2)
{
  int i, len, rnmiss = 0;
  int    grid    = field1.grid;
  int    nmiss   = field1.nmiss;
  double missval = field1.missval;
  double *array  = field1.ptr;
  double *w      = field1.weight;
  double rsum = 0, rsumw = 0, rvar = 0;
  double rsumq = 0, rsumwq = 0;

  len    = gridInqSize(grid);

  if ( nmiss > 0 )
    {
      for ( i = 0; i < len; i++ ) 
	if ( array[i] != missval && w[i] != missval )
	  {
	    rsum   += w[i] * array[i];
	    rsumq  += w[i] * array[i] * array[i];
	    rsumw  += w[i];
	    rsumwq += w[i] * w[i];
	  }
    }
  else
    {
      for ( i = 0; i < len; i++ ) 
	{
	  rsum   += w[i] * array[i];
	  rsumq  += w[i] * array[i] * array[i];
	  rsumw  += w[i];
	  rsumwq += w[i] * w[i];
	}
    }

  rvar = rsumw ? (rsumq*rsumw - rsum*rsum) / (rsumw*rsumw) : missval;

  if ( rvar == missval ) rnmiss++;

  field2->ptr[0] = rvar;
  field2->nmiss  = rnmiss;
}

void fldstd(FIELD field1, FIELD *field2)
{
  int rnmiss = 0;
  double missval = field1.missval;
  double rvar, rstd;

  fldvar(field1, field2);

  rvar = field2->ptr[0];
  rstd = (rvar && rvar != missval) ? sqrt(rvar) : missval;

  if ( rvar == missval ) rnmiss++;

  field2->ptr[0] = rstd;
  field2->nmiss  = rnmiss;
}

void fldrms(FIELD field1, FIELD field2, FIELD *field3)
{
  int i, len, rnmiss = 0;
  int    grid1    = field1.grid;
  int    nmiss1   = field1.nmiss;
  double *array1  = field1.ptr;
  int    grid2    = field2.grid;
  int    nmiss2   = field2.nmiss;
  double *array2  = field2.ptr;
  double missval1 = field1.missval;
  double missval2 = field2.missval;
  double *w       = field1.weight;
  double rsum = 0, rsumw = 0, ravg = 0;

  len    = gridInqSize(grid1);
  if ( len != gridInqSize(grid2) )
    cdoAbort("fields have different size!");

  /*
  if ( nmiss1 > 0 )
  */
    {
      for ( i = 0; i < len; i++ ) 
	if ( w[i] != missval1 )
	  {
	    rsum  = ADD(rsum, MUL(w[i], MUL(SUB(array2[i], array1[i]),
                                            SUB(array2[i], array1[i]))));
	    rsumw = ADD(rsumw, w[i]);
	  }
    }
    /*
  else
    {
      for ( i = 0; i < len; i++ ) 
	{
	  rsum  += w[i] * array1[i];
	  rsumw += w[i];
	}
    }
    */

  ravg = ROOT(DIV(rsum, rsumw));

  if ( ravg == missval1 ) rnmiss++;

  field3->ptr[0] = ravg;
  field3->nmiss  = rnmiss;
}

void varrms(FIELD field1, FIELD field2, FIELD *field3)
{
  int i, k, nlev, len, rnmiss = 0;
  int    zaxis    = field1.zaxis;
  int    grid1    = field1.grid;
  int    nmiss1   = field1.nmiss;
  double *array1  = field1.ptr;
  int    grid2    = field2.grid;
  int    nmiss2   = field2.nmiss;
  double *array2  = field2.ptr;
  double missval1 = field1.missval;
  double missval2 = field2.missval;
  double *w       = field1.weight;
  double rsum = 0, rsumw = 0, ravg = 0;

  nlev   = zaxisInqSize(zaxis);
  len    = gridInqSize(grid1);
  if ( len != gridInqSize(grid2) )
    cdoAbort("fields have different size!");

  /*
  if ( nmiss1 > 0 )
  */
    {
      for ( k = 0; k < nlev; k++ )
	for ( i = 0; i < len; i++ )
	  /*	  if ( w[i] != missval1 ) */
	    {
	      rsum  = ADD(rsum, MUL(w[i], MUL(SUB(array2[k*len+i], array1[k*len+i]),
                                              SUB(array2[k*len+i], array1[k*len+i]))));
	      rsumw = ADD(rsumw, w[i]);
	    }
    }
    /*
  else
    {
      for ( i = 0; i < len; i++ )
	{
	  rsum  += w[i] * array1[i];
	  rsumw += w[i];
	}
    }
    */

  ravg = ROOT(DIV(rsum, rsumw));

  if ( ravg == missval1 ) rnmiss++;

  field3->ptr[0] = ravg;
  field3->nmiss  = rnmiss;
}

double minval(FIELD field)
{
  int i, k, nlev, len, rnmiss = 0;
  int    grid    = field.grid;
  int    nmiss   = field.nmiss;
  double *array  = field.ptr;
  double missval = field.missval;
  double rmin    = 0;

  len    = gridInqSize(grid);
  if ( nmiss > 0 )
    {
      rmin = DBL_MAX;
      for ( i = 0; i < len; i++ ) 
	if ( array[i] != missval )
	  if ( array[i] < rmin ) rmin = array[i];

      if ( rmin == DBL_MAX )
	{
	  rnmiss++;
	  rmin = missval;
	}
    }
  else
    {
      rmin = array[0];
      for ( i = 1; i < len; i++ ) 
	if ( array[i] < rmin )  rmin = array[i];
    }

  return (rmin);
}

double maxval(FIELD field)
{
}
