/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate Data files.

  Copyright (C) 2003-2005 Uwe Schulzweida, schulzweida@dkrz.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#if  defined  (HAVE_CONFIG_H)
#  include "config.h"
#endif

#include <string.h>
#include <stdarg.h>
#include <errno.h>

#include "cdi.h"
#include "cdo.h"
#include "modules.h"
#include "pstream_int.h"
#include "cdo_int.h"
#include "util.h"
#include "process.h"
#include "pipe.h"
#include "error.h"
#include "dmemory.h"

static int StreamDebug = 0;

PSTREAMTABLE *pstreamTable = 0;
static int pstreamTableSize = 0;

#if  defined  (HAVE_PTHREAD_H)
pthread_mutex_t streamOpenReadMutex  = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t streamOpenWriteMutex = PTHREAD_MUTEX_INITIALIZER;
#endif


static void pstreamTableInitEntry(int streamID)
{
  static char func[] = "pstreamTableInitEntry";

  if ( streamID < 0 || streamID >= pstreamTableSize )
    Error(func, "streamID %d out of range!", streamID);

  pstreamTable[streamID].used       = FALSE;
  pstreamTable[streamID].fileID     = -1;
  pstreamTable[streamID].vlistID    = -1;
  pstreamTable[streamID].tsID       = -1;
  pstreamTable[streamID].filetype   = -1;
  pstreamTable[streamID].ispipe     = FALSE;
  pstreamTable[streamID].isopen     = FALSE;
  pstreamTable[streamID].name       = NULL;
#if  defined  (HAVE_PTHREAD_H)
  pstreamTable[streamID].pipe       = NULL;
#endif
  pstreamTable[streamID].rthreadID  = 0;
  pstreamTable[streamID].wthreadID  = 0;
}

static int pstreamTableNewEntry(void)
{
  static char func[] = "pstreamTableNewEntry";
  int pstreamTableMax = 1024;
  int streamID = 0;

  /*
    Look for a free slot in pstreamTable.
    (Create the table the first time through).
  */
  if ( ! pstreamTableSize )
    {
      int i;

      pstreamTableSize = 2;
      pstreamTable = (PSTREAMTABLE *) malloc(pstreamTableSize*sizeof(PSTREAMTABLE));
      if( pstreamTable == NULL )
	{
          Message(func, "pstreamTableSize = %d", pstreamTableSize);
	  SysError(func, "Allocation of PSTREAMTABLE failed");
	}

      for( i = 0; i < pstreamTableSize; i++ )
	pstreamTableInitEntry(i);
    }
  else
    {
      while( streamID < pstreamTableSize )
	{
	  if ( ! pstreamTable[streamID].used ) break;
	  streamID++;
	}
    }
  /*
    If the table overflows, double its size.
  */
  if( streamID == pstreamTableSize )
    {
      int i;

      if ( pstreamTableSize == pstreamTableMax )
	Error(func, "limit of %d streams reached", pstreamTableMax);

      pstreamTableSize = 2*pstreamTableSize;
      pstreamTable = (PSTREAMTABLE *) realloc(pstreamTable, pstreamTableSize*sizeof(PSTREAMTABLE));
      if( pstreamTable == NULL )
	{
          Message(func, "pstreamTableSize = %d", pstreamTableSize);
	  SysError(func, "Reallocation of PSTREAMTABLE failed");
	}
      streamID = pstreamTableSize/2;

      for( i = streamID; i < pstreamTableSize; i++ )
	pstreamTableInitEntry(i);
    }

  pstreamTableInitEntry(streamID);

  pstreamTable[streamID].used = 1;
  pstreamTable[streamID].isopen = TRUE;

  return (streamID);
}

static int pstreamFindID(const char *name)
{
  int streamID;

  for ( streamID = 0; streamID < pstreamTableSize; streamID++ )
    {
      if ( pstreamTable[streamID].used )
	if ( strcmp(pstreamTable[streamID].name, name) == 0 ) break;
    }

  if ( streamID == pstreamTableSize ) pstreamTableSize = -1;

  return (streamID);
}

int pstreamIsPipe(int pstreamID)
{
  return (pstreamTable[pstreamID].ispipe);
}

int pstreamOpenRead(const char *argument)
{
  static char func[] = "pstreamOpenRead";
  char *operatorArg = NULL;
  char *operatorName = NULL;
  int ispipe = FALSE;
  int fileID;
  int streamID;

  streamID = pstreamTableNewEntry();

  ispipe = argument[0] == '-';

  if ( ispipe )
    {
#if  defined  (HAVE_LIBPTHREAD)
      char *newarg;
      char *pipename = (char *) malloc(10);
      int rval;
      pthread_t thrID;
      size_t len;

      operatorArg = getOperator(argument);
      operatorName = getOperatorName(operatorArg);
      free(operatorArg);

      len = strlen(argument);
      newarg = (char *) malloc(len+11);
      strcpy(newarg, argument);
      sprintf(pipename, "(pipe%d.%d)", processSelf() + 1, processInqChildNum() + 1);
      newarg[len] = ' ';
      strcpy(&newarg[len+1], pipename);

      pstreamTable[streamID].ispipe = TRUE;
      pstreamTable[streamID].name   = pipename;
      pstreamTable[streamID].pipe   = (PIPE *) malloc(sizeof(PIPE));
      pstreamTable[streamID].rthreadID = (int) pthread_self();
      pipeInit(pstreamTable[streamID].pipe);
 
      fprintf(stderr, "%s: Started child process \"%s\".\n", processInqPrompt(), newarg+1);
   
      rval = pthread_create(&thrID, NULL, operatorModule(operatorName), newarg);
      if ( rval != 0 )
	{
	  errno = rval;
	  SysError(func, "pthread_create failed for '%s'\n", newarg+1);
	}

      /* free(operatorName); */
      processAddStream(streamID);
      /*      pipeInqInfo(streamID); */
      if ( StreamDebug ) Message(func, "pipe %s", pipename);
#else
      cdoAbort("Can't use pipes, pthread library not available!");
#endif
    }
  else
    {
      extern int cdoDefaultFileType/*, cdoDefaultInstID*/;
      char *filename = (char *) malloc(strlen(argument)+1);

      if ( StreamDebug ) Message(func, "file %s", argument);

#if  defined  (HAVE_PTHREAD_H)
      pthread_mutex_lock(&streamOpenReadMutex);
#endif
      fileID = streamOpenRead(argument);
#if  defined  (HAVE_PTHREAD_H)
      pthread_mutex_unlock(&streamOpenReadMutex);
#endif
      if ( fileID < 0 ) cdiError(fileID, "Open failed on %s", argument);

      if ( cdoDefaultFileType == CDI_UNDEFID )
	cdoDefaultFileType = streamInqFiletype(fileID);
      /*
      if ( cdoDefaultInstID == CDI_UNDEFID )
	cdoDefaultInstID = streamInqInstID(fileID);
      */
      cdoInqHistory(fileID);

      pstreamTable[streamID].ispipe = FALSE;
      strcpy(filename, argument);
      pstreamTable[streamID].name   = filename;
      pstreamTable[streamID].fileID = fileID;
    }

  return (streamID);
}

int pstreamOpenWrite(const char *argument, int filetype)
{
  static char func[] = "pstreamOpenWrite";
  int fileID;
  int pstreamID;
  int ispipe;

  ispipe = strncmp(argument, "(pipe", 5) == 0;

  if ( ispipe )
    {
      if ( StreamDebug ) Message(func, "pipe %s", argument);
      pstreamID = pstreamFindID(argument);
      if ( pstreamID == -1 )
	Error(func, "%s not open", argument);

      pstreamTable[pstreamID].wthreadID = (int) pthread_self();
      pstreamTable[pstreamID].filetype = filetype;
      processAddStream(pstreamID);
    }
  else
    {
      /* extern int cdoDefaultInstID; */
      char *filename = (char *) malloc(strlen(argument)+1);

      pstreamID = pstreamTableNewEntry();
      if ( StreamDebug ) Message(func, "file %s", argument);

      if ( filetype == CDI_UNDEFID ) filetype = FILETYPE_GRB;

#if  defined  (HAVE_PTHREAD_H)
      pthread_mutex_lock(&streamOpenWriteMutex);
#endif
      fileID = streamOpenWrite(argument, filetype);
#if  defined  (HAVE_PTHREAD_H)
      pthread_mutex_unlock(&streamOpenWriteMutex);
#endif
      if ( fileID < 0 ) cdiError(fileID, "Open failed on %s", argument);

      cdoDefHistory(fileID, commandLine());

      if ( cdoDefaultByteorder != CDI_UNDEFID )
	streamDefByteorder(fileID, cdoDefaultByteorder);
      /*
      if ( cdoDefaultInstID != CDI_UNDEFID )
	streamDefInstID(fileID, cdoDefaultInstID);
      */
      pstreamTable[pstreamID].ispipe = FALSE;
      strcpy(filename, argument);
      pstreamTable[pstreamID].name   = filename;
      pstreamTable[pstreamID].fileID = fileID;
    }

  return (pstreamID);
}

int pstreamOpenAppend(const char *argument)
{
  static char func[] = "pstreamOpenAppend";
  int fileID;
  int pstreamID = -1;
  int ispipe;

  ispipe = strncmp(argument, "(pipe", 5) == 0;

  if ( ispipe )
    {
      if ( StreamDebug ) Message(func, "pipe %s", argument);
      cdoAbort("this operator doesn't work with pipes!");
    }
  else
    {
      char *filename = (char *) malloc(strlen(argument)+1);

      pstreamID = pstreamTableNewEntry();
      if ( StreamDebug ) Message(func, "file %s", argument);

      fileID = streamOpenAppend(argument);
      if ( fileID < 0 ) cdiError(fileID, "Open failed on %s", argument);
      /*
      cdoInqHistory(fileID);
      cdoDefHistory(fileID, commandLine());
      */
      pstreamTable[pstreamID].ispipe = FALSE;
      strcpy(filename, argument);
      pstreamTable[pstreamID].name   = filename;
      pstreamTable[pstreamID].fileID = fileID;
    }

  return (pstreamID);
}

void pstreamClose(int pstreamID)
{
  static char func[] = "pstreamClose";

  if ( pstreamTable[pstreamID].ispipe )
    {
#if  defined  (HAVE_LIBPTHREAD)
      PIPE *pipe;
      int lread = FALSE, lwrite = FALSE;
      int threadID = (int) pthread_self();

      if      ( threadID == pstreamTable[pstreamID].rthreadID ) lread  = TRUE;
      else if ( threadID == pstreamTable[pstreamID].wthreadID ) lwrite = TRUE;
      else Error(func, "Internal problem! Close pipe %s", pstreamTable[pstreamID].name);

      if ( lread )
	{
	  int cthreadID;
	  cthreadID = pstreamTable[pstreamID].wthreadID;
	  pipe = pstreamTable[pstreamID].pipe;
	  pthread_mutex_lock(pipe->mutex);
	  pipe->EOP = TRUE;
	  if ( StreamDebug )
	    Message(func, "%s read closed", pstreamTable[pstreamID].name);
	  pthread_mutex_unlock(pipe->mutex);     
	  pthread_cond_signal(pipe->tsDef);
	  pthread_cond_signal(pipe->tsInq);

	  pthread_mutex_lock(pipe->mutex);
	  pstreamTable[pstreamID].isopen = FALSE;
	  pthread_mutex_unlock(pipe->mutex);     
	  pthread_cond_signal(pipe->isclosed);

	  pthread_join((pthread_t) cthreadID, NULL);
	}
      else
	{
	  pipe = pstreamTable[pstreamID].pipe;
	  pthread_mutex_lock(pipe->mutex);
	  pipe->EOP = TRUE;
	  if ( StreamDebug )
	    Message(func, "%s write closed", pstreamTable[pstreamID].name);
	  pthread_mutex_unlock(pipe->mutex);     
	  pthread_cond_signal(pipe->tsDef);
	  pthread_cond_signal(pipe->tsInq);

	  pthread_mutex_lock(pipe->mutex);
	  while ( pstreamTable[pstreamID].isopen )
	    {
	      if ( StreamDebug )
		Message(func, "wait of read close");
	      pthread_cond_wait(pipe->isclosed, pipe->mutex);
	    }
	  pthread_mutex_unlock(pipe->mutex);
   
	  pstreamTable[pstreamID].used = 0;
	}
#else
      cdoAbort("Can't use pipes, pthread library not available!");
#endif
    }
  else
    {
      if ( StreamDebug )
	Message(func, "%s fileID %d\n",
		pstreamTable[pstreamID].name, pstreamTable[pstreamID].fileID);

      streamClose(pstreamTable[pstreamID].fileID);

      pstreamTable[pstreamID].isopen = FALSE;
      pstreamTable[pstreamID].used = 0;
    }
}

int pstreamInqVlist(int pstreamID)
{
  int vlistID;

#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    vlistID = pipeInqVlist(pstreamID);
  else
#endif
    {
      extern int cdoDefaultTimeType;
      vlistID = streamInqVlist(pstreamTable[pstreamID].fileID);
      if ( cdoDefaultTimeType != CDI_UNDEFID )
	taxisDefType(vlistInqTaxis(vlistID), cdoDefaultTimeType);
    }

  pstreamTable[pstreamID].vlistID = vlistID;

  processDefVarNum(vlistNvars(vlistID), pstreamID);

  return (vlistID);
}

const char *cdoComment(void)
{
  static char comment[256];
  static int init = 0;
  int size = 0;
  extern char CDO_Version[];

  if ( ! init )
    {
      init = 1;

      size = strlen(CDO_Version);

      strncat(comment, CDO_Version, size);
    }

  return (comment);
}

void pstreamDefVlist(int pstreamID, int vlistID)
{
  pstreamTable[pstreamID].vlistID = vlistID;

#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    {
      /*    pipeDefVlist(pstreamID, vlistID);*/
      pipeDefVlist(pstreamID, vlistDuplicate(vlistID));
    }
  else
#endif
    {
      extern int cdoDefaultDataType;
      if ( cdoDefaultDataType != CDI_UNDEFID )
	{
	  int varID, nvars = vlistNvars(vlistID);
	  for ( varID = 0; varID < nvars; varID++ )
	    vlistDefVarDatatype(vlistID, varID, cdoDefaultDataType);
	}
      vlistDefAttribute(vlistID, "CDO", cdoComment());
      streamDefVlist(pstreamTable[pstreamID].fileID, vlistID);
    }
}

int pstreamInqRecord(int pstreamID, int *varID, int *levelID)
{
#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    pipeInqRecord(pstreamID, varID, levelID);
  else
#endif
    streamInqRecord(pstreamTable[pstreamID].fileID, varID, levelID);

  return (0);
}

void pstreamDefRecord(int pstreamID, int varID, int levelID)
{
#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    pipeDefRecord(pstreamID, varID, levelID);
  else
#endif
    streamDefRecord(pstreamTable[pstreamID].fileID, varID, levelID);
}

void pstreamReadRecord(int pstreamID, double *data, int *nmiss)
{
#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    pipeReadRecord(pstreamID, data, nmiss);
  else
#endif
    streamReadRecord(pstreamTable[pstreamID].fileID, data, nmiss);
}

void pstreamWriteRecord(int pstreamID, double *data, int nmiss)
{
#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    pipeWriteRecord(pstreamID, data, nmiss);
  else
#endif
    streamWriteRecord(pstreamTable[pstreamID].fileID, data, nmiss);
}

int pstreamInqTimestep(int pstreamID, int tsID)
{
  int nrecs = 0;

#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    nrecs = pipeInqTimestep(pstreamID, tsID);
  else
#endif
    {
      extern int cdoDefaultTimeType;
      nrecs = streamInqTimestep(pstreamTable[pstreamID].fileID, tsID);
      if ( cdoDefaultTimeType != CDI_UNDEFID )
	taxisDefType(vlistInqTaxis(pstreamTable[pstreamID].vlistID), cdoDefaultTimeType);
    }

  if ( nrecs && tsID != pstreamTable[pstreamID].tsID )
    {
      processDefTimesteps(pstreamID);
      pstreamTable[pstreamID].tsID = tsID;
    }

  return (nrecs);
}

void pstreamDefTimestep(int pstreamID, int tsID)
{
#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    pipeDefTimestep(pstreamID, tsID);
  else
#endif
    {
      extern int cdoDefaultTimeType;
      if ( cdoDefaultTimeType != CDI_UNDEFID )
	{
	  int taxisID, vlistID;
	  vlistID = pstreamTable[pstreamID].vlistID;
	  taxisID = vlistInqTaxis(vlistID);
      	  taxisDefType(taxisID, cdoDefaultTimeType);
	}

      streamDefTimestep(pstreamTable[pstreamID].fileID, tsID);
    }
}

void pstreamCopyRecord(int pstreamIDdest, int pstreamIDsrc)
{
  static char func[] = "pstreamCopyRecord";

  if ( StreamDebug )
    Message(func, "pstreamIDdest = %d  pstreamIDsrc = %d", pstreamIDdest, pstreamIDsrc);

#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamIDdest].ispipe || pstreamTable[pstreamIDsrc].ispipe )
    pipeCopyRecord(pstreamIDdest, pstreamIDsrc);
  else
#endif
    streamCopyRecord(pstreamTable[pstreamIDdest].fileID, pstreamTable[pstreamIDsrc].fileID);
}

void pstreamDebug(int debug)
{
  StreamDebug = debug;
}

void cdoInitialize(void *argument)
{
  static char func[] = "cdoInitialize";
  int processID;

  processID = processCreate();

#if  defined  (HAVE_PTHREAD_H)
  if ( StreamDebug )
     Message(func, "process %d  thread %d", processSelf(), (int) pthread_self());
#endif

  processDefArgument((const char*) argument);
}

void cdoFinish(void)
{
  static char func[] = "cdoFinish";
  int processID = processSelf();
  int sindex, streamID;
  int nstream;
  int nvars, ntimesteps;
  char memstring[32] = {""};
  double s_utime, s_stime;
  double e_utime, e_stime;
  double c_cputime = 0, c_usertime = 0, c_systime = 0;
  double p_cputime = 0, p_usertime = 0, p_systime = 0;

  if ( StreamDebug )
    Message(func, "process %d  thread %d", processID, (int) pthread_self());

  nvars = processInqVarNum();
  ntimesteps = processInqTimesteps();

  fprintf(stderr, "%s : Processed %d variable%s %d timestep%s.",
	  processInqPrompt(), nvars, nvars > 1 ? "s" : "",
	  ntimesteps, ntimesteps > 1 ? "s" : "");

  processStartTime(&s_utime, &s_stime);
  cdoTimer(&e_utime, &e_stime);

  c_usertime = e_utime - s_utime;
  c_systime  = e_stime - s_stime;
  c_cputime  = c_usertime + c_systime;

  processAccuTime(c_usertime, c_systime);

  if ( processID == 0 )
    {
      int mu[] = {'B', 'K', 'M', 'G', 'T'};
      int muindex = 0;
      long memmax;

      memmax = memTotal();
      while ( memmax > 9999 )
	{
	  memmax /= 1024;
	  muindex++;
	}

      sprintf(memstring, " %ld%c ", memmax, mu[muindex]);

      processEndTime(&p_usertime, &p_systime);
      p_cputime  = p_usertime + p_systime;

      userlog(processInqPrompt(), p_cputime); 
    }

  if ( cdoBenchmark )
    fprintf(stderr, " ( %.2fs %.2fs %.2fs %s)\n", c_usertime, c_systime, c_cputime, memstring);
  else
    fprintf(stderr, " ( %.2fs %s)\n", c_cputime, memstring);

  if ( cdoBenchmark && processID == 0 )
    fprintf(stderr, "total: user %.2fs  sys %.2fs  cpu %.2fs  mem%s\n",
	    p_usertime, p_systime, p_cputime, memstring);

  nstream = processInqStreamNum();
  for ( sindex = 0; sindex < nstream; sindex++ )
    {
      streamID = processInqStreamID(sindex);
      if ( StreamDebug )
	Message(func, "process %d  stream %d  close streamID %d",
		processID, sindex, streamID);
      if ( pstreamTable[streamID].isopen ) streamClose(streamID);
    }
}

int pstreamInqFiletype(int pstreamID)
{
  int filetype;

#if  defined  (HAVE_PTHREAD_H)
  if ( pstreamTable[pstreamID].ispipe )
    filetype = pstreamTable[pstreamID].filetype;
  else
#endif
    filetype = streamInqFiletype(pstreamTable[pstreamID].fileID);

  return (filetype);
}
