#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>

#include "cdi.h"
#include "cdo.h"
#include "cdo_int.h"
#include "remap.h"


void remapGridFree(REMAPGRID *rg)
{
  static char func[] = "remapGridFree";

  if ( rg->pinit == TRUE )
    {
      rg->pinit = FALSE;

      free(rg->grid1_mask);
      free(rg->grid2_mask);
      free(rg->grid1_center_lat);
      free(rg->grid1_center_lon);
      free(rg->grid2_center_lat);
      free(rg->grid2_center_lon);
      free(rg->grid1_area);
      free(rg->grid2_area);
      free(rg->grid1_frac);
      free(rg->grid2_frac);

      free(rg->grid1_corner_lat);
      free(rg->grid1_corner_lon);
      free(rg->grid2_corner_lat);
      free(rg->grid2_corner_lon);

      free(rg->grid1_bound_box);
      free(rg->grid2_bound_box);

      free(rg->bin_addr1);
      free(rg->bin_addr2);
      free(rg->bin_lats);
      free(rg->bin_lons);
    }

} /* remapGridFree */


void remapVarsFree(REMAPVARS *rv)
{
  static char func[] = "remapVarsFree";
  int i;

  if ( rv->pinit == TRUE )
    {
      rv->pinit = FALSE;

      free(rv->grid1_add);
      free(rv->grid2_add);
      for ( i = 0; i < 4; i++ )
	free(rv->wts[i]);
    }

} /* remapVarsFree */


void genXbounds(int xsize, int ysize, double *grid_center_lon, double *grid_corner_lon)
{
  int i, j, index;
  double dlon, minlon, maxlon;

  dlon = 360./xsize;

  if ( xsize == 1 || (grid_center_lon[xsize-1]-grid_center_lon[0]+dlon) < 359 )
    cdoAbort("can't calculate Xbounds for %d vals with dlon = %g", xsize, dlon);

  for ( i = 0; i < xsize; i++ )
    {
      minlon = grid_center_lon[i] - half*dlon;
      maxlon = grid_center_lon[i] + half*dlon;
      for ( j = 0; j < ysize; j++ )
	{
	  index = j*4*xsize + 4*i;
	  grid_corner_lon[index+0] = minlon;
	  grid_corner_lon[index+1] = maxlon;
	  grid_corner_lon[index+2] = maxlon;
	  grid_corner_lon[index+3] = minlon;
	}
    }
}


void genYbounds(int xsize, int ysize, double *grid_center_lat, double *grid_corner_lat)
{
  int i, j, index;
  double minlat, maxlat;
  double firstlat, lastlat;

  firstlat = grid_center_lat[0];
  lastlat  = grid_center_lat[xsize*ysize-1];

  if ( ysize == 1 ) cdoAbort("can't calculate Ybounds for 1 value");

  for ( j = 0; j < ysize; j++ )
    {
      index = j*xsize;
      if ( firstlat > lastlat )
	{
	  if ( j == 0 )
	    maxlat = 90;
	  else
	    maxlat = 0.5*(grid_center_lat[index]+grid_center_lat[index-xsize]);

	  if ( j == (ysize-1) )
	    minlat = -90;
	  else
	    minlat = 0.5*(grid_center_lat[index]+grid_center_lat[index+xsize]);
	}
      else
	{
	  if ( j == 0 )
	    minlat = -90;
	  else
	    minlat = 0.5*(grid_center_lat[index]+grid_center_lat[index-xsize]);

	  if ( j == (ysize-1) )
	    maxlat = 90;
	  else
	    maxlat = 0.5*(grid_center_lat[index]+grid_center_lat[index+xsize]);
	}

      for ( i = 0; i < xsize; i++ )
	{
	  index = j*4*xsize + 4*i;
	  grid_corner_lat[index+0] = minlat;
	  grid_corner_lat[index+1] = minlat;
	  grid_corner_lat[index+2] = maxlat;
	  grid_corner_lat[index+3] = maxlat;
	}
    }
}


void remapGridInit(int map_type, int gridID1, int gridID2, REMAPGRID *rg)
{
  static char func[] = "remapGridInit";
  char units[128];
  int nbins, nalloc;
  int k, nc;
  int n;      /* loop counter                  */
  int nele;   /* element loop counter          */
  int i,j;    /* logical 2d addresses          */
  int ip1,jp1;
  int n_add, e_add, ne_add;
  int nx, ny;
  int lgrid1_gen_bounds = FALSE, lgrid2_gen_bounds = FALSE;
  double  dlat, dlon;                /* lat/lon intervals for search bins  */
  double  tmp_lats[4], tmp_lons[4];  /* temps for computing bounding boxes */

  /* initialize all pointer */
  if ( rg->pinit == FALSE )
    {
      rg->pinit = TRUE;

      rg->grid1_mask       = NULL;
      rg->grid2_mask       = NULL;
      rg->grid1_center_lat = NULL;
      rg->grid1_center_lon = NULL;
      rg->grid2_center_lat = NULL;
      rg->grid2_center_lon = NULL;
      rg->grid1_area       = NULL;
      rg->grid2_area       = NULL;
      rg->grid1_frac       = NULL;
      rg->grid2_frac       = NULL;

      rg->grid1_corner_lat = NULL;
      rg->grid1_corner_lon = NULL;
      rg->grid2_corner_lat = NULL;
      rg->grid2_corner_lon = NULL;

      rg->grid1_bound_box  = NULL;
      rg->grid2_bound_box  = NULL;

      rg->bin_addr1 = NULL;
      rg->bin_addr2 = NULL;
      rg->bin_lats  = NULL;
      rg->bin_lons  = NULL;
    }

  if ( gridInqType(gridID1) != GRID_CELL && gridInqType(gridID1) != GRID_CURVILINEAR )
    {
      if ( gridInqType(gridID1) == GRID_TRIANGULAR )
	{
	  gridID1 = gridToCell(gridID1);
	}
      else
	{
	  gridID1 = gridToCurvilinear(gridID1);
	  lgrid1_gen_bounds = TRUE;
	}
    }

  if ( gridInqType(gridID2) != GRID_CELL && gridInqType(gridID2) != GRID_CURVILINEAR )
    {
      if ( gridInqType(gridID2) == GRID_TRIANGULAR )
	{
	  gridID2 = gridToCell(gridID2);
	}
      else
	{
	  gridID2 = gridToCurvilinear(gridID2);
	  lgrid2_gen_bounds = TRUE;
	}
    }

  rg->grid1_size = gridInqSize(gridID1);
  rg->grid2_size = gridInqSize(gridID2);

  if ( gridInqType(gridID1) == GRID_CELL )
    rg->grid1_rank = 1;
  else
    rg->grid1_rank = 2;

  if ( gridInqType(gridID2) == GRID_CELL )
    rg->grid2_rank = 1;
  else
    rg->grid2_rank = 2;

  if ( gridInqType(gridID1) == GRID_CELL )
    rg->grid1_corners = gridInqNvertex(gridID1);
  else
    rg->grid1_corners = 4;

  if ( gridInqType(gridID2) == GRID_CELL )
    rg->grid2_corners = gridInqNvertex(gridID2);
  else
    rg->grid2_corners = 4;

  rg->grid1_mask       = (int *) realloc(rg->grid1_mask, rg->grid1_size*sizeof(int));
  rg->grid2_mask       = (int *) realloc(rg->grid2_mask, rg->grid2_size*sizeof(int));
  rg->grid1_center_lat = (double *) realloc(rg->grid1_center_lat, rg->grid1_size*sizeof(double));
  rg->grid1_center_lon = (double *) realloc(rg->grid1_center_lon, rg->grid1_size*sizeof(double));
  rg->grid2_center_lat = (double *) realloc(rg->grid2_center_lat, rg->grid2_size*sizeof(double));
  rg->grid2_center_lon = (double *) realloc(rg->grid2_center_lon, rg->grid2_size*sizeof(double));

  if ( map_type == map_type_conserv )
    {
      rg->grid1_area = (double *) realloc(rg->grid1_area, rg->grid1_size*sizeof(double));
      rg->grid2_area = (double *) realloc(rg->grid2_area, rg->grid2_size*sizeof(double));

      memset(rg->grid1_area, 0, rg->grid1_size*sizeof(double));
      memset(rg->grid2_area, 0, rg->grid2_size*sizeof(double));
    }

  rg->grid1_frac = (double *) realloc(rg->grid1_frac, rg->grid1_size*sizeof(double));
  rg->grid2_frac = (double *) realloc(rg->grid2_frac, rg->grid2_size*sizeof(double));

  memset(rg->grid1_frac, 0, rg->grid1_size*sizeof(double));
  memset(rg->grid2_frac, 0, rg->grid2_size*sizeof(double));

  if ( ! rg->luse_grid_centers )
    {
      nalloc = rg->grid1_corners*rg->grid1_size;
      rg->grid1_corner_lat = (double *) realloc(rg->grid1_corner_lat, nalloc*sizeof(double));
      rg->grid1_corner_lon = (double *) realloc(rg->grid1_corner_lon, nalloc*sizeof(double));
      nalloc = rg->grid2_corners*rg->grid2_size;
      rg->grid2_corner_lat = (double *) realloc(rg->grid2_corner_lat, nalloc*sizeof(double));
      rg->grid2_corner_lon = (double *) realloc(rg->grid2_corner_lon, nalloc*sizeof(double));
    }
  
  rg->grid1_bound_box  = (double *) realloc(rg->grid1_bound_box, 4*rg->grid1_size*sizeof(double));
  rg->grid2_bound_box  = (double *) realloc(rg->grid2_bound_box, 4*rg->grid2_size*sizeof(double));

  rg->grid1_dims[0] = gridInqXsize(gridID1);
  rg->grid1_dims[1] = gridInqYsize(gridID1);
 
  gridInqYvals(gridID1, rg->grid1_center_lat);
  gridInqXvals(gridID1, rg->grid1_center_lon);

  if ( ! rg->luse_grid_centers )
    {
      if ( gridInqYbounds(gridID1, NULL) && gridInqXbounds(gridID1, NULL) )
	{
	  gridInqYbounds(gridID1, rg->grid1_corner_lat);
	  gridInqXbounds(gridID1, rg->grid1_corner_lon);
	}
      else
	{
	  if ( lgrid1_gen_bounds )
	    {
	      genXbounds(rg->grid1_dims[0], rg->grid1_dims[1], rg->grid1_center_lon, rg->grid1_corner_lon);
	      genYbounds(rg->grid1_dims[0], rg->grid1_dims[1], rg->grid1_center_lat, rg->grid1_corner_lat);
	    }
	  else
	    cdoAbort("grid1 corner missing!");
	}
    }

  /* initialize logical mask and convert lat/lon units if required */

  for ( i = 0; i < rg->grid1_size; i++ ) rg->grid1_mask[i] = TRUE;

  gridInqYunits(gridID1, units);

  if ( strncmp(units, "degrees", 7) == 0 )
    {
      for ( i = 0; i < rg->grid1_size; i++ )
	{
	  rg->grid1_center_lat[i] = rg->grid1_center_lat[i]*deg2rad;
	  rg->grid1_center_lon[i] = rg->grid1_center_lon[i]*deg2rad;
	}

      /* note: using units from latitude instead from bounds */
      if ( ! rg->luse_grid_centers )
	for ( i = 0; i < rg->grid1_corners*rg->grid1_size; i++ )
	  {
	    rg->grid1_corner_lat[i] = rg->grid1_corner_lat[i]*deg2rad;
	    rg->grid1_corner_lon[i] = rg->grid1_corner_lon[i]*deg2rad;
	  }
    }
  else if ( strncmp(units, "radians", 7) == 0 )
    {
      /* no conversion necessary */
    }
  else
    {
      cdoWarning("unknown units supplied for grid1 center lat/lon: "
		 "proceeding assuming radians");
    }

  /* data for grid 2 */

  rg->grid2_dims[0] = gridInqXsize(gridID2);
  rg->grid2_dims[1] = gridInqYsize(gridID2);

  gridInqYvals(gridID2, rg->grid2_center_lat);
  gridInqXvals(gridID2, rg->grid2_center_lon);

  if ( ! rg->luse_grid_centers )
    {
      if ( gridInqYbounds(gridID2, NULL) && gridInqXbounds(gridID2, NULL) )
	{
	  gridInqYbounds(gridID2, rg->grid2_corner_lat);
	  gridInqXbounds(gridID2, rg->grid2_corner_lon);
	}
      else
	{
	  if ( lgrid2_gen_bounds )
	    {
	      genXbounds(rg->grid2_dims[0], rg->grid2_dims[1], rg->grid2_center_lon, rg->grid2_corner_lon);
	      genYbounds(rg->grid2_dims[0], rg->grid2_dims[1], rg->grid2_center_lat, rg->grid2_corner_lat);
	    }
	  else
	    cdoAbort("grid2 corner missing!");
	}
    }

  /* initialize logical mask and convert lat/lon units if required */

  for ( i = 0; i < rg->grid2_size; i++ ) rg->grid2_mask[i] = TRUE;

  gridInqYunits(gridID2, units);

  if ( strncmp(units, "degrees", 7) == 0 )
    {
      for ( i = 0; i < rg->grid2_size; i++ )
	{
	  rg->grid2_center_lat[i] *= deg2rad;
	  rg->grid2_center_lon[i] *= deg2rad;
	}

      /* note: using units from latitude instead from bounds */
      if ( ! rg->luse_grid_centers )
	for ( i = 0; i < rg->grid2_corners*rg->grid2_size; i++ )
	  {
	    rg->grid2_corner_lat[i] *= deg2rad;
	    rg->grid2_corner_lon[i] *= deg2rad;
	  }
    }
  else if ( strncmp(units, "radians", 7) == 0 )
    {
      /* no conversion necessary */
    }
  else
    {
      cdoWarning("unknown units supplied for grid2 center lat/lon: "
		 "proceeding assuming radians");
    }

  /*     convert longitudes to 0,2pi interval */

  for ( i = 0; i < rg->grid1_size; i++ )
    {
      if ( rg->grid1_center_lon[i] > pi2 )  rg->grid1_center_lon[i] -= pi2;
      if ( rg->grid1_center_lon[i] < zero ) rg->grid1_center_lon[i] += pi2;
    }
  for ( i = 0; i < rg->grid2_size; i++ )
    {
      if ( rg->grid2_center_lon[i] > pi2 )  rg->grid2_center_lon[i] -= pi2;
      if ( rg->grid2_center_lon[i] < zero ) rg->grid2_center_lon[i] += pi2;
    }

  if ( ! rg->luse_grid_centers )
    for ( i = 0; i < rg->grid1_corners*rg->grid1_size; i++ )
      {
	if ( rg->grid1_corner_lon[i] > pi2 )  rg->grid1_corner_lon[i] -= pi2;
	if ( rg->grid1_corner_lon[i] < zero ) rg->grid1_corner_lon[i] += pi2;
      }

  if ( ! rg->luse_grid_centers )
    for ( i = 0; i < rg->grid2_corners*rg->grid2_size; i++ )
      {
	if ( rg->grid2_corner_lon[i] > pi2 )  rg->grid2_corner_lon[i] -= pi2;
	if ( rg->grid2_corner_lon[i] < zero ) rg->grid2_corner_lon[i] += pi2;
      }

  /*  make sure input latitude range is within the machine values for +/- pi/2 */

  for ( i = 0; i < rg->grid1_size; i++ )
    {
      if ( rg->grid1_center_lat[i] >  pih ) rg->grid1_center_lat[i] =  pih;
      if ( rg->grid1_center_lat[i] < -pih ) rg->grid1_center_lat[i] = -pih;
    }

  if ( ! rg->luse_grid_centers )
    for ( i = 0; i < rg->grid1_corners*rg->grid1_size; i++ )
      {
	if ( rg->grid1_corner_lat[i] >  pih ) rg->grid1_corner_lat[i] =  pih;
	if ( rg->grid1_corner_lat[i] < -pih ) rg->grid1_corner_lat[i] = -pih;
      }

  for ( i = 0; i < rg->grid2_size; i++ )
    {
      if ( rg->grid2_center_lat[i] >  pih ) rg->grid2_center_lat[i] =  pih;
      if ( rg->grid2_center_lat[i] < -pih ) rg->grid2_center_lat[i] = -pih;
    }

   if ( ! rg->luse_grid_centers )
     for ( i = 0; i < rg->grid2_corners*rg->grid2_size; i++ )
       {
	 if ( rg->grid2_corner_lat[i] >  pih ) rg->grid2_corner_lat[i] =  pih;
	 if ( rg->grid2_corner_lat[i] < -pih ) rg->grid2_corner_lat[i] = -pih;
       }

  /*  compute bounding boxes for restricting future grid searches */

  if ( ! rg->luse_grid_centers )
    {
      nc = rg->grid1_corners;
      for ( i = 0; i < rg->grid1_size; i++ )
	{
	  rg->grid1_bound_box[4*i+0] = rg->grid1_corner_lat[i*nc];
	  rg->grid1_bound_box[4*i+1] = rg->grid1_corner_lat[i*nc];
	  rg->grid1_bound_box[4*i+2] = rg->grid1_corner_lon[i*nc];
	  rg->grid1_bound_box[4*i+3] = rg->grid1_corner_lon[i*nc];
	  for ( j = 1; j < nc; j++ )
	    {
	      if ( rg->grid1_corner_lat[i*nc+j] < rg->grid1_bound_box[4*i+0] )
		rg->grid1_bound_box[4*i+0] = rg->grid1_corner_lat[i*nc+j];
	      if ( rg->grid1_corner_lat[i*nc+j] > rg->grid1_bound_box[4*i+1] )
		rg->grid1_bound_box[4*i+1] = rg->grid1_corner_lat[i*nc+j];
	      if ( rg->grid1_corner_lon[i*nc+j] < rg->grid1_bound_box[4*i+2] )
		rg->grid1_bound_box[4*i+2] = rg->grid1_corner_lon[i*nc+j];
	      if ( rg->grid1_corner_lon[i*nc+j] > rg->grid1_bound_box[4*i+3] )
		rg->grid1_bound_box[4*i+3] = rg->grid1_corner_lon[i*nc+j];
	    }
	}

      nc = rg->grid2_corners;
      for ( i = 0; i < rg->grid2_size; i++ )
	{
	  rg->grid2_bound_box[4*i+0] = rg->grid2_corner_lat[i*nc];
	  rg->grid2_bound_box[4*i+1] = rg->grid2_corner_lat[i*nc];
	  rg->grid2_bound_box[4*i+2] = rg->grid2_corner_lon[i*nc];
	  rg->grid2_bound_box[4*i+3] = rg->grid2_corner_lon[i*nc];
	  for ( j = 1; j < nc; j++ )
	    {
	      if ( rg->grid2_corner_lat[i*nc+j] < rg->grid2_bound_box[4*i+0] )
		rg->grid2_bound_box[4*i+0] = rg->grid2_corner_lat[i*nc+j];
	      if ( rg->grid2_corner_lat[i*nc+j] > rg->grid2_bound_box[4*i+1] )
		rg->grid2_bound_box[4*i+1] = rg->grid2_corner_lat[i*nc+j];
	      if ( rg->grid2_corner_lon[i*nc+j] < rg->grid2_bound_box[4*i+2] )
		rg->grid2_bound_box[4*i+2] = rg->grid2_corner_lon[i*nc+j];
	      if ( rg->grid2_corner_lon[i*nc+j] > rg->grid2_bound_box[4*i+3] )
		rg->grid2_bound_box[4*i+3] = rg->grid2_corner_lon[i*nc+j];
	    }
	}
    }
  else
    {
      nx = rg->grid1_dims[0];
      ny = rg->grid1_dims[1];

      for ( n = 0; n < rg->grid1_size; n++ )
	{
          /* find N,S and NE points to this grid point */

          j = n/nx + 1;
          i = n - (j-1)*nx + 1;

          if ( i < nx )
            ip1 = i + 1;
          else
	    {
	      /* assume cyclic */
	      ip1 = 1;
	      /* but if it is not, correct */
	      e_add = (j - 1)*nx + ip1 - 1;
	      if ( fabs(rg->grid1_center_lat[e_add] - rg->grid1_center_lat[n]) > pih ) ip1 = i;
	      /*
	      ip1 = i;
	      */
	    }

          if ( j < ny )
            jp1 = j + 1;
          else
	    {
	      /* assume cyclic */
	      jp1 = 1;
	      /* but if it is not, correct */
	      n_add = (jp1 - 1)*nx + i - 1;
	      if ( fabs(rg->grid1_center_lat[n_add] - rg->grid1_center_lat[n]) > pih ) jp1 = j;
	      /*
	      jp1 = j;
	      */
	    }

          n_add  = (jp1 - 1)*nx + i - 1;
          e_add  = (j - 1)*nx + ip1 - 1;
          ne_add = (jp1 - 1)*nx + ip1 - 1;

          /* find N,S and NE lat/lon coords and check bounding box */

          tmp_lats[0] = rg->grid1_center_lat[n];
          tmp_lats[1] = rg->grid1_center_lat[e_add];
          tmp_lats[2] = rg->grid1_center_lat[ne_add];
          tmp_lats[3] = rg->grid1_center_lat[n_add];

          tmp_lons[0] = rg->grid1_center_lon[n];
          tmp_lons[1] = rg->grid1_center_lon[e_add];
          tmp_lons[2] = rg->grid1_center_lon[ne_add];
          tmp_lons[3] = rg->grid1_center_lon[n_add];

	  /* test */
	  /*
	  if ( ip1 == 1 )
	    {
	      if ( tmp_lons[1] <= 0 ) tmp_lons[1] += pi2;
	      if ( tmp_lons[2] <= 0 ) tmp_lons[2] += pi2;
	    }
	  */
	  
          rg->grid1_bound_box[4*n+0] = tmp_lats[0];
          rg->grid1_bound_box[4*n+1] = tmp_lats[0];
          rg->grid1_bound_box[4*n+2] = tmp_lons[0];
          rg->grid1_bound_box[4*n+3] = tmp_lons[0];
	  for ( k = 1; k < 4; k++ )
	    {
	      if ( tmp_lats[k] < rg->grid1_bound_box[4*n+0] ) rg->grid1_bound_box[4*n+0] = tmp_lats[k];
	      if ( tmp_lats[k] > rg->grid1_bound_box[4*n+1] ) rg->grid1_bound_box[4*n+1] = tmp_lats[k];
	      if ( tmp_lons[k] < rg->grid1_bound_box[4*n+2] ) rg->grid1_bound_box[4*n+2] = tmp_lons[k];
	      if ( tmp_lons[k] > rg->grid1_bound_box[4*n+3] ) rg->grid1_bound_box[4*n+3] = tmp_lons[k];
	    }
        }

      nx = rg->grid2_dims[0];
      ny = rg->grid2_dims[1];

      for ( n = 0; n < rg->grid2_size; n++ )
	{
	  /* find N,S and NE points to this grid point */

          j = n/nx + 1;
          i = n - (j-1)*nx + 1;

          if ( i < nx )
            ip1 = i + 1;
          else
	    {
	      /* assume cyclic */
	      ip1 = 1;
	      /* but if it is not, correct */
	      e_add = (j - 1)*nx + ip1 - 1;
	      if ( fabs(rg->grid2_center_lat[e_add] - rg->grid2_center_lat[n]) > pih ) ip1 = i;
	    }

          if ( j < ny )
            jp1 = j+1;
          else
	    {
	      /* assume cyclic */
	      jp1 = 1;
	      /* but if it is not, correct */
	      n_add = (jp1 - 1)*nx + i - 1;
	      if ( fabs(rg->grid2_center_lat[n_add] - rg->grid2_center_lat[n]) > pih ) jp1 = j;
	    }

          n_add = (jp1 - 1)*nx + i - 1;
          e_add = (j - 1)*nx + ip1 - 1;
          ne_add = (jp1 - 1)*nx + ip1 - 1;

          /* find N,S and NE lat/lon coords and check bounding box */

          tmp_lats[0] = rg->grid2_center_lat[n];
          tmp_lats[1] = rg->grid2_center_lat[e_add];
          tmp_lats[2] = rg->grid2_center_lat[ne_add];
          tmp_lats[3] = rg->grid2_center_lat[n_add];

          tmp_lons[0] = rg->grid2_center_lon[n];
          tmp_lons[1] = rg->grid2_center_lon[e_add];
          tmp_lons[2] = rg->grid2_center_lon[ne_add];
          tmp_lons[3] = rg->grid2_center_lon[n_add];

	  /* test */
	  /*
	  if ( ip1 == 1 )
	    {
	      if ( tmp_lons[1] <= 0 ) tmp_lons[1] += pi2;
	      if ( tmp_lons[2] <= 0 ) tmp_lons[2] += pi2;
	    }
	  */

          rg->grid2_bound_box[4*n+0] = tmp_lats[0];
          rg->grid2_bound_box[4*n+1] = tmp_lats[0];
          rg->grid2_bound_box[4*n+2] = tmp_lons[0];
          rg->grid2_bound_box[4*n+3] = tmp_lons[0];
	  for ( k = 1; k < 4; k++ )
	    {
	      if ( tmp_lats[k] < rg->grid2_bound_box[4*n+0] ) rg->grid2_bound_box[4*n+0] = tmp_lats[k];
	      if ( tmp_lats[k] > rg->grid2_bound_box[4*n+1] ) rg->grid2_bound_box[4*n+1] = tmp_lats[k];
	      if ( tmp_lons[k] < rg->grid2_bound_box[4*n+2] ) rg->grid2_bound_box[4*n+2] = tmp_lons[k];
	      if ( tmp_lons[k] > rg->grid2_bound_box[4*n+3] ) rg->grid2_bound_box[4*n+3] = tmp_lons[k];
	    }
        }
    }

  for ( n = 0; n < rg->grid1_size; n++ )
    if ( fabs(rg->grid1_bound_box[4*n+3] - rg->grid1_bound_box[4*n+2]) > pi )
      {
	rg->grid1_bound_box[4*n+2] = zero;
	rg->grid1_bound_box[4*n+3] = pi2;
      }

  for ( n = 0; n < rg->grid2_size; n++ )
    if ( fabs(rg->grid2_bound_box[4*n+3] - rg->grid2_bound_box[4*n+2]) > pi )
      {
        rg->grid2_bound_box[4*n+2] = zero;
        rg->grid2_bound_box[4*n+3] = pi2;
      }

  /* try to check for cells that overlap poles */

  for ( n = 0; n < rg->grid1_size; n++ )
    if ( rg->grid1_center_lat[n] > rg->grid1_bound_box[4*n+1] )
      rg->grid1_bound_box[4*n+1] = pih;

  for ( n = 0; n < rg->grid1_size; n++ )
    if ( rg->grid1_center_lat[n] < rg->grid1_bound_box[4*n+0] )
      rg->grid1_bound_box[4*n+0] = -pih;

  for ( n = 0; n < rg->grid2_size; n++ )
    if ( rg->grid2_center_lat[n] > rg->grid2_bound_box[4*n+1] )
      rg->grid2_bound_box[4*n+1] = pih;

  for ( n = 0; n < rg->grid2_size; n++ )
    if ( rg->grid2_center_lat[n] < rg->grid2_bound_box[4*n+0] )
      rg->grid2_bound_box[4*n+0] = -pih;


  /* ##################### */
  /*
  nx = rg->grid1_dims[0];
  ny = rg->grid1_dims[1];
  for ( n = 0; n < rg->grid1_size; n++ )
    {
      j = n/nx + 1;
      i = n - (j-1)*nx + 1;
      printf("1 %d %d %d %g %g %g %g\n", n, j, i,
	     rad2deg*rg->grid1_bound_box[4*n+0], rad2deg*rg->grid1_bound_box[4*n+1],
	     rad2deg*rg->grid1_bound_box[4*n+2], rad2deg*rg->grid1_bound_box[4*n+3]);
    }

  nx = rg->grid2_dims[0];
  ny = rg->grid2_dims[1];
  for ( n = 0; n < rg->grid2_size; n++ )
    {
      j = n/nx + 1;
      i = n - (j-1)*nx + 1;
      printf("2 %d %d %d %g %g %g %g\n", n, j, i,
	     rad2deg*rg->grid2_bound_box[4*n+0], rad2deg*rg->grid2_bound_box[4*n+1],
	     rad2deg*rg->grid2_bound_box[4*n+2], rad2deg*rg->grid2_bound_box[4*n+3]);
    }
  */

  /*
    set up and assign address ranges to search bins in order to 
    further restrict later searches
  */
  if ( rg->restrict_type == restrict_latitude )
    {
      nbins = rg->num_srch_bins;

      if ( cdoVerbose )
	cdoPrint("Using %d latitude bins to restrict search.", nbins);

      rg->bin_addr1 = (int *) realloc(rg->bin_addr1, 2*nbins*sizeof(int));
      rg->bin_addr2 = (int *) realloc(rg->bin_addr2, 2*nbins*sizeof(int));
      rg->bin_lats  = (double *) realloc(rg->bin_lats, 2*nbins*sizeof(double));
      rg->bin_lons  = (double *) realloc(rg->bin_lons, 2*nbins*sizeof(double));

      dlat = pi/rg->num_srch_bins;

      for ( n = 0; n < rg->num_srch_bins; n++ )
	{
	  rg->bin_lats[2*n+0] = (n  )*dlat - pih;
	  rg->bin_lats[2*n+1] = (n+1)*dlat - pih;
	  rg->bin_lons[2*n+0] = zero;
	  rg->bin_lons[2*n+1] = pi2;
	  rg->bin_addr1[2*n+0] = rg->grid1_size;
	  rg->bin_addr1[2*n+1] = 0;
	  rg->bin_addr2[2*n+0] = rg->grid2_size;
	  rg->bin_addr2[2*n+1] = 0;
	}

      for ( nele = 0; nele < rg->grid1_size; nele++ )
	for ( n = 0; n < rg->num_srch_bins; n++ )
	  if ( rg->grid1_bound_box[4*nele+0] <= rg->bin_lats[2*n+1] &&
	       rg->grid1_bound_box[4*nele+1] >= rg->bin_lats[2*n+0] )
	    {
	      rg->bin_addr1[2*n+0] = MIN(nele, rg->bin_addr1[2*n+0]);
	      rg->bin_addr1[2*n+1] = MAX(nele, rg->bin_addr1[2*n+1]);
	    }

      for ( nele = 0; nele < rg->grid2_size; nele++ )
	for ( n = 0; n < rg->num_srch_bins; n++ )
	  if ( rg->grid2_bound_box[4*nele+0] <= rg->bin_lats[2*n+1] &&
	       rg->grid2_bound_box[4*nele+1] >= rg->bin_lats[2*n+0] )
	    {
	      rg->bin_addr2[2*n+0] = MIN(nele, rg->bin_addr2[2*n+0]);
	      rg->bin_addr2[2*n+1] = MAX(nele, rg->bin_addr2[2*n+1]);
	    }
    }
  else if ( rg->restrict_type == restrict_latlon )
    {
      dlat = pi /rg->num_srch_bins;
      dlon = pi2/rg->num_srch_bins;

      nbins = rg->num_srch_bins*rg->num_srch_bins;

      if ( cdoVerbose )
	cdoPrint("Using %d lat/lon boxes to restrict search.", nbins);

      rg->bin_addr1 = (int *) realloc(rg->bin_addr1, 2*nbins*sizeof(int));
      rg->bin_addr2 = (int *) realloc(rg->bin_addr2, 2*nbins*sizeof(int));
      rg->bin_lats  = (double *) realloc(rg->bin_lats, 2*nbins*sizeof(double));
      rg->bin_lons  = (double *) realloc(rg->bin_lons, 2*nbins*sizeof(double));

      n = 0;
      for ( j = 0; j < rg->num_srch_bins; j++ )
	for ( i = 0; i < rg->num_srch_bins; i++ )
	  {
	    rg->bin_lats[2*n+0]  = (j  )*dlat - pih;
	    rg->bin_lats[2*n+1]  = (j+1)*dlat - pih;
	    rg->bin_lons[2*n+0]  = (i  )*dlon;
	    rg->bin_lons[2*n+1]  = (i+1)*dlon;
	    rg->bin_addr1[2*n+0] = rg->grid1_size;
	    rg->bin_addr1[2*n+1] = 0;
	    rg->bin_addr2[2*n+0] = rg->grid2_size;
	    rg->bin_addr2[2*n+1] = 0;

	    n++;
	  }

      rg->num_srch_bins = rg->num_srch_bins*rg->num_srch_bins;

      for ( nele = 0; nele < rg->grid1_size; nele++ )
	for ( n = 0; n < rg->num_srch_bins; n++ )
	  if ( rg->grid1_bound_box[4*nele+0] <= rg->bin_lats[2*n+1] &&
	       rg->grid1_bound_box[4*nele+1] >= rg->bin_lats[2*n+0] &&
	       rg->grid1_bound_box[4*nele+2] <= rg->bin_lons[2*n+1] &&
	       rg->grid1_bound_box[4*nele+3] >= rg->bin_lons[2*n+0] )
	    {
	      rg->bin_addr1[2*n+0] = MIN(nele,rg->bin_addr1[2*n+0]);
	      rg->bin_addr1[2*n+1] = MAX(nele,rg->bin_addr1[2*n+1]);
	    }

      for ( nele = 0; nele < rg->grid2_size; nele++ )
	for ( n = 0; n < rg->num_srch_bins; n++ )
	  if ( rg->grid2_bound_box[4*nele+0] <= rg->bin_lats[2*n+1] &&
	       rg->grid2_bound_box[4*nele+1] >= rg->bin_lats[2*n+0] &&
	       rg->grid2_bound_box[4*nele+2] <= rg->bin_lons[2*n+1] &&
	       rg->grid2_bound_box[4*nele+3] >= rg->bin_lons[2*n+0] )
	    {
	      rg->bin_addr2[2*n+0] = MIN(nele,rg->bin_addr2[2*n+0]);
	      rg->bin_addr2[2*n+1] = MAX(nele,rg->bin_addr2[2*n+1]);
	    }
    }
  else
    cdoAbort("unknown search restriction method");

}  /* remapGridInit */


/*
    this routine initializes some variables and provides an initial
    allocation of arrays (fairly large so frequent resizing unnecessary).
*/
void remapVarsInit(int map_type, REMAPGRID *rg, REMAPVARS *rv)
{
  static char func[] = "remapVarsInit";
  int i;

  /* initialize all pointer */
  if ( rv->pinit == FALSE )
    {
      rv->pinit = TRUE;

      rv->grid1_add = NULL;
      rv->grid2_add = NULL;
      for ( i = 0; i < 4; i++ )
	rv->wts[i] = NULL;
    }

  /* determine the number of weights */

  if      ( map_type == map_type_conserv  ) rv->num_wts = 3;
  else if ( map_type == map_type_bilinear ) rv->num_wts = 1;
  else if ( map_type == map_type_bicubic  ) rv->num_wts = 4;
  else if ( map_type == map_type_distwgt  ) rv->num_wts = 1;
  else if ( map_type == map_type_distwgt1 ) rv->num_wts = 1;
  else cdoAbort("unknown mapping method");

  /*
    initialize num_links and set max_links to four times the largest 
    of the destination grid sizes initially (can be changed later).
    set a default resize increment to increase the size of link
    arrays if the number of links exceeds the initial size
  */
  rv->num_links = 0;
  rv->max_links = 4 * rg->grid2_size;

  rv->resize_increment = (int) (0.1 * MAX(rg->grid1_size, rg->grid2_size));

  /*  allocate address and weight arrays for mapping 1 */

  rv->grid1_add = (int *) realloc(rv->grid1_add, rv->max_links*sizeof(int));
  rv->grid2_add = (int *) realloc(rv->grid2_add, rv->max_links*sizeof(int));

  for ( i = 0; i < rv->num_wts; i++ )
    rv->wts[i] = (double *) realloc(rv->wts[i], rv->max_links*sizeof(double));

} /* remapVarsInit */


/*
   this routine resizes remapping arrays by increasing(decreasing)
   the max_links by increment
*/
void resize_remap_vars(REMAPVARS *rv, int increment)
{
  static char func[] = "resize_remap_vars";
  /*
    input variables:
    int  increment  ! the number of links to add(subtract) to arrays
  */
  int i;

  /*  reallocate arrays at new size */

  rv->max_links += increment;

  if ( rv->max_links )
    {
      rv->grid1_add = (int *) realloc(rv->grid1_add, rv->max_links*sizeof(int));
      rv->grid2_add = (int *) realloc(rv->grid2_add, rv->max_links*sizeof(int));

      for ( i = 0; i < rv->num_wts; i++ )
	rv->wts[i] = (double *) realloc(rv->wts[i], rv->max_links*sizeof(double));
    }

} /* resize_remap_vars */


/*
  -----------------------------------------------------------------------

  performs the remapping based on weights computed elsewhere
     
  -----------------------------------------------------------------------
*/
void remap(double *dst_array, double missval, int dst_size, int num_links, double **map_wts, int num_wts,
	   int *dst_add, int *src_add, 
	   double *src_array, double *src_grad1, double *src_grad2, double *src_grad3)
{
  /*
    input arrays:

    int *dst_add         ! destination address for each link
    int *src_add         ! source      address for each link

    int num_wts          ! num of weights used in remapping

    double **map_wts     ! remapping weights for each link

    double *src_array    ! array with source field to be remapped

    optional:

    double *src_grad1    ! gradient arrays on source grid necessary for
    double *src_grad2    ! higher-order remappings
    double *src_grad3

    output variables:

    double *dst_array    ! array for remapped field on destination grid
  */

  /* local variables */

  int n, iorder;

  /* check the order of the interpolation */

  if ( src_grad1 )
    iorder = 2;
  else
    iorder = 1;

  for ( n = 0; n < dst_size; n++ ) dst_array[n] = missval;

  /* Unvectorized loop: Unvectorizable dependency */
  for ( n = 0; n < num_links; n++ )
    if ( dst_array[dst_add[n]] == missval ) dst_array[dst_add[n]] = zero;

  if ( iorder == 1 )   /* first order remapping */
    {
      /* Unvectorized loop: (0.1-44% CPU time) Unvectorizable dependency */
      for ( n = 0; n < num_links; n++ )
	{
          dst_array[dst_add[n]] += src_array[src_add[n]]*map_wts[0][n];
	}
    }
  else                 /* second order remapping */
    {
      if ( num_wts == 3 )
	{
	  /* Unvectorized loop: Unvectorizable dependency */
	  for ( n = 0; n < num_links; n++ )
	    {
	      dst_array[dst_add[n]] += src_array[src_add[n]]*map_wts[0][n] +
                                       src_grad1[src_add[n]]*map_wts[1][n] +
                                       src_grad2[src_add[n]]*map_wts[2][n];
	    }
	}
      else if ( num_wts == 4 )
	{
	  /* Unvectorized loop: Unvectorizable dependency */
      	  for ( n = 0; n < num_links; n++ )
	    {
              dst_array[dst_add[n]] += src_array[src_add[n]]*map_wts[0][n] +
                                       src_grad1[src_add[n]]*map_wts[1][n] +
                                       src_grad2[src_add[n]]*map_wts[2][n] +
                                       src_grad3[src_add[n]]*map_wts[3][n];
	    }
	}
    }
}


/*
  -----------------------------------------------------------------------

  performs the remapping based on weights computed elsewhere
     
  -----------------------------------------------------------------------
*/

void remap_con1(double *dst_array, double missval, int dst_size, int num_links, double **map_wts,
		int num_wts, int *dst_add, int *src_add, double *src_array)
{
  /*
    input arrays:

    int *dst_add         ! destination address for each link
    int *src_add         ! source      address for each link

    int num_wts          ! num of weights used in remapping

    double **map_wts     ! remapping weights for each link

    double *src_array    ! array with source field to be remapped

    output variables:

    double *dst_array    ! array for remapped field on destination grid
  */

  /* local variables */
  static char func[] = "remap_con1";
  int i, n, k, ncls, imax;
  int max_cls = 1000;
  int max_cls_inc = 1000;
  double wts;
  double *src_cls;
  double *src_wts;

  src_cls = (double *) malloc(max_cls*sizeof(double));
  src_wts = (double *) malloc(max_cls*sizeof(double));

  memset(src_cls, 0, max_cls*sizeof(double));
  memset(src_wts, 0, max_cls*sizeof(double));

  for ( i = 0; i < dst_size; i++ ) dst_array[i] = missval;

  for ( n = 0; n < num_links; n++ )
    if ( dst_array[dst_add[n]] == missval ) dst_array[dst_add[n]] = zero;

  for ( i = 0; i < dst_size; i++ )
    {
      ncls = 0;
      for ( n = 0; n < num_links; n++ )
	{
	  if ( i == dst_add[n] )
	    {
	      for ( k = 0; k < ncls; k++ )
		if ( src_array[src_add[n]] == src_cls[k] ) break;
	      
	      if ( k == ncls )
		{
		  if ( ncls == max_cls )
		    {
		      max_cls += max_cls_inc;
		      src_cls = (double *) realloc(src_cls, max_cls*sizeof(double));
		      src_wts = (double *) realloc(src_wts, max_cls*sizeof(double));

		      memset(src_cls+ncls, 0, (max_cls-ncls)*sizeof(double));
		      memset(src_wts+ncls, 0, (max_cls-ncls)*sizeof(double));
		    }

		  src_cls[k] = src_array[src_add[n]];
		  ncls++;
		}
	      
	      src_wts[k] += map_wts[0][n];
	    }
	}

      if ( ncls )
	{
	  imax = 0;
	  wts = src_wts[0];
	  for ( k = 1; k < ncls; k++ )
	    {
	      if ( src_wts[k] > wts )
		{
		  wts  = src_wts[k];
		  imax = k;
		}
	    }

	  dst_array[i] = src_cls[imax];

	  memset(src_cls, 0, ncls*sizeof(double));
	  memset(src_wts, 0, ncls*sizeof(double));
	}
    }

  free(src_cls);
  free(src_wts);
}


static int     max_iter = 100;     /* max iteration count for i,j iteration */
static double  converge = 1.e-10;  /* convergence criterion */

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                         */
/*      BILINEAR INTERPOLATION                                             */
/*                                                                         */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/*
   this routine finds the location of the search point plat, plon
   in the source grid and returns the corners needed for a bilinear
   interpolation.
*/
void grid_search(REMAPGRID *rg, int *src_add, double *src_lats, double *src_lons, 
		 double plat, double plon, int *src_grid_dims,
		 double *src_center_lat, double *src_center_lon,
		 double *src_grid_bound_box,
		 int *src_bin_add, int *dst_bin_add)
{
  /*
    output variables:

    int    src_add[4]              ! address of each corner point enclosing P
    double src_lats[4]             ! latitudes  of the four corner points
    double src_lons[4]             ! longitudes of the four corner points
  */

  /*
    input variables:

    double plat                    ! latitude  of the search point
    double plon                    ! longitude of the search point

    int src_grid_dims[2]           ! size of each src grid dimension

    double src_center_lat[]        ! latitude  of each src grid center 
    double src_center_lon[]        ! longitude of each src grid center

    double src_grid_bound_box[][4] ! bound box for source grid

    int src_bin_add[][2]           ! latitude bins for restricting
    int dst_bin_add[][2]           ! searches
  */
  /*  local variables */

  int n, next_n, srch_add;                   /* dummy indices */
  int nx, ny;                                /* dimensions of src grid */
  int min_add, max_add;                      /* addresses for restricting search */
  int i, j, jp1, ip1, n_add, e_add, ne_add;  /* addresses */

  /* vectors for cross-product check */
  double vec1_lat, vec1_lon;
  double vec2_lat, vec2_lon, cross_product, cross_product_last = 0;
  double coslat_dst, sinlat_dst, coslon_dst, sinlon_dst;
  double dist_min, distance; /* for computing dist-weighted avg */

  /*
    restrict search first using bins
  */
  for ( n = 0; n < 4; n++ ) src_add[n] = 0;

  min_add = rg->grid1_size-1;
  max_add = 0;

  for ( n = 0; n < rg->num_srch_bins; n++ )
    if ( plat >= rg->bin_lats[2*n+0] && plat <= rg->bin_lats[2*n+1] &&
	 plon >= rg->bin_lons[2*n+0] && plon <= rg->bin_lons[2*n+1] )
      {
	if ( src_bin_add[2*n+0] < min_add ) min_add = src_bin_add[2*n+0];
	if ( src_bin_add[2*n+1] > max_add ) max_add = src_bin_add[2*n+1];
      }
 
  /*
    now perform a more detailed search 
  */
  nx = src_grid_dims[0];
  ny = src_grid_dims[1];

  /* srch_loop */
  /* Unvectorized loop: break and return */
  for ( srch_add = min_add; srch_add <= max_add; srch_add++ )
    {
      /*
	first check bounding box
      */
      if ( plat >= src_grid_bound_box[4*srch_add+0] && 
           plat <= src_grid_bound_box[4*srch_add+1] &&
           plon >= src_grid_bound_box[4*srch_add+2] &&
	   plon <= src_grid_bound_box[4*srch_add+3] )
	{
	  /*
	    we are within bounding box so get really serious
	  */

          /* determine neighbor addresses */

          j = srch_add/nx + 1;
          i = srch_add - (j-1)*nx + 1;

          if ( i < nx )
            ip1 = i + 1;
          else
            ip1 = 1;

          if ( j < ny )
            jp1 = j + 1;
          else
            jp1 = 1;

          n_add = (jp1 - 1)*nx + i - 1;
          e_add = (j - 1)*nx + ip1 - 1;
          ne_add = (jp1 - 1)*nx + ip1 - 1;

          src_lats[0] = src_center_lat[srch_add];
          src_lats[1] = src_center_lat[e_add];
          src_lats[2] = src_center_lat[ne_add];
          src_lats[3] = src_center_lat[n_add];

          src_lons[0] = src_center_lon[srch_add];
          src_lons[1] = src_center_lon[e_add];
          src_lons[2] = src_center_lon[ne_add];
          src_lons[3] = src_center_lon[n_add];

	  /*
	     for consistency, we must make sure all lons are in same 2pi interval
	  */

          vec1_lon = src_lons[0] - plon;
          if      ( vec1_lon >  pi ) src_lons[0] -= pi2;
          else if ( vec1_lon < -pi ) src_lons[0] += pi2;

          for ( n = 1; n < 4; n++ )
	    {
	      vec1_lon = src_lons[n] - src_lons[0];
	      if      ( vec1_lon >  pi ) src_lons[n] -= pi2;
	      else if ( vec1_lon < -pi ) src_lons[n] += pi2;
	    }

          /* corner_loop */
          for ( n = 0; n < 4; n++ )
	    {
	      next_n = (n+1)%4;

	      /*
		here we take the cross product of the vector making 
		up each box side with the vector formed by the vertex
		and search point.  if all the cross products are 
		positive, the point is contained in the box.
	      */
	      vec1_lat = src_lats[next_n] - src_lats[n];
	      vec1_lon = src_lons[next_n] - src_lons[n];
	      vec2_lat = plat - src_lats[n];
	      vec2_lon = plon - src_lons[n];

	      /* check for 0,2pi crossings */

	      if      ( vec1_lon >  three*pih ) vec1_lon -= pi2;
	      else if ( vec1_lon < -three*pih ) vec1_lon += pi2;

	      if      ( vec2_lon >  three*pih ) vec2_lon -= pi2;
	      else if ( vec2_lon < -three*pih ) vec2_lon += pi2;

	      cross_product = vec1_lon*vec2_lat - vec2_lon*vec1_lat;

	      /*
		if cross product is less than zero, this cell doesn't work
	      */
	      if ( n == 0 ) cross_product_last = cross_product;
	      if ( cross_product*cross_product_last < zero ) break;

	      cross_product_last = cross_product;

	    } /* corner_loop */

	  /*
	    if cross products all same sign, we found the location
	  */
          if ( n >= 4 )
	    {
	      src_add[0] = srch_add + 1;
	      src_add[1] = e_add + 1;
	      src_add[2] = ne_add + 1;
	      src_add[3] = n_add + 1;

	      return;
	    }

	  /* otherwise move on to next cell */

        } /* bounding box check */
    } /* srch_loop */

  /*
       if no cell found, point is likely either in a box that
       straddles either pole or is outside the grid.  fall back
       to a distance-weighted average of the four closest
       points.  go ahead and compute weights here, but store
       in src_lats and return -add to prevent the parent
       routine from computing bilinear weights
  */
      
  /*
    !print *,'Could not find location for ',plat,plon
    !print *,'Using nearest-neighbor average for this point'
  */

  coslat_dst = cos(plat);
  sinlat_dst = sin(plat);
  coslon_dst = cos(plon);
  sinlon_dst = sin(plon);

  dist_min = bignum;
  for ( n = 0; n < 4; n++ ) src_lats[n] = bignum;
  for ( srch_add = min_add; srch_add <= max_add; srch_add++ )
    {
      distance = acos(coslat_dst*cos(src_center_lat[srch_add])*
		     (coslon_dst*cos(src_center_lon[srch_add]) +
                      sinlon_dst*sin(src_center_lon[srch_add]))+
		      sinlat_dst*sin(src_center_lat[srch_add]));

      if ( distance < dist_min )
	{
          for ( n = 0; n < 4; n++ )
	    {
	      if ( distance < src_lats[n] )
		{
		  for ( i = 3; i > n; i-- )
		    {
		      src_add [i] = src_add [i-1];
		      src_lats[i] = src_lats[i-1];
		    }
		  src_add [n] = -srch_add - 1;
		  src_lats[n] = distance;
		  dist_min = src_lats[3];
		  break;
		}
	    }
        }
    }

  for ( n = 0; n < 4; n++ ) src_lons[n] = one/(src_lats[n] + tiny);
  distance = 0.0;
  for ( n = 0; n < 4; n++ ) distance += src_lons[n];
  for ( n = 0; n < 4; n++ ) src_lats[n] = src_lons[n]/distance;

}  /* grid_search */


/*
  this routine stores the address and weight for four links 
  associated with one destination point in the appropriate address 
  and weight arrays and resizes those arrays if necessary.
*/
void store_link_bilin(REMAPVARS *rv, int dst_add, int *src_add, double *weights)
{
  /*
    input variables:
    int dst_add       ! address on destination grid
    int src_add[4]    ! addresses on source grid
    double weights[4] ! array of remapping weights for these links
  */
  int n;             /* dummy index */
  int num_links_old; /* placeholder for old link number */

  /*
     increment number of links and check to see if remap arrays need
     to be increased to accomodate the new link.  then store the link.
  */
  num_links_old = rv->num_links;
  rv->num_links = num_links_old + 4;

  if ( rv->num_links >= rv->max_links ) 
    resize_remap_vars(rv, rv->resize_increment);

  for ( n = 0; n < 4; n++ )
    {
      rv->grid1_add[num_links_old+n] = src_add[n]-1;
      rv->grid2_add[num_links_old+n] = dst_add;
      rv->wts   [0][num_links_old+n] = weights[n];
    }

} /* store_link_bilin */


/*
  -----------------------------------------------------------------------

  this routine computes the weights for a bilinear interpolation.

  -----------------------------------------------------------------------
*/
void remap_bilin(REMAPGRID *rg, REMAPVARS *rv)
{
  /*   local variables */
  int n,icount;
  int dst_add;        /*  destination addresss */
  int iter;           /*  iteration counters   */

  int src_add[4];     /* address for the four source points */

  double src_lats[4]; /*  latitudes  of four bilinear corners */
  double src_lons[4]; /*  longitudes of four bilinear corners */
  double wgts[4];     /*  bilinear weights for four corners   */

  double plat, plon;             /*  lat/lon coords of destination point    */
  double iguess, jguess;         /*  current guess for bilinear coordinate  */
  double deli, delj;             /*  corrections to i,j                     */
  double dth1, dth2, dth3;       /*  some latitude  differences             */
  double dph1, dph2, dph3;       /*  some longitude differences             */
  double dthp, dphp;             /*  difference between point and sw corner */
  double mat1, mat2, mat3, mat4; /*  matrix elements                        */
  double determinant;            /*  matrix determinant                     */
  double sum_wgts;               /*  sum of weights for normalization       */

  /*
    compute mappings from grid1 to grid2
  */
  if ( rg->grid1_rank != 2 )
    cdoAbort("Can not do bilinear interpolation when grid1_rank != 2"); 

  /*
    loop over destination grid 
  */
  /* grid_loop1 */
  for ( dst_add = 0; dst_add < rg->grid2_size; dst_add++ )
    {
      if ( ! rg->grid2_mask[dst_add] ) continue;

      plat = rg->grid2_center_lat[dst_add];
      plon = rg->grid2_center_lon[dst_add];

      /*  find nearest square of grid points on source grid  */
      grid_search(rg, src_add, src_lats, src_lons, 
		  plat, plon, rg->grid1_dims,
		  rg->grid1_center_lat, rg->grid1_center_lon,
		  rg->grid1_bound_box, rg->bin_addr1, rg->bin_addr2);

      /* check to see if points are land points */

      if ( src_add[0] > 0 )
	for ( n = 0; n < 4; n++ )
	  if ( src_add[n] > 0 ) /* Uwe Schulzweida: check if src_add is valid first */
	    if ( ! rg->grid1_mask[src_add[n]-1] ) src_add[0] = 0;

      /*  if point found, find local i,j coordinates for weights  */

      if ( src_add[0] > 0 )
	{
          rg->grid2_frac[dst_add] = one;

	  /*  iterate to find i,j for bilinear approximation  */

          dth1 = src_lats[1] - src_lats[0];
          dth2 = src_lats[3] - src_lats[0];
          dth3 = src_lats[2] - src_lats[1] - dth2;

          dph1 = src_lons[1] - src_lons[0];
          dph2 = src_lons[3] - src_lons[0];
          dph3 = src_lons[2] - src_lons[1];

          if ( dph1 >  three*pih ) dph1 -= pi2;
          if ( dph2 >  three*pih ) dph2 -= pi2;
          if ( dph3 >  three*pih ) dph3 -= pi2;
          if ( dph1 < -three*pih ) dph1 += pi2;
          if ( dph2 < -three*pih ) dph2 += pi2;
          if ( dph3 < -three*pih ) dph3 += pi2;

          dph3 = dph3 - dph2;

          iguess = half;
          jguess = half;

	  /* Unvectorized loop: break */
          for ( iter = 0; iter < max_iter; iter++ )
	    {
	      dthp = plat - src_lats[0] - dth1*iguess - dth2*jguess - dth3*iguess*jguess;
	      dphp = plon - src_lons[0];

	      if ( dphp >  three*pih ) dphp -= pi2;
	      if ( dphp < -three*pih ) dphp += pi2;

	      dphp = dphp - dph1*iguess - dph2*jguess - dph3*iguess*jguess;

	      mat1 = dth1 + dth3*jguess;
	      mat2 = dth2 + dth3*iguess;
	      mat3 = dph1 + dph3*jguess;
	      mat4 = dph2 + dph3*iguess;

	      determinant = mat1*mat4 - mat2*mat3;

	      deli = (dthp*mat4 - mat2*dphp)/determinant;
	      delj = (mat1*dphp - dthp*mat3)/determinant;

	      if ( fabs(deli) < converge && fabs(delj) < converge ) break;

	      iguess += deli;
	      jguess += delj;
	    }

          if ( iter < max_iter )
	    {
	      /* successfully found i,j - compute weights */

	      wgts[0] = (one-iguess)*(one-jguess);
	      wgts[1] = iguess*(one-jguess);
	      wgts[2] = iguess*jguess;
	      wgts[3] = (one-iguess)*jguess;

	      store_link_bilin(rv, dst_add, src_add, wgts);
	    }
          else
	    {
	      cdoPrint("Point coords: %g %g", plat, plon);
	      cdoPrint("Dest grid lats: %g %g %g %g", src_lats[0], src_lats[1], src_lats[2], src_lats[3]);
	      cdoPrint("Dest grid lons: %g %g %g %g", src_lons[0], src_lons[1], src_lons[2], src_lons[3]);
	      cdoPrint("Dest grid addresses: %d %d %d %d", src_add[0], src_add[1], src_add[2], src_add[3]);
	      cdoPrint("Current i,j : %g %g", iguess, jguess);
	      cdoAbort("Iteration for i,j exceed max iteration count of %d", max_iter);
	    }

	  /*
	    search for bilinear failed - use a distance-weighted
	    average instead (this is typically near the pole)
	  */
	}
      else if ( src_add[0] < 0 )
	{
          for ( n = 0; n < 4; n++ ) src_add[n] = src_add[n] < 0 ? -1*src_add[n] : src_add[n];
          icount = 0;
          for ( n = 0; n < 4; n++ )
	    {
	      if ( rg->grid1_mask[src_add[n]-1] )
		icount++;
	      else
		src_lats[n] = zero;
	    }

          if ( icount > 0 )
	    {
	      /* renormalize weights */
	      sum_wgts = 0.0;
	      for ( n = 0; n < 4; n++ ) sum_wgts += src_lats[n];
	      wgts[0] = src_lats[0]/sum_wgts;
	      wgts[1] = src_lats[1]/sum_wgts;
	      wgts[2] = src_lats[2]/sum_wgts;
	      wgts[3] = src_lats[3]/sum_wgts;

	      rg->grid2_frac[dst_add] = one;

	      store_link_bilin(rv, dst_add, src_add, wgts);
	    }
        }
    } /* grid_loop1 */

} /* remap_bilin */


/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                         */
/*      BICUBIC INTERPOLATION                                              */
/*                                                                         */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/*
  this routine stores the address and weight for four links 
  associated with one destination point in the appropriate address 
  and weight arrays and resizes those arrays if necessary.
*/
void store_link_bicub(REMAPVARS *rv, int dst_add, int *src_add, double weights[4][4])
{
  /*
    input variables:
    int dst_add          ! address on destination grid
    int src_add[4]       ! addresses on source grid
    double weights[4][4] ! array of remapping weights for these links
  */
  int n, k;          /* dummy index */
  int num_links_old; /* placeholder for old link number */

  /*
     increment number of links and check to see if remap arrays need
     to be increased to accomodate the new link.  then store the link.
  */
  num_links_old = rv->num_links;
  rv->num_links = num_links_old + 4;

  if ( rv->num_links >= rv->max_links ) 
    resize_remap_vars(rv, rv->resize_increment);

  for ( n = 0; n < 4; n++ )
    {
      rv->grid1_add[num_links_old+n] = src_add[n]-1;
      rv->grid2_add[num_links_old+n] = dst_add;
      for ( k = 0; k < 4; k++ )
	rv->wts[k][num_links_old+n] = weights[k][n];
    }

} /* store_link_bicub */


/*
  -----------------------------------------------------------------------

  this routine computes the weights for a bicubic interpolation.

  -----------------------------------------------------------------------
*/
void remap_bicub(REMAPGRID *rg, REMAPVARS *rv)
{
  /*   local variables */
  int n,icount;
  int dst_add;        /*  destination addresss */
  int iter;           /*  iteration counters   */

  int src_add[4];     /* address for the four source points */

  double src_lats[4]; /*  latitudes  of four bilinear corners */
  double src_lons[4]; /*  longitudes of four bilinear corners */
  double wgts[4][4];  /*  bicubic weights for four corners    */

  double plat, plon;             /*  lat/lon coords of destination point    */
  double iguess, jguess;         /*  current guess for bilinear coordinate  */
  double deli, delj;             /*  corrections to i,j                     */
  double dth1, dth2, dth3;       /*  some latitude  differences             */
  double dph1, dph2, dph3;       /*  some longitude differences             */
  double dthp, dphp;             /*  difference between point and sw corner */
  double mat1, mat2, mat3, mat4; /*  matrix elements                        */
  double determinant;            /*  matrix determinant                     */
  double sum_wgts;               /*  sum of weights for normalization       */

  /*
    compute mappings from grid1 to grid2
  */
  if ( rg->grid1_rank != 2 )
    cdoAbort("Can not do bicubic interpolation when grid1_rank != 2"); 

  /* loop over destination grid */

  /* grid_loop1 */
  for ( dst_add = 0; dst_add < rg->grid2_size; dst_add++ )
    {
      if ( ! rg->grid2_mask[dst_add] ) continue;

      plat = rg->grid2_center_lat[dst_add];
      plon = rg->grid2_center_lon[dst_add];

      /*  find nearest square of grid points on source grid  */
      grid_search(rg, src_add, src_lats, src_lons, 
		  plat, plon, rg->grid1_dims,
		  rg->grid1_center_lat, rg->grid1_center_lon,
		  rg->grid1_bound_box, rg->bin_addr1, rg->bin_addr2);

      /* check to see if points are land points */

      if ( src_add[0] > 0 )
	for ( n = 0; n < 4; n++ )
	  if ( src_add[n] > 0 ) /* Uwe Schulzweida: check if src_add is valid first */
	    if ( ! rg->grid1_mask[src_add[n]-1] ) src_add[0] = 0;

      /*  if point found, find local i,j coordinates for weights  */

      if ( src_add[0] > 0 )
	{
          rg->grid2_frac[dst_add] = one;

	  /*  iterate to find i,j for bilinear approximation  */

          dth1 = src_lats[1] - src_lats[0];
          dth2 = src_lats[3] - src_lats[0];
          dth3 = src_lats[2] - src_lats[1] - dth2;

          dph1 = src_lons[1] - src_lons[0];
          dph2 = src_lons[3] - src_lons[0];
          dph3 = src_lons[2] - src_lons[1];

          if ( dph1 >  three*pih ) dph1 -= pi2;
          if ( dph2 >  three*pih ) dph2 -= pi2;
          if ( dph3 >  three*pih ) dph3 -= pi2;
          if ( dph1 < -three*pih ) dph1 += pi2;
          if ( dph2 < -three*pih ) dph2 += pi2;
          if ( dph3 < -three*pih ) dph3 += pi2;

          dph3 = dph3 - dph2;

          iguess = half;
          jguess = half;

          for ( iter = 0; iter < max_iter; iter++ )
	    {
	      dthp = plat - src_lats[0] - dth1*iguess - dth2*jguess - dth3*iguess*jguess;
	      dphp = plon - src_lons[0];

	      if ( dphp >  three*pih ) dphp -= pi2;
	      if ( dphp < -three*pih ) dphp += pi2;

	      dphp = dphp - dph1*iguess - dph2*jguess - dph3*iguess*jguess;

	      mat1 = dth1 + dth3*jguess;
	      mat2 = dth2 + dth3*iguess;
	      mat3 = dph1 + dph3*jguess;
	      mat4 = dph2 + dph3*iguess;

	      determinant = mat1*mat4 - mat2*mat3;

	      deli = (dthp*mat4 - mat2*dphp)/determinant;
	      delj = (mat1*dphp - dthp*mat3)/determinant;

	      if ( fabs(deli) < converge && fabs(delj) < converge ) break;

	      iguess += deli;
	      jguess += delj;
	    }

          if ( iter < max_iter )
	    {
	      /* successfully found i,j - compute weights */

	      wgts[0][0] = (one - jguess*jguess*(three-two*jguess))*
		           (one - iguess*iguess*(three-two*iguess));
	      wgts[0][1] = (one - jguess*jguess*(three-two*jguess))*
                                  iguess*iguess*(three-two*iguess);
	      wgts[0][2] =        jguess*jguess*(three-two*jguess)*
                                  iguess*iguess*(three-two*iguess);
	      wgts[0][3] =        jguess*jguess*(three-two*jguess)*
                           (one - iguess*iguess*(three-two*iguess));
	      wgts[1][0] = (one - jguess*jguess*(three-two*jguess))*
                                  iguess*(iguess-one)*(iguess-one);
	      wgts[1][1] = (one - jguess*jguess*(three-two*jguess))*
                                  iguess*iguess*(iguess-one);
	      wgts[1][2] =        jguess*jguess*(three-two*jguess)*
                                  iguess*iguess*(iguess-one);
	      wgts[1][3] =        jguess*jguess*(three-two*jguess)*
                                  iguess*(iguess-one)*(iguess-one);
	      wgts[2][0] =        jguess*(jguess-one)*(jguess-one)*
                           (one - iguess*iguess*(three-two*iguess));
	      wgts[2][1] =        jguess*(jguess-one)*(jguess-one)*
                                  iguess*iguess*(three-two*iguess);
	      wgts[2][2] =        jguess*jguess*(jguess-one)*
                                  iguess*iguess*(three-two*iguess);
	      wgts[2][3] =        jguess*jguess*(jguess-one)*
                           (one - iguess*iguess*(three-two*iguess));
	      wgts[3][0] =        iguess*(iguess-one)*(iguess-one)*
                                  jguess*(jguess-one)*(jguess-one);
              wgts[3][1] =        iguess*iguess*(iguess-one)*
                                  jguess*(jguess-one)*(jguess-one);
	      wgts[3][2] =        iguess*iguess*(iguess-one)*
                                  jguess*jguess*(jguess-one);
	      wgts[3][3] =        iguess*(iguess-one)*(iguess-one)*
                                  jguess*jguess*(jguess-one);

	      store_link_bicub(rv, dst_add, src_add, wgts);
	    }
          else
	    {
	      cdoAbort("Iteration for i,j exceed max iteration count of %d", max_iter);
	    }
	  
	  /*
	    search for bilinear failed - use a distance-weighted
	    average instead (this is typically near the pole)
	  */
	}
      else if ( src_add[0] < 0 )
	{
          for ( n = 0; n < 4; n++ ) src_add[n] = src_add[n] < 0 ? -1*src_add[n] : src_add[n];
          icount = 0;
          for ( n = 0; n < 4; n++ )
	    {
	      if ( rg->grid1_mask[src_add[n]-1] )
		icount++;
	      else
		src_lats[n] = zero;
	    }

          if ( icount > 0 )
	    {
	      /* renormalize weights */
	      sum_wgts = 0.0;
	      for ( n = 0; n < 4; n++ ) sum_wgts += src_lats[n];
	      for ( n = 0; n < 4; n++ ) wgts[0][n] = src_lats[n]/sum_wgts;
	      for ( n = 0; n < 4; n++ ) wgts[1][n] = zero;
	      for ( n = 0; n < 4; n++ ) wgts[2][n] = zero;
	      for ( n = 0; n < 4; n++ ) wgts[3][n] = zero;

	      rg->grid2_frac[dst_add] = one;
	      store_link_bicub(rv, dst_add, src_add, wgts);
	    }
        }
    } /* grid_loop1 */

} /* remap_bicub */


/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                         */
/*      INTERPOLATION USING A DISTANCE-WEIGHTED AVERAGE                    */
/*                                                                         */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

#define  num_neighbors  4  /* num nearest neighbors to interpolate from */

/*
   this routine finds the closest num_neighbor points to a search 
   point and computes a distance to each of the neighbors.
*/
void grid_search_nbr(REMAPGRID *rg, int *nbr_add, double *nbr_dist, double plat, double plon, 
		     double coslat_dst, double coslon_dst, double sinlat_dst, double sinlon_dst,
		     int *src_bin_add, int *dst_bin_add,
		     double *sinlat, double *coslat, double *sinlon, double *coslon)
{
  /*
    output variables:

    int nbr_add[num_neighbors]     ! address of each of the closest points
    double nbr_dist[num_neighbors] ! distance to each of the closest points
  */

  /*
    input variables:

    int src_bin_add[][2]  ! search bins for restricting search
    int dst_bin_add[][2]   

    double plat,         ! latitude  of the search point
    double plon,         ! longitude of the search point
    double coslat_dst,   ! cos(lat)  of the search point
    double coslon_dst,   ! cos(lon)  of the search point
    double sinlat_dst,   ! sin(lat)  of the search point
    double sinlon_dst    ! sin(lon)  of the search point
  */
  /*  local variables */

  int n, nmax, nadd, nchk;
  int min_add = 0, max_add = 0, nm1, np1, i, j, ip1, im1, jp1, jm1;

  double distance;     /* angular distance */

  /*  loop over source grid and find nearest neighbors */

  /*
    restrict the search using search bins
    expand the bins to catch neighbors
  */
  if ( rg->restrict_type == restrict_latitude )
    {
      for ( n = 0; n < rg->num_srch_bins; n++ )
	{
	  if ( plat >= rg->bin_lats[2*n+0] && plat <= rg->bin_lats[2*n+1] )
	    {
	      min_add = src_bin_add[2*n+0];
	      max_add = src_bin_add[2*n+1];

	      nm1 = MAX(n-1, 0);
	      np1 = MIN(n+1, rg->num_srch_bins-1);

	      min_add = MIN(min_add, src_bin_add[2*nm1+0]);
	      max_add = MAX(max_add, src_bin_add[2*nm1+1]);
	      min_add = MIN(min_add, src_bin_add[2*np1+0]);
	      max_add = MAX(max_add, src_bin_add[2*np1+1]);
	    }
	}
    }
  else if ( rg->restrict_type == restrict_latlon )
    {
      n = 0;
      nmax = NINT(sqrt((double)rg->num_srch_bins)) - 1;
      for ( j = 0; j < nmax; j++ )
	{
	  jp1 = MIN(j+1,nmax);
	  jm1 = MAX(j-1,0);
	  for ( i = 0; i < nmax; i++ )
	    {
	      ip1 = MIN(i+1, nmax);
	      im1 = MAX(i-1, 0);

	      n = n+1;
	      if ( plat >= rg->bin_lats[2*n+0] && plat <= rg->bin_lats[2*n+1] &&
		   plon >= rg->bin_lons[2*n+0] && plon <= rg->bin_lons[2*n+2] )
		{
		  min_add = src_bin_add[2*n+0];
		  max_add = src_bin_add[2*n+1];

		  nm1 = (jm1-1)*nmax + im1;
		  np1 = (jp1-1)*nmax + ip1;
		  nm1 = MAX(nm1, 0);
		  np1 = MIN(np1, rg->num_srch_bins-1);

		  min_add = MIN(min_add, src_bin_add[2*nm1+0]);
		  max_add = MAX(max_add, src_bin_add[2*nm1+1]);
		  min_add = MIN(min_add, src_bin_add[2*np1+0]);
		  max_add = MAX(max_add, src_bin_add[2*np1+1]);
		}
	    }
	}
    }
  else
    cdoAbort("unknown search restriction method");


  /* initialize distance and address arrays */

  for ( n = 0; n < num_neighbors; n++ )
    {
      nbr_add[n]  = 0;
      nbr_dist[n] = bignum;
    }

  /* Unvectorized loop: break (86% CPU time) (num_neighbors is constant 4) */
  for ( nadd = min_add; nadd <= max_add; nadd++ )
    {
      /* find distance to this point */

      distance = acos(sinlat_dst*sinlat[nadd] + coslat_dst*coslat[nadd]*
                     (coslon_dst*coslon[nadd] + sinlon_dst*sinlon[nadd]));
      /*
         store the address and distance if this is one of the
	 smallest four so far
      */
      for ( nchk = 0; nchk < num_neighbors; nchk++ )
	{
          if ( distance < nbr_dist[nchk] )
	    {
	      for ( n = num_neighbors-1; n > nchk; n-- )
		{
		  nbr_add[n]  = nbr_add[n-1];
		  nbr_dist[n] = nbr_dist[n-1];
		}
	      nbr_add[nchk]  = nadd + 1;
	      nbr_dist[nchk] = distance;
	      break;
	    }
        }
    }

}  /*  grid_search_nbr  */


/*
  this routine stores the address and weight for this link in
  the appropriate address and weight arrays and resizes those
  arrays if necessary.
*/
void store_link_nbr(REMAPVARS *rv, int add1, int add2, double *weights)
{
  /*
    input variables:
    int  add1         ! address on grid1
    int  add2         ! address on grid2
    double weights[]  ! array of remapping weights for this link
  */
  int k;

  /*
     increment number of links and check to see if remap arrays need
     to be increased to accomodate the new link.  then store the link.
  */
  rv->num_links = rv->num_links + 1;

  if ( rv->num_links >= rv->max_links ) 
    resize_remap_vars(rv, rv->resize_increment);

  rv->grid1_add[rv->num_links-1] = add1;
  rv->grid2_add[rv->num_links-1] = add2;

#if defined (SX)
#pragma vdir novector
#endif
  for ( k = 0; k < rv->num_wts; k++ )
    rv->wts[k][rv->num_links-1] = weights[k];

} /* store_link_nbr */


/*
  -----------------------------------------------------------------------

   this routine computes the inverse-distance weights for a
   nearest-neighbor interpolation.

  -----------------------------------------------------------------------
*/
void remap_distwgt(REMAPGRID *rg, REMAPVARS *rv)
{
  static char func[] = "remap_distwgt";

  /*   local variables */

  int grid1_size;
  int grid2_size;
  int nbr_mask[num_neighbors];    /* mask at nearest neighbors */
  int n;
  int dst_add;                    /* destination address */
  int nbr_add[num_neighbors];     /* source address at nearest neighbors     */
  double nbr_dist[num_neighbors]; /* angular distance four nearest neighbors */
  double coslat_dst;       /* cos(lat) of destination grid point */
  double coslon_dst;       /* cos(lon) of destination grid point */
  double sinlat_dst;       /* sin(lat) of destination grid point */
  double sinlon_dst;       /* sin(lon) of destination grid point */
  double dist_tot;         /* sum of neighbor distances (for normalizing) */
  double *coslat, *sinlat; /* cosine, sine of grid lats (for distance)    */
  double *coslon, *sinlon; /* cosine, sine of grid lons (for distance)    */
  double *wgtstmp;         /* an array to hold the link weight            */

  /*
    compute mappings from grid1 to grid2
  */

  grid1_size = rg->grid1_size;
  grid2_size = rg->grid2_size;

  /*  allocate wgtstmp to be consistent with store_link interface */

  wgtstmp = (double *) malloc(rv->num_wts*sizeof(double));

  /* compute cos, sin of lat/lon on source grid for distance calculations */

  coslat = (double *) malloc(grid1_size*sizeof(double));
  coslon = (double *) malloc(grid1_size*sizeof(double));
  sinlat = (double *) malloc(grid1_size*sizeof(double));
  sinlon = (double *) malloc(grid1_size*sizeof(double));

  for ( n = 0; n < grid1_size; n++ )
    {
      coslat[n] = cos(rg->grid1_center_lat[n]);
      coslon[n] = cos(rg->grid1_center_lon[n]);
      sinlat[n] = sin(rg->grid1_center_lat[n]);
      sinlon[n] = sin(rg->grid1_center_lon[n]);
    }

  /* loop over destination grid  */

  /* grid_loop1 */
  for ( dst_add = 0; dst_add < grid2_size; dst_add++ )
    {
      if ( ! rg->grid2_mask[dst_add] ) continue;

      coslat_dst = cos(rg->grid2_center_lat[dst_add]);
      coslon_dst = cos(rg->grid2_center_lon[dst_add]);
      sinlat_dst = sin(rg->grid2_center_lat[dst_add]);
      sinlon_dst = sin(rg->grid2_center_lon[dst_add]);
	
      /* find nearest grid points on source grid and  distances to each point */
      grid_search_nbr(rg, nbr_add, nbr_dist, 
		      rg->grid2_center_lat[dst_add],
		      rg->grid2_center_lon[dst_add],
		      coslat_dst, coslon_dst, 
		      sinlat_dst, sinlon_dst,
		      rg->bin_addr1, rg->bin_addr2,
		      sinlat, coslat, sinlon, coslon);

      /*
         compute weights based on inverse distance
         if mask is false, eliminate those points
      */
      dist_tot = zero;
      for ( n = 0; n < num_neighbors; n++ )
	{
	  nbr_mask[n] = FALSE;

	  /* Uwe Schulzweida: check if nbr_add is valid first */
	  if ( nbr_add[n] > 0 )
	    if ( rg->grid1_mask[nbr_add[n]-1] )
	      {
		nbr_dist[n] = one/nbr_dist[n];
		dist_tot = dist_tot + nbr_dist[n];
		nbr_mask[n] = TRUE;
	      }
	}

      /* normalize weights and store the link */

      for ( n = 0; n < num_neighbors; n++ )
	{
          if ( nbr_mask[n] )
	    {
	      wgtstmp[0] = nbr_dist[n]/dist_tot;
	      store_link_nbr(rv, nbr_add[n]-1, dst_add, wgtstmp);
	      rg->grid2_frac[dst_add] = one;
	    }
	}

    } /* grid_loop1 */

    free(coslat);
    free(coslon);
    free(sinlat);
    free(sinlon);

    free(wgtstmp);

}  /* remap_distwgt */


/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                         */
/*      INTERPOLATION USING A DISTANCE-WEIGHTED AVERAGE WITH 1 NEIGHBOR    */
/*                                                                         */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/*
   this routine finds the closest num_neighbor points to a search 
   point and computes a distance to each of the neighbors.
*/
void grid_search_nbr1(REMAPGRID *rg, int *nbr_add, double *nbr_dist, double plat, double plon, 
		      double coslat_dst, double coslon_dst, double sinlat_dst, double sinlon_dst,
		      int *src_bin_add, int *dst_bin_add,
		      double *sinlat, double *coslat, double *sinlon, double *coslon)
{
  /*
    output variables:

    int nbr_add[0]     ! address of each of the closest points
    double nbr_dist[0] ! distance to each of the closest points
  */

  /*
    input variables:

    int src_bin_add[][2]  ! search bins for restricting search
    int dst_bin_add[][2]   

    double plat,         ! latitude  of the search point
    double plon,         ! longitude of the search point
    double coslat_dst,   ! cos(lat)  of the search point
    double coslon_dst,   ! cos(lon)  of the search point
    double sinlat_dst,   ! sin(lat)  of the search point
    double sinlon_dst    ! sin(lon)  of the search point
  */
  /*  local variables */

  int n, nmax, nadd;
  int min_add = 0, max_add = 0, nm1, np1, i, j, ip1, im1, jp1, jm1;

  double distance;     /* angular distance */

  /*  loop over source grid and find nearest neighbors */

  /*
    restrict the search using search bins
    expand the bins to catch neighbors
  */
  if ( rg->restrict_type == restrict_latitude )
    {
      for ( n = 0; n < rg->num_srch_bins; n++ )
	{
	  if ( plat >= rg->bin_lats[2*n+0] && plat <= rg->bin_lats[2*n+1] )
	    {
	      min_add = src_bin_add[2*n+0];
	      max_add = src_bin_add[2*n+1];

	      nm1 = MAX(n-1, 0);
	      np1 = MIN(n+1, rg->num_srch_bins-1);

	      min_add = MIN(min_add, src_bin_add[2*nm1+0]);
	      max_add = MAX(max_add, src_bin_add[2*nm1+1]);
	      min_add = MIN(min_add, src_bin_add[2*np1+0]);
	      max_add = MAX(max_add, src_bin_add[2*np1+1]);
	    }
	}
    }
  else if ( rg->restrict_type == restrict_latlon )
    {
      n = 0;
      nmax = NINT(sqrt((double)rg->num_srch_bins)) - 1;
      for ( j = 0; j < nmax; j++ )
	{
	  jp1 = MIN(j+1,nmax);
	  jm1 = MAX(j-1,0);
	  for ( i = 0; i < nmax; i++ )
	    {
	      ip1 = MIN(i+1, nmax);
	      im1 = MAX(i-1, 0);

	      n = n+1;
	      if ( plat >= rg->bin_lats[2*n+0] && plat <= rg->bin_lats[2*n+1] &&
		   plon >= rg->bin_lons[2*n+0] && plon <= rg->bin_lons[2*n+2] )
		{
		  min_add = src_bin_add[2*n+0];
		  max_add = src_bin_add[2*n+1];

		  nm1 = (jm1-1)*nmax + im1;
		  np1 = (jp1-1)*nmax + ip1;
		  nm1 = MAX(nm1, 0);
		  np1 = MIN(np1, rg->num_srch_bins-1);

		  min_add = MIN(min_add, src_bin_add[2*nm1+0]);
		  max_add = MAX(max_add, src_bin_add[2*nm1+1]);
		  min_add = MIN(min_add, src_bin_add[2*np1+0]);
		  max_add = MAX(max_add, src_bin_add[2*np1+1]);
		}
	    }
	}
    }
  else
    cdoAbort("unknown search restriction method");


  /* initialize distance and address arrays */

  nbr_add[0]  = 0;
  nbr_dist[0] = bignum;

  for ( nadd = min_add; nadd <= max_add; nadd++ )
    {
      /* find distance to this point */

      distance = acos(sinlat_dst*sinlat[nadd] + coslat_dst*coslat[nadd]*
                     (coslon_dst*coslon[nadd] + sinlon_dst*sinlon[nadd]));
      /*
         store the address and distance if this is the smallest so far
      */
      if ( distance < nbr_dist[0] )
	{
	  nbr_add[0]  = nadd + 1;
	  nbr_dist[0] = distance;
        }
    }

}  /*  grid_search_nbr1  */


/*
  this routine stores the address and weight for this link in
  the appropriate address and weight arrays and resizes those
  arrays if necessary.
*/
void store_link_nbr1(REMAPVARS *rv, int add1, int add2, double *weights)
{
  /*
    input variables:
    int  add1         ! address on grid1
    int  add2         ! address on grid2
    double weights[]  ! array of remapping weights for this link
  */
  int k;

  /*
     increment number of links and check to see if remap arrays need
     to be increased to accomodate the new link.  then store the link.
  */
  rv->num_links = rv->num_links + 1;

  if ( rv->num_links >= rv->max_links ) 
    resize_remap_vars(rv, rv->resize_increment);

  rv->grid1_add[rv->num_links-1] = add1;
  rv->grid2_add[rv->num_links-1] = add2;

#if defined (SX)
#pragma vdir novector
#endif
  for ( k = 0; k < rv->num_wts; k++ )
    rv->wts[k][rv->num_links-1] = weights[k];

} /* store_link_nbr1 */


/*
  -----------------------------------------------------------------------

   this routine computes the inverse-distance weights for a
   nearest-neighbor interpolation.

  -----------------------------------------------------------------------
*/
void remap_distwgt1(REMAPGRID *rg, REMAPVARS *rv)
{
  static char func[] = "remap_distwgt";

  /*   local variables */

  int grid1_size;
  int grid2_size;
  int nbr_mask;    /* mask at nearest neighbors */
  int n;
  int dst_add;                    /* destination address */
  int nbr_add;     /* source address at nearest neighbors     */
  double nbr_dist; /* angular distance four nearest neighbors */
  double coslat_dst;       /* cos(lat) of destination grid point */
  double coslon_dst;       /* cos(lon) of destination grid point */
  double sinlat_dst;       /* sin(lat) of destination grid point */
  double sinlon_dst;       /* sin(lon) of destination grid point */
  double dist_tot;         /* sum of neighbor distances (for normalizing) */
  double *coslat, *sinlat; /* cosine, sine of grid lats (for distance)    */
  double *coslon, *sinlon; /* cosine, sine of grid lons (for distance)    */
  double *wgtstmp;         /* an array to hold the link weight            */

  /*
    compute mappings from grid1 to grid2
  */

  grid1_size = rg->grid1_size;
  grid2_size = rg->grid2_size;

  /*  allocate wgtstmp to be consistent with store_link interface */

  wgtstmp = (double *) malloc(rv->num_wts*sizeof(double));

  /* compute cos, sin of lat/lon on source grid for distance calculations */

  coslat = (double *) malloc(grid1_size*sizeof(double));
  coslon = (double *) malloc(grid1_size*sizeof(double));
  sinlat = (double *) malloc(grid1_size*sizeof(double));
  sinlon = (double *) malloc(grid1_size*sizeof(double));

  for ( n = 0; n < grid1_size; n++ )
    {
      coslat[n] = cos(rg->grid1_center_lat[n]);
      coslon[n] = cos(rg->grid1_center_lon[n]);
      sinlat[n] = sin(rg->grid1_center_lat[n]);
      sinlon[n] = sin(rg->grid1_center_lon[n]);
    }

  /* loop over destination grid  */

  /* grid_loop1 */
  for ( dst_add = 0; dst_add < grid2_size; dst_add++ )
    {
      if ( ! rg->grid2_mask[dst_add] ) continue;

      coslat_dst = cos(rg->grid2_center_lat[dst_add]);
      coslon_dst = cos(rg->grid2_center_lon[dst_add]);
      sinlat_dst = sin(rg->grid2_center_lat[dst_add]);
      sinlon_dst = sin(rg->grid2_center_lon[dst_add]);
	
      /* find nearest grid points on source grid and  distances to each point */

      grid_search_nbr1(rg, &nbr_add, &nbr_dist,
		       rg->grid2_center_lat[dst_add],
		       rg->grid2_center_lon[dst_add],
		       coslat_dst, coslon_dst, 
		       sinlat_dst, sinlon_dst,
		       rg->bin_addr1, rg->bin_addr2,
		       sinlat, coslat, sinlon, coslon);

      /*
         compute weights based on inverse distance
         if mask is false, eliminate those points
      */
      dist_tot = zero;

      nbr_mask = FALSE;

      /* Uwe Schulzweida: check if nbr_add is valid first */
      if ( nbr_add > 0 )
	if ( rg->grid1_mask[nbr_add-1] )
	  {
	    nbr_dist = one/nbr_dist;
	    dist_tot = dist_tot + nbr_dist;
	    nbr_mask = TRUE;
	  }

      /* normalize weights and store the link */

      if ( nbr_mask )
	{
	  wgtstmp[0] = nbr_dist/dist_tot;
	  store_link_nbr1(rv, nbr_add-1, dst_add, wgtstmp);
	  rg->grid2_frac[dst_add] = one;
	}

    } /* grid_loop1 */

    free(coslat);
    free(coslon);
    free(sinlat);
    free(sinlon);

    free(wgtstmp);

}  /* remap_distwgt1 */


/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                         */
/*      CONSERVATIVE INTERPOLATION                                         */
/*                                                                         */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/* static double north_thresh = 1.45;  */ /* threshold for coord transf. */
static double north_thresh = 2.00;  /* threshold for coord transf. */
static double south_thresh =-2.00;  /* threshold for coord transf. */

/*
    this routine is identical to the intersection routine except
    that a coordinate transformation (using a Lambert azimuthal
    equivalent projection) is performed to treat polar cells more
    accurately.
*/
void pole_intersection(int *location,
		       double *intrsct_lat, double *intrsct_lon, int *lcoinc, int *lthresh,
		       double beglat, double beglon, double endlat, double endlon,
		       double *begseg, int lrevers,
		       int num_srch_cells, int srch_corners, int *srch_add,
		       double *srch_corner_lat, double *srch_corner_lon,
		       int *luse_last, double *intrsct_x, double *intrsct_y,
		       int *avoid_pole_count, double *avoid_pole_offset)
{
  static char func[] = "pole_intersection";
  /*
    intent(in): 
    double beglat, beglon,  ! beginning lat/lon endpoints for segment
    double endlat, endlon   ! ending    lat/lon endpoints for segment
    int    lrevers          ! flag true if segment integrated in reverse

    intent(inout) :
    double begseg[2] ! begin lat/lon of full segment
    int *location    ! address in destination array containing this
                     ! segment -- also may contain last location on entry
    int *lthresh     ! flag segment crossing threshold boundary

    intent(out): 
    *int lcoinc      ! flag segment coincident with grid line
    double *intrsct_lat, *intrsct_lon ! lat/lon coords of next intersect.
  */
  /* local variables */
  int n, next_n, cell;
  int loutside; /* flags points outside grid */

  double pi4, rns;           /*  north/south conversion */
  double x1, x2;             /*  local x variables for segment */
  double y1, y2;             /*  local y variables for segment */
  double begx, begy;         /*  beginning x,y variables for segment */
  double endx, endy;         /*  beginning x,y variables for segment */
  double begsegx, begsegy;   /*  beginning x,y variables for segment */
  double grdx1, grdx2;       /*  local x variables for grid cell */
  double grdy1, grdy2;       /*  local y variables for grid cell */
  double vec1_y, vec1_x;     /*  vectors and cross products used */
  double vec2_y, vec2_x;     /*  during grid search */
  double cross_product, eps; /*  eps=small offset away from intersect */
  double s1, s2, determ;     /*  variables used for linear solve to   */
  double mat1, mat2, mat3, mat4, rhs1, rhs2;  /* find intersection */

  double *srch_corner_x;     /*  x of each corner of srch cells */
  double *srch_corner_y;     /*  y of each corner of srch cells */


  /* initialize defaults, flags, etc. */

  if ( ! *lthresh ) *location = -1;
  *lcoinc = FALSE;
  *intrsct_lat = endlat;
  *intrsct_lon = endlon;

  loutside = FALSE;
  s1 = zero;

  /* convert coordinates */

  srch_corner_x = (double *) malloc(srch_corners*num_srch_cells*sizeof(double));
  srch_corner_y = (double *) malloc(srch_corners*num_srch_cells*sizeof(double));

  if ( beglat > zero )
    {
      pi4 = quart*pi;
      rns = one;
    }
  else
    {
      pi4 = -quart*pi;
      rns = -one;
    }

  if ( *luse_last )
    {
      x1 = *intrsct_x;
      y1 = *intrsct_y;
    }
  else
    {
      x1 = rns*two*sin(pi4 - half*beglat)*cos(beglon);
      y1 =     two*sin(pi4 - half*beglat)*sin(beglon);
      *luse_last = TRUE;
    }

  x2 = rns*two*sin(pi4 - half*endlat)*cos(endlon);
  y2 =     two*sin(pi4 - half*endlat)*sin(endlon);

  for ( n = 0; n < srch_corners*num_srch_cells; n++ )
    {
      srch_corner_x[n] = rns*two*sin(pi4 - half*srch_corner_lat[n])*
	                         cos(srch_corner_lon[n]);
      srch_corner_y[n] =     two*sin(pi4 - half*srch_corner_lat[n])*
	                         sin(srch_corner_lon[n]);
    }

  begx = x1;
  begy = y1;
  endx = x2;
  endy = y2;
  begsegx = rns*two*sin(pi4 - half*begseg[0])*cos(begseg[1]);
  begsegy =     two*sin(pi4 - half*begseg[0])*sin(begseg[1]);
  *intrsct_x = endx;
  *intrsct_y = endy;

  /*
     search for location of this segment in ocean grid using cross
     product method to determine whether a point is enclosed by a cell
  */
  /*    call timer_start(12) */

  while (TRUE) /* srch_loop */
    {
      /* if last segment crossed threshold, use that location */

      if ( *lthresh )
	{
	  for ( cell=0; cell < num_srch_cells; cell++ )
	    if ( srch_add[cell] == *location )
	      {
		eps = tiny;
		goto after_srch_loop;
	      }
	}

      /* otherwise normal search algorithm */

      for ( cell = 0; cell < num_srch_cells; cell++ ) /* cell_loop  */
	{
	  for ( n = 0; n < srch_corners; n++ ) /* corner_loop */
	    {
	      next_n = (n+1)%srch_corners;

	      /*
		here we take the cross product of the vector making 
		up each cell side with the vector formed by the vertex
		and search point.  if all the cross products are 
		positive, the point is contained in the cell.
	      */
	      vec1_x = srch_corner_x[cell*srch_corners+next_n] - srch_corner_x[cell*srch_corners+n];
	      vec1_y = srch_corner_y[cell*srch_corners+next_n] - srch_corner_y[cell*srch_corners+n];
	      vec2_x = x1 - srch_corner_x[cell*srch_corners+n];
	      vec2_y = y1 - srch_corner_y[cell*srch_corners+n];

	      /* if endpoint coincident with vertex, offset the endpoint */

	      if ( vec2_x == 0 && vec2_y == 0 )
		{
		  x1 = x1 + 1.e-10*(x2 - x1);
		  y1 = y1 + 1.e-10*(y2 - y1);
		  vec2_x = x1 - srch_corner_x[cell*srch_corners+n];
		  vec2_y = y1 - srch_corner_y[cell*srch_corners+n];
		}

	      cross_product = vec1_x*vec2_y - vec2_x*vec1_y;

	      /*
		if the cross product for a side is zero, the point 
                  lies exactly on the side or the length of a side
                  is zero.  if the length is zero set det > 0.
                  otherwise, perform another cross 
                  product between the side and the segment itself. 
	        if this cross product is also zero, the line is 
	          coincident with the cell boundary - perform the 
                  dot product and only choose the cell if the dot 
                  product is positive (parallel vs anti-parallel).
	      */
	      if ( cross_product == zero )
		{
		  if ( vec1_x != zero || vec1_y != 0 )
		    {
		      vec2_x = x2 - x1;
		      vec2_y = y2 - y1;
		      cross_product = vec1_x*vec2_y - vec2_x*vec1_y;
		    }
		  else
		    cross_product = one;

		  if ( cross_product == zero )
		    {
		      *lcoinc = TRUE;
		      cross_product = vec1_x*vec2_x + vec1_y*vec2_y;
		      if ( lrevers ) cross_product = -cross_product;
		    }
		}

	      /* if cross product is less than zero, this cell doesn't work */

	      if ( cross_product < zero ) break; /* corner_loop */
	     
	    } /* corner_loop */

	  /* if cross products all positive, we found the location */

	  if  ( n >= srch_corners )
	    {
	      *location = srch_add[cell];
	      /*
		if the beginning of this segment was outside the
		grid, invert the segment so the intersection found
		will be the first intersection with the grid
	      */
	      if ( loutside )
		{
		  x2 = begx;
		  y2 = begy;
		  *location = -1;
		  eps  = -tiny;
		}
	      else
		eps  = tiny;
            
	      goto after_srch_loop;
	    }

	  /* otherwise move on to next cell */

	} /* cell_loop */

     /*
       if no cell found, the point lies outside the grid.
       take some baby steps along the segment to see if any
       part of the segment lies inside the grid.  
     */
      loutside = TRUE;
      s1 = s1 + 0.001;
      x1 = begx + s1*(x2 - begx);
      y1 = begy + s1*(y2 - begy);

      /* reached the end of the segment and still outside the grid return no intersection */

      if ( s1 >= one )
	{
          free(srch_corner_y);
          free(srch_corner_x);
          *luse_last = FALSE;
          return;
	}
    } /* srch_loop */

 after_srch_loop:

  /*      call timer_stop(12)  */

  /*
    now that a cell is found, search for the next intersection.
    loop over sides of the cell to find intersection with side
    must check all sides for coincidences or intersections
  */
  /*   call timer_start(13) */

  for ( n = 0; n < srch_corners; n++ ) /* intrsct_loop */
    {
      next_n = (n+1)%srch_corners;

      grdy1 = srch_corner_y[cell*srch_corners+n];
      grdy2 = srch_corner_y[cell*srch_corners+next_n];
      grdx1 = srch_corner_x[cell*srch_corners+n];
      grdx2 = srch_corner_x[cell*srch_corners+next_n];

      /* set up linear system to solve for intersection */

      mat1 = x2 - x1;
      mat2 = grdx1 - grdx2;
      mat3 = y2 - y1;
      mat4 = grdy1 - grdy2;
      rhs1 = grdx1 - x1;
      rhs2 = grdy1 - y1;

      determ = mat1*mat4 - mat2*mat3;

      /*
         if the determinant is zero, the segments are either 
           parallel or coincident.  coincidences were detected 
           above so do nothing.
         if the determinant is non-zero, solve for the linear 
           parameters s for the intersection point on each line 
           segment.
         if 0<s1,s2<1 then the segment intersects with this side.
           return the point of intersection (adding a small
           number so the intersection is off the grid line).
      */
      if ( fabs(determ) > 1.e-30 )
	{
          s1 = (rhs1*mat4 - mat2*rhs2)/determ;
          s2 = (mat1*rhs2 - rhs1*mat3)/determ;

          if ( s2 >= zero && s2 <= one && s1 > zero && s1 <= one )
	    {
	      /*
		recompute intersection based on full segment
		so intersections are consistent for both sweeps
	      */
	      if ( ! loutside )
		{
		  mat1 = x2 - begsegx;
		  mat3 = y2 - begsegy;
		  rhs1 = grdx1 - begsegx;
		  rhs2 = grdy1 - begsegy;
		}
	      else
		{
		  mat1 = x2 - endx;
		  mat3 = y2 - endy;
		  rhs1 = grdx1 - endx;
		  rhs2 = grdy1 - endy;
		}

	      determ = mat1*mat4 - mat2*mat3;

	      /*
		sometimes due to roundoff, the previous 
		determinant is non-zero, but the lines
		are actually coincident.  if this is the
		case, skip the rest.
	      */
             if ( determ != zero )
	       {
		 s1 = (rhs1*mat4 - mat2*rhs2)/determ;
		 s2 = (mat1*rhs2 - rhs1*mat3)/determ;

		 if ( ! loutside )
		   {
		     *intrsct_x = begsegx + s1*mat1;
		     *intrsct_y = begsegy + s1*mat3;
		   }
		 else 
		   {
		     *intrsct_x = endx + s1*mat1;
		     *intrsct_y = endy + s1*mat3;
		   }

		 /* convert back to lat/lon coordinates */

		 *intrsct_lon = rns*atan2(*intrsct_y, *intrsct_x);
		 if ( *intrsct_lon < zero ) 
		   *intrsct_lon = *intrsct_lon + pi2;
		 
		 if ( fabs(*intrsct_x) > 1.e-10 )
		   *intrsct_lat = (pi4 - asin(rns*half*(*intrsct_x)/cos(*intrsct_lon)))*two;
		 else if ( fabs(*intrsct_y) > 1.e-10 )
		   *intrsct_lat = (pi4 - asin(half*(*intrsct_y)/sin(*intrsct_lon)))*two;
		 else
		   *intrsct_lat = two*pi4;

		 /* add offset in transformed space for next pass. */

		 if ( s1 - eps/determ < one )
		   {
		     *intrsct_x = *intrsct_x - mat1*(eps/determ);
		     *intrsct_y = *intrsct_y - mat3*(eps/determ);
		   }
		 else
		   {
		     if ( ! loutside)
		       {
			 *intrsct_x = endx;
			 *intrsct_y = endy;
			 *intrsct_lat = endlat;
			 *intrsct_lon = endlon;
		       }
		     else 
		       {
			 *intrsct_x = begsegx;
			 *intrsct_y = begsegy;
			 *intrsct_lat = begseg[0];
			 *intrsct_lon = begseg[1];
		       }
		   }

		 break; /* intrsct_loop */
	       }
	    }
	}
 
      /* no intersection this side, move on to next side */

    } /* intrsct_loop */

  /*  call timer_stop(13) */

  free(srch_corner_y);
  free(srch_corner_x);

  /*
     if segment manages to cross over pole, shift the beginning 
     endpoint in order to avoid hitting pole directly
     (it is ok for endpoint to be pole point)
  */

  if ( fabs(*intrsct_x) < 1.e-10 && fabs(*intrsct_y) < 1.e-10 &&
       (endx != zero  && endy != 0) )
    {
      if ( *avoid_pole_count > 2 )
	{
	  *avoid_pole_count  = 0;
	  *avoid_pole_offset = 10.*(*avoid_pole_offset);
        }

      cross_product = begsegx*(endy-begsegy) - begsegy*(endx-begsegx);
      *intrsct_lat = begseg[0];
      if ( cross_product*(*intrsct_lat) > zero )
	{
          *intrsct_lon = beglon    + *avoid_pole_offset;
          begseg[1]    = begseg[1] + *avoid_pole_offset;
	}
      else
	{
          *intrsct_lon = beglon    - *avoid_pole_offset;
          begseg[1]    = begseg[1] - *avoid_pole_offset;
        }

      *avoid_pole_count = *avoid_pole_count + 1;
      *luse_last = FALSE;
    }
  else
    {
      *avoid_pole_count  = 0;
      *avoid_pole_offset = tiny;
    }

  /*
     if the segment crosses a pole threshold, reset the intersection
     to be the threshold latitude and do not reuse x,y intersect
     on next entry.  only check if did not cross threshold last
     time - sometimes the coordinate transformation can place a
     segment on the other side of the threshold again
  */
  if ( *lthresh )
    {
      if ( *intrsct_lat > north_thresh || *intrsct_lat < south_thresh )
	*lthresh = FALSE;
    }
  else if ( beglat > zero && *intrsct_lat < north_thresh )
    {
      mat4 = endlat - begseg[0];
      mat3 = endlon - begseg[1];
      if ( mat3 >  pi ) mat3 = mat3 - pi2;
      if ( mat3 < -pi ) mat3 = mat3 + pi2;
      *intrsct_lat = north_thresh - tiny;
      s1 = (north_thresh - begseg[0])/mat4;
      *intrsct_lon = begseg[1] + s1*mat3;
      *luse_last = FALSE;
      *lthresh = TRUE;
    }
  else if ( beglat < zero && *intrsct_lat > south_thresh )
    {
      mat4 = endlat - begseg[0];
      mat3 = endlon - begseg[1];
      if ( mat3 >  pi ) mat3 = mat3 - pi2;
      if ( mat3 < -pi ) mat3 = mat3 + pi2;
      *intrsct_lat = south_thresh + tiny;
      s1 = (south_thresh - begseg[0])/mat4;
      *intrsct_lon = begseg[1] + s1*mat3;
      *luse_last = FALSE;
      *lthresh = TRUE;
    }

  /* if reached end of segment, do not use x,y intersect on next entry */

  if ( *intrsct_lat == endlat && *intrsct_lon == endlon ) *luse_last = FALSE;

}  /* pole_intersection */


/*
   this routine finds the next intersection of a destination grid 
   line with the line segment given by beglon, endlon, etc.
   a coincidence flag is returned if the segment is entirely 
   coincident with an ocean grid line.  the cells in which to search
   for an intersection must have already been restricted in the
   calling routine.
*/
void intersection(int *location, double *intrsct_lat, double *intrsct_lon, int *lcoinc,
		  double beglat, double beglon, double endlat, double endlon, double *begseg,
		  int lbegin, int lrevers,
		  int num_srch_cells, int srch_corners, int *srch_add,
		  double *srch_corner_lat, double *srch_corner_lon,
		  int *last_loc, int *lthresh, double *intrsct_lat_off, double *intrsct_lon_off,
		  int *luse_last, double *intrsct_x, double *intrsct_y,
		  int *avoid_pole_count, double *avoid_pole_offset)
{
  /*
    intent(in): 
    int lbegin, ! flag for first integration along this segment
    int lrevers ! flag whether segment integrated in reverse
    double beglat, beglon,  ! beginning lat/lon endpoints for segment
    double endlat, endlon   ! ending    lat/lon endpoints for segment

    intent(inout) :: 
    double *begseg ! begin lat/lon of full segment

    intent(out): 
    int *location  ! address in destination array containing this segment
    int *lcoinc    ! flag segments which are entirely coincident with a grid line
    double *intrsct_lat, *intrsct_lon ! lat/lon coords of next intersect.
  */
  /* local variables */
  int n, next_n, cell;
  int ioffset;

  int  loutside;  /* flags points outside grid */

  double lon1, lon2;         /* local longitude variables for segment */
  double lat1, lat2;         /* local latitude  variables for segment */
  double grdlon1, grdlon2;   /* local longitude variables for grid cell */
  double grdlat1, grdlat2;   /* local latitude  variables for grid cell */
  double vec1_lat, vec1_lon; /* vectors and cross products used */
  double vec2_lat, vec2_lon; /* during grid search */
  double cross_product; 
  double eps, offset;        /* small offset away from intersect */
  double s1, s2, determ;     /* variables used for linear solve to */
  double mat1 = 0, mat2, mat3 = 0, mat4, rhs1, rhs2;  /* find intersection */

  /* initialize defaults, flags, etc. */

  *location = -1;
  *lcoinc = FALSE;
  *intrsct_lat = endlat;
  *intrsct_lon = endlon;

  if ( num_srch_cells == 0 ) return;

  if ( beglat > north_thresh || beglat < south_thresh )
    {
      if ( *lthresh ) *location = *last_loc;
      pole_intersection(location,
			intrsct_lat, intrsct_lon, lcoinc, lthresh,
			beglat, beglon, endlat, endlon, begseg, lrevers,
			num_srch_cells, srch_corners, srch_add,
			srch_corner_lat, srch_corner_lon,
			luse_last, intrsct_x, intrsct_y,
			avoid_pole_count, avoid_pole_offset);

      if ( *lthresh )
	{
          *last_loc = *location;
          *intrsct_lat_off = *intrsct_lat;
          *intrsct_lon_off = *intrsct_lon;
        }
      return;
    }

  loutside = FALSE;
  if ( lbegin )
    {
      lat1 = beglat;
      lon1 = beglon;
    }
  else
    {
      lat1 = *intrsct_lat_off;
      lon1 = *intrsct_lon_off;
    }

  lat2 = endlat;
  lon2 = endlon;
  if ( (lon2-lon1) > three*pih )
    lon2 = lon2 - pi2;
  else if ( (lon2-lon1) < -three*pih )
    lon2 = lon2 + pi2;

  s1 = zero;

  /*
     search for location of this segment in ocean grid using cross
     product method to determine whether a point is enclosed by a cell
  */

  /*    call timer_start(12) */

  while (TRUE) /* srch_loop */
    {
      /* if last segment crossed threshold, use that location */

      if ( *lthresh )
       {
         for ( cell=0; cell < num_srch_cells; cell++ )
	   if ( srch_add[cell] == *last_loc )
	     {
               *location = *last_loc;
               eps = tiny;
               goto after_srch_loop;
	     }
       }

      /* otherwise normal search algorithm */

      for ( cell = 0; cell < num_srch_cells; cell++ ) /* cell_loop  */
	{
	  ioffset = cell*srch_corners;
	  for ( n = 0; n < srch_corners; n++ ) /* corner_loop */
	    {
	      next_n = (n+1)%srch_corners;
	      /*
		here we take the cross product of the vector making 
		up each cell side with the vector formed by the vertex
		and search point.  if all the cross products are 
		positive, the point is contained in the cell.
	      */
	      vec1_lat = srch_corner_lat[ioffset+next_n] - srch_corner_lat[ioffset+n];
	      vec1_lon = srch_corner_lon[ioffset+next_n] - srch_corner_lon[ioffset+n];
	      vec2_lat = lat1 - srch_corner_lat[ioffset+n];
	      vec2_lon = lon1 - srch_corner_lon[ioffset+n];

	      /* if endpoint coincident with vertex, offset the endpoint */

	      if ( vec2_lat == 0 && vec2_lon == 0 )
		{
		  lat1 = lat1 + 1.e-10*(lat2-lat1);
		  lon1 = lon1 + 1.e-10*(lon2-lon1);
		  vec2_lat = lat1 - srch_corner_lat[ioffset+n];
		  vec2_lon = lon1 - srch_corner_lon[ioffset+n];
		}

	      /* check for 0,2pi crossings */

	      if      ( vec1_lon >  pi ) vec1_lon -= pi2;
	      else if ( vec1_lon < -pi ) vec1_lon += pi2;

	      if      ( vec2_lon >  pi ) vec2_lon -= pi2;
	      else if ( vec2_lon < -pi ) vec2_lon += pi2;

	      cross_product = vec1_lon*vec2_lat - vec2_lon*vec1_lat;

	      /*
	       if the cross product for a side is zero, the point 
                 lies exactly on the side or the side is degenerate
                 (zero length).  if degenerate, set the cross 
                 product to a positive number.  otherwise perform 
                 another cross product between the side and the 
                 segment itself. 
	       if this cross product is also zero, the line is 
                 coincident with the cell boundary - perform the 
                 dot product and only choose the cell if the dot 
                 product is positive (parallel vs anti-parallel).
	      */
	      if ( cross_product == zero )
		{
		  if ( vec1_lat != zero || vec1_lon != zero )
		    {
		      vec2_lat = lat2 - lat1;
		      vec2_lon = lon2 - lon1;

		      if      ( vec2_lon >  pi ) vec2_lon -= pi2;
		      else if ( vec2_lon < -pi ) vec2_lon += pi2;

		      cross_product = vec1_lon*vec2_lat - vec2_lon*vec1_lat;
		    }
		  else
		    cross_product = one;

		  if ( cross_product == zero )
		    {
		      *lcoinc = TRUE;
		      cross_product = vec1_lon*vec2_lon + vec1_lat*vec2_lat;
		      if ( lrevers ) cross_product = -cross_product;
		    }
		}

	      /* if cross product is less than zero, this cell doesn't work */

	      if ( cross_product < zero ) break; /* corner_loop */

	    } /* corner_loop */

	  /* if cross products all positive, we found the location */

	  if ( n >= srch_corners )
	    {
	      *location = srch_add[cell];
	      /*
		if the beginning of this segment was outside the
		grid, invert the segment so the intersection found
		will be the first intersection with the grid
	      */
	      if ( loutside )
		{
		  lat2 = beglat;
		  lon2 = beglon;
		  *location = -1;
		  eps  = -tiny;
		}
	      else
		eps  = tiny;

	      goto after_srch_loop;
	    }

	  /* otherwise move on to next cell */

	} /* cell_loop */

      /*
	if still no cell found, the point lies outside the grid.
	take some baby steps along the segment to see if any
	part of the segment lies inside the grid.  
      */
      loutside = TRUE;
      s1 = s1 + 0.001;
      lat1 = beglat + s1*(endlat - beglat);
      lon1 = beglon + s1*(lon2   - beglon);

      /* reached the end of the segment and still outside the grid return no intersection */

      if ( s1 >= one ) return;

    } /* srch_loop */

 after_srch_loop:

  /* call timer_stop(12) */

  /*
    now that a cell is found, search for the next intersection.
    loop over sides of the cell to find intersection with side
    must check all sides for coincidences or intersections
  */
  /*   call timer_start(13) */

  for ( n = 0; n < srch_corners; n++ ) /* intrsct_loop */
    {
      next_n = (n+1)%srch_corners;

      grdlon1 = srch_corner_lon[cell*srch_corners+n];
      grdlon2 = srch_corner_lon[cell*srch_corners+next_n];
      grdlat1 = srch_corner_lat[cell*srch_corners+n];
      grdlat2 = srch_corner_lat[cell*srch_corners+next_n];

      /* set up linear system to solve for intersection */

      mat1 = lat2 - lat1;
      mat2 = grdlat1 - grdlat2;
      mat3 = lon2 - lon1;
      mat4 = grdlon1 - grdlon2;
      rhs1 = grdlat1 - lat1;
      rhs2 = grdlon1 - lon1;

      if      ( mat3 >  pi ) mat3 -= pi2;
      else if ( mat3 < -pi ) mat3 += pi2;

      if      ( mat4 >  pi ) mat4 -= pi2;
      else if ( mat4 < -pi ) mat4 += pi2;

      if      ( rhs2 >  pi ) rhs2 -= pi2;
      else if ( rhs2 < -pi ) rhs2 += pi2;

      determ = mat1*mat4 - mat2*mat3;

      /*
         if the determinant is zero, the segments are either 
           parallel or coincident.  coincidences were detected 
           above so do nothing.
         if the determinant is non-zero, solve for the linear 
           parameters s for the intersection point on each line 
           segment.
         if 0<s1,s2<1 then the segment intersects with this side.
           return the point of intersection (adding a small
           number so the intersection is off the grid line).
      */
      if ( fabs(determ) > 1.e-30 )
	{
	  s1 = (rhs1*mat4 - mat2*rhs2)/determ;
	  s2 = (mat1*rhs2 - rhs1*mat3)/determ;

	  if ( s2 >= zero && s2 <= one && s1 > zero && s1 <= one )
	    {
	      /*
		recompute intersection based on full segment
		so intersections are consistent for both sweeps
	      */
	      if ( ! loutside )
		{
		  mat1 = lat2 - begseg[0];
		  mat3 = lon2 - begseg[1];
		  rhs1 = grdlat1 - begseg[0];
		  rhs2 = grdlon1 - begseg[1];
		}
	      else
		{
		  mat1 = begseg[0] - endlat;
		  mat3 = begseg[1] - endlon;
		  rhs1 = grdlat1 - endlat;
		  rhs2 = grdlon1 - endlon;
		}

	      if      ( mat3 >  pi ) mat3 -= pi2;
	      else if ( mat3 < -pi ) mat3 += pi2;

	      if      ( rhs2 > pi  ) rhs2 -= pi2;
	      else if ( rhs2 < -pi ) rhs2 += pi2;

	      determ = mat1*mat4 - mat2*mat3;

	      /*
		sometimes due to roundoff, the previous 
		determinant is non-zero, but the lines
		are actually coincident.  if this is the
		case, skip the rest.
	      */
	      if ( determ != zero )
		{
		  s1 = (rhs1*mat4 - mat2*rhs2)/determ;
		  s2 = (mat1*rhs2 - rhs1*mat3)/determ;

		  offset = s1 + eps/determ;
		  if ( offset > one ) offset = one;

		  if ( ! loutside )
		    {
		      *intrsct_lat = begseg[0] + mat1*s1;
		      *intrsct_lon = begseg[1] + mat3*s1;
		      *intrsct_lat_off = begseg[0] + mat1*offset;
		      *intrsct_lon_off = begseg[1] + mat3*offset;
		    }
		  else
		    {
		      *intrsct_lat = endlat + mat1*s1;
		      *intrsct_lon = endlon + mat3*s1;
		      *intrsct_lat_off = endlat + mat1*offset;
		      *intrsct_lon_off = endlon + mat3*offset;
		    }
		  break; /* intrsct_loop */
		}
	    }
	}

      /* no intersection this side, move on to next side */

    } /* intrsct_loop */

  /* call timer_stop(13) */

  /*
     if the segment crosses a pole threshold, reset the intersection
     to be the threshold latitude.  only check if this was not a
     threshold segment since sometimes coordinate transform can end
     up on other side of threshold again.
  */
  if ( *lthresh )
    {
      if ( *intrsct_lat < north_thresh || *intrsct_lat > south_thresh )
	*lthresh = FALSE;
    }
  else if ( lat1 > zero && *intrsct_lat > north_thresh )
    {
      *intrsct_lat = north_thresh + tiny;
      *intrsct_lat_off = north_thresh + eps*mat1;
      s1 = (*intrsct_lat - begseg[0])/mat1;
      *intrsct_lon     = begseg[1] + s1*mat3;
      *intrsct_lon_off = begseg[1] + (s1+eps)*mat3;
      *last_loc = *location;
      *lthresh = TRUE;
    }
  else if ( lat1 < zero && *intrsct_lat < south_thresh )
    {
      *intrsct_lat = south_thresh - tiny;
      *intrsct_lat_off = south_thresh + eps*mat1;
      s1 = (*intrsct_lat - begseg[0])/mat1;
      *intrsct_lon     = begseg[1] + s1*mat3;
      *intrsct_lon_off = begseg[1] + (s1+eps)*mat3;
      *last_loc = *location;
      *lthresh = TRUE;
    }

}  /* intersection */


/*
   this routine computes the line integral of the flux function 
   that results in the interpolation weights.  the line is defined
   by the input lat/lon of the endpoints.
*/
void line_integral(double *weights, int num_wts, 
		   double in_phi1, double in_phi2, double theta1, double theta2,
		   double grid1_lat, double grid1_lon, double grid2_lat, double grid2_lon)
{
  /*
    intent(in): 
    int    num_wts               ! number of weights to compute
    double in_phi1, in_phi2,     ! longitude endpoints for the segment
    double theta1, theta2,       ! latitude  endpoints for the segment
    double grid1_lat, grid1_lon, ! reference coordinates for each
    double grid2_lat, grid2_lon  ! grid (to ensure correct 0,2pi interv.

    intent(out):
    double weights[2*num_wts]    ! line integral contribution to weights
  */

  /*  local variables  */

  double dphi, sinth1, sinth2, costh1, costh2, fac;
  double phi1, phi2;
  double f1, f2, fint;

  /*  weights for the general case based on a trapezoidal approx to the integrals. */

  sinth1 = sin(theta1);
  sinth2 = sin(theta2);
  costh1 = cos(theta1);
  costh2 = cos(theta2);

  dphi = in_phi1 - in_phi2;
  if      ( dphi >  pi ) dphi -= pi2;
  else if ( dphi < -pi ) dphi += pi2;
      
  dphi = half*dphi;

  /*
     the first weight is the area overlap integral. the second and
     fourth are second-order latitude gradient weights.
  */
  weights[        0] = dphi*(sinth1 + sinth2);
  weights[num_wts+0] = dphi*(sinth1 + sinth2);
  weights[        1] = dphi*(costh1 + costh2 + (theta1*sinth1 + theta2*sinth2));
  weights[num_wts+1] = dphi*(costh1 + costh2 + (theta1*sinth1 + theta2*sinth2));

  /*
     the third and fifth weights are for the second-order phi gradient
     component.  must be careful of longitude range.
  */
  f1 = half*(costh1*sinth1 + theta1);
  f2 = half*(costh2*sinth2 + theta2);

  phi1 = in_phi1 - grid1_lon;
  if      ( phi1 >  pi ) phi1 -= pi2;
  else if ( phi1 < -pi ) phi1 += pi2;

  phi2 = in_phi2 - grid1_lon;
  if      ( phi2 >  pi ) phi2 -= pi2;
  else if ( phi2 < -pi ) phi2 += pi2;

  if ( (phi2-phi1) <  pi && (phi2-phi1) > -pi )
    weights[2] = dphi*(phi1*f1 + phi2*f2);
  else
    {
      if ( phi1 > zero )
	fac = pi;
      else
	fac = -pi;

      fint = f1 + (f2-f1)*(fac-phi1)/fabs(dphi);
      weights[2] = half*phi1*(phi1-fac)*f1 -
	           half*phi2*(phi2+fac)*f2 +
	           half*fac*(phi1+phi2)*fint;
    }

  phi1 = in_phi1 - grid2_lon;
  if      ( phi1 >  pi ) phi1 -= pi2;
  else if ( phi1 < -pi ) phi1 += pi2;

  phi2 = in_phi2 - grid2_lon;
  if      ( phi2 >  pi ) phi2 -= pi2;
  else if ( phi2 < -pi ) phi2 += pi2;

  if ( (phi2-phi1) <  pi  && (phi2-phi1) > -pi )
    weights[num_wts+2] = dphi*(phi1*f1 + phi2*f2);
  else
    {
      if ( phi1 > zero ) fac =  pi;
      else               fac = -pi;

      fint = f1 + (f2-f1)*(fac-phi1)/fabs(dphi);
      weights[num_wts+2] = half*phi1*(phi1-fac)*f1 -
                           half*phi2*(phi2+fac)*f2 +
                           half*fac*(phi1+phi2)*fint;
    }

}  /* line_integral */


/*
    this routine stores the address and weight for this link in
    the appropriate address and weight arrays and resizes those
    arrays if necessary.
*/
void store_link_cnsrv(REMAPGRID *rg, REMAPVARS *rv, int add1, int add2, double *weights,
		      int *link_add1[2], int *link_add2[2])
{
  /*
    input variables:
    int  add1         ! address on grid1
    int  add2         ! address on grid2
    double weights[]  ! array of remapping weights for this link
  */
  /* local variables */
  int n;
  int nlink, min_link, max_link; /* link index */

  /*   if all weights are zero, do not bother storing the link */

  if ( weights[0] == zero && weights[1] == zero && weights[2] == zero &&
       weights[3] == zero && weights[4] == zero && weights[5] == zero ) return;

  /*  restrict the range of links to search for existing links */

  min_link = MIN(link_add1[0][add1], link_add2[0][add2]);
  max_link = MAX(link_add1[1][add1], link_add2[1][add2]);
  if ( min_link == -1 )
    {
      min_link = 0;
      max_link = -1;
    }

  /*  if the link already exists, add the weight to the current weight arrays */

  /* Unvectorized loop: return (97% CPU time) */
  for ( nlink = min_link; nlink <= max_link; nlink++ )
    {
      if ( add1 == rv->grid1_add[nlink] )
	if ( add2 == rv->grid2_add[nlink] )
	  {
	    for ( n = 0; n < rv->num_wts; n++ )
	      rv->wts[n][nlink] += weights[n];

	    return;
	  }
    }

  /*
     If the link does not yet exist, increment number of links and 
     check to see if remap arrays need to be increased to accomodate 
     the new link. Then store the link.
  */
  rv->num_links = rv->num_links + 1;
  if ( rv->num_links >= rv->max_links )
    resize_remap_vars(rv, rv->resize_increment);

  rv->grid1_add[rv->num_links-1] = add1;
  rv->grid2_add[rv->num_links-1] = add2;
  for ( n = 0; n < rv->num_wts; n++ )
    rv->wts[n][rv->num_links-1] = weights[n];

  if ( link_add1[0][add1] == -1 ) link_add1[0][add1] = rv->num_links-1;
  if ( link_add2[0][add2] == -1 ) link_add2[0][add2] = rv->num_links-1;
  link_add1[1][add1] = rv->num_links-1;
  link_add2[1][add2] = rv->num_links-1;

}  /* store_link_cnsrv */


/*
  -----------------------------------------------------------------------

   this routine traces the perimeters of every grid cell on each
   grid checking for intersections with the other grid and computing
   line integrals for each subsegment.

  -----------------------------------------------------------------------
*/
void remap_conserv(REMAPGRID *rg, REMAPVARS *rv)
{
  static char func[] = "remap_conserv";

  /* local variables */

  int max_subseg = 10000; /* max number of subsegments per segment to prevent infinite loop */

  int grid1_size;
  int grid2_size;
  int grid1_corners;
  int grid2_corners;
  int grid1_add;   /* current linear address for grid1 cell   */
  int grid2_add;   /* current linear address for grid2 cell   */
  int min_add;     /* addresses for restricting search of     */
  int max_add;     /* destination grid                        */
  int n, k, nwgt;  /* generic counters                        */
  int corner;      /* corner of cell that segment starts from */
  int next_corn;   /* corner of cell that segment ends on     */
  int num_subseg;  /* number of subsegments                   */

  int lcoinc;      /* flag for coincident segments            */
  int lrevers;     /* flag for reversing direction of segment */
  int lbegin;      /* flag for first integration of a segment */

  int *srch_mask; /* mask for restricting searches */

  double intrsct_lat, intrsct_lon;         /* lat/lon of next intersect  */
  double beglat, endlat, beglon, endlon;   /* endpoints of current seg.  */
  double norm_factor = 0;                  /* factor for normalizing wts */

  double *grid2_centroid_lat, *grid2_centroid_lon; /* centroid coords  */
  double *grid1_centroid_lat, *grid1_centroid_lon; /* on each grid     */

  double begseg[2];  /* begin lat/lon for full segment */
  double weights[6]; /* local wgt array */

  int     num_srch_cells;   /* num cells in restricted search arrays  */
  int     srch_corners;     /* num of corners of srch cells           */
  int    *srch_add;         /* global address of cells in srch arrays */
  double *srch_corner_lat;  /* lat of each corner of srch cells */
  double *srch_corner_lon;  /* lon of each corner of srch cells */

  int *link_add1[2];  /* min,max link add to restrict search */
  int *link_add2[2];  /* min,max link add to restrict search */

  /* intersection */
  int last_loc = FALSE; /* save location when crossing threshold */
  int lthresh = FALSE;  /* flags segments crossing threshold bndy */
  double intrsct_lat_off = 0, intrsct_lon_off = 0; /* lat/lon coords offset for next search */

  /* pole_intersection */
  /* save last intersection to avoid roundoff during coord transformation */
  int luse_last = FALSE;
  double intrsct_x, intrsct_y;  /* x,y for intersection */
  /* variables necessary if segment manages to hit pole */
  int avoid_pole_count = 0;  /* count attempts to avoid pole */
  double avoid_pole_offset = tiny;  /* endpoint offset to avoid pole */


  grid1_size = rg->grid1_size;
  grid2_size = rg->grid2_size;

  grid1_corners = rg->grid1_corners;
  grid2_corners = rg->grid2_corners;

  link_add1[0] = (int *) malloc(grid1_size*sizeof(int));
  link_add1[1] = (int *) malloc(grid1_size*sizeof(int));
  link_add2[0] = (int *) malloc(grid2_size*sizeof(int));
  link_add2[1] = (int *) malloc(grid2_size*sizeof(int));

  /* Unvectorized loop: ??? */
  for ( n = 0; n < grid1_size; n++ )
    {
      link_add1[0][n] = -1;
      link_add1[1][n] = -1;
    }

  /* Unvectorized loop: ??? */
  for ( n = 0; n < grid2_size; n++ )
    {
      link_add2[0][n] = -1;
      link_add2[1][n] = -1;
    }

  /* initialize centroid arrays */

  grid1_centroid_lat = (double *) malloc(grid1_size*sizeof(double));
  grid1_centroid_lon = (double *) malloc(grid1_size*sizeof(double));
  grid2_centroid_lat = (double *) malloc(grid2_size*sizeof(double));
  grid2_centroid_lon = (double *) malloc(grid2_size*sizeof(double));

  for ( n = 0; n < grid1_size; n++ )
    {
      grid1_centroid_lat[n] = zero;
      grid1_centroid_lon[n] = zero;
    }

  for ( n = 0; n < grid2_size; n++ )
    {
      grid2_centroid_lat[n] = zero;
      grid2_centroid_lon[n] = zero;
    }

  /*   integrate around each cell on grid1 */

  srch_mask = (int *) malloc(grid2_size*sizeof(int));

  lthresh = FALSE;
  luse_last = FALSE;
  avoid_pole_count = 0;
  avoid_pole_offset = tiny;

  for ( grid1_add = 0; grid1_add < grid1_size; grid1_add++ )
    {
      /* restrict searches first using search bins */

      /*   call timer_start(1) */
      min_add = grid2_size - 1;
      max_add = 0;
      for ( n = 0; n < rg->num_srch_bins; n++ )
	if ( grid1_add >= rg->bin_addr1[2*n+0] && grid1_add <= rg->bin_addr1[2*n+1] )
	  {
            min_add = MIN(min_add, rg->bin_addr2[2*n+0]);
            max_add = MAX(max_add, rg->bin_addr2[2*n+1]);
          }

      /* further restrict searches using bounding boxes */

      num_srch_cells = 0;
      for ( grid2_add = min_add; grid2_add <= max_add; grid2_add++ )
	{
          srch_mask[grid2_add] = (rg->grid2_bound_box[4*grid2_add+0] <= 
				  rg->grid1_bound_box[4*grid1_add+1])  &&
	                         (rg->grid2_bound_box[4*grid2_add+1] >= 
                                  rg->grid1_bound_box[4*grid1_add+0])  &&
                                 (rg->grid2_bound_box[4*grid2_add+2] <= 
                                  rg->grid1_bound_box[4*grid1_add+3])  &&
                                 (rg->grid2_bound_box[4*grid2_add+3] >= 
                                  rg->grid1_bound_box[4*grid1_add+2]);

          if ( srch_mask[grid2_add] ) num_srch_cells++;
        }

      if ( num_srch_cells == 0 ) continue;

      /* create search arrays */
      
      srch_corners = grid2_corners;
      srch_add = (int *) malloc(num_srch_cells*sizeof(int));
      srch_corner_lat = (double *) malloc(srch_corners*num_srch_cells*sizeof(double));
      srch_corner_lon = (double *) malloc(srch_corners*num_srch_cells*sizeof(double));

      n = 0;
      /* gather1 */
      for ( grid2_add = min_add; grid2_add <= max_add; grid2_add++ )
	if ( srch_mask[grid2_add] )
	  {
            srch_add[n] = grid2_add;
	    /* Unvectorized loop */
	    for ( k = 0; k < srch_corners; k++ )
	      {
		srch_corner_lat[n*srch_corners+k] = rg->grid2_corner_lat[grid2_add*srch_corners+k];
		srch_corner_lon[n*srch_corners+k] = rg->grid2_corner_lon[grid2_add*srch_corners+k];
	      }
            n++;
          }

      /*        call timer_stop(1) */

      /* integrate around this cell */

      for ( corner = 0; corner < grid1_corners; corner++ )
	{
          next_corn = (corner+1)%grid1_corners;

          /* define endpoints of the current segment */

          beglat = rg->grid1_corner_lat[grid1_add*grid1_corners+corner];
          beglon = rg->grid1_corner_lon[grid1_add*grid1_corners+corner];
          endlat = rg->grid1_corner_lat[grid1_add*grid1_corners+next_corn];
          endlon = rg->grid1_corner_lon[grid1_add*grid1_corners+next_corn];
          lrevers = FALSE;

          /* 
	     to ensure exact path taken during both
	     sweeps, always integrate segments in the same direction (SW to NE).
          */
          if ( (endlat < beglat) || (endlat == beglat && endlon < beglon) )
	    {
	      beglat = rg->grid1_corner_lat[grid1_add*grid1_corners+next_corn];
	      beglon = rg->grid1_corner_lon[grid1_add*grid1_corners+next_corn];
	      endlat = rg->grid1_corner_lat[grid1_add*grid1_corners+corner];
	      endlon = rg->grid1_corner_lon[grid1_add*grid1_corners+corner];
	      lrevers = TRUE;
	    }

          begseg[0] = beglat;
          begseg[1] = beglon;
          lbegin = TRUE;

          /*
	    if this is a constant-longitude segment, skip the rest 
	    since the line integral contribution will be zero.
          */
          if ( endlon != beglon )
	    {
	      num_subseg = 0;
	      /*
		integrate along this segment, detecting intersections 
		and computing the line integral for each sub-segment
	      */
	      while ( beglat != endlat || beglon != endlon )
		{
 		  /*
		    prevent infinite loops if integration gets stuck
		    near cell or threshold boundary
		  */
		  num_subseg++;
		  
		  if ( num_subseg >= max_subseg )
		    cdoAbort("integration stalled: num_subseg exceeded limit");

		  /* find next intersection of this segment with a gridline on grid 2. */

		  /*  call timer_start(2) */

		  intersection(&grid2_add, &intrsct_lat, &intrsct_lon, &lcoinc,
			       beglat, beglon, endlat, endlon, begseg, 
			       lbegin, lrevers,
                               num_srch_cells, srch_corners, srch_add,
			       srch_corner_lat, srch_corner_lon,
			       &last_loc, &lthresh, &intrsct_lat_off, &intrsct_lon_off,
			       &luse_last, &intrsct_x, &intrsct_y,
			       &avoid_pole_count, &avoid_pole_offset);

		  /*  call timer_stop(2) */

		  lbegin = FALSE;

		  /* compute line integral for this subsegment. */

		  /* call timer_start(3) */
		  if ( grid2_add != -1 )
		    line_integral(weights, rv->num_wts,
				  beglon, intrsct_lon, beglat, intrsct_lat,
				  rg->grid1_center_lat[grid1_add], 
				  rg->grid1_center_lon[grid1_add],
				  rg->grid2_center_lat[grid2_add], 
				  rg->grid2_center_lon[grid2_add]);
		  else
		    line_integral(weights, rv->num_wts,
				  beglon, intrsct_lon, beglat, intrsct_lat,
				  rg->grid1_center_lat[grid1_add], 
				  rg->grid1_center_lon[grid1_add],
				  rg->grid1_center_lat[grid1_add], 
				  rg->grid1_center_lon[grid1_add]);

		  /* call timer_stop(3) */

		  /* if integrating in reverse order, change sign of weights */

		  if ( lrevers ) for ( k = 0; k < 6; k++ ) weights[k] = -weights[k];

		  /*
		    store the appropriate addresses and weights. 
		    also add contributions to cell areas and centroids.
		  */
		  if ( grid2_add != -1 )
		    if ( rg->grid1_mask[grid1_add] )
		      {
			/*     call timer_start(4) */
			store_link_cnsrv(rg, rv, grid1_add, grid2_add, weights, link_add1, link_add2);
			/* call timer_stop(4) */
			rg->grid1_frac[grid1_add] += weights[0];
			rg->grid2_frac[grid2_add] += weights[rv->num_wts];
		      }

		  rg->grid1_area[grid1_add]     += weights[0];
		  grid1_centroid_lat[grid1_add] += weights[1];
		  grid1_centroid_lon[grid1_add] += weights[2];

		  /* reset beglat and beglon for next subsegment. */

		  beglat = intrsct_lat;
		  beglon = intrsct_lon;
		}
	    }
          /* end of segment */
        }

      /* finished with this cell: deallocate search array and start on next cell */

      free(srch_corner_lon);
      free(srch_corner_lat);
      free(srch_add);
    }

  free(srch_mask);

  /* integrate around each cell on grid2 */

  srch_mask = (int *) malloc(grid1_size*sizeof(int));

  lthresh = FALSE;
  luse_last = FALSE;
  avoid_pole_count = 0;
  avoid_pole_offset = tiny;

  for ( grid2_add = 0; grid2_add < grid2_size; grid2_add++ )
    {
      /* restrict searches first using search bins */

      /*   call timer_start(5) */
      min_add = grid1_size - 1;
      max_add = 0;
      for ( n = 0; n < rg->num_srch_bins; n++ )
	if ( grid2_add >= rg->bin_addr2[2*n+0] && grid2_add <= rg->bin_addr2[2*n+1] )
	  {
            min_add = MIN(min_add, rg->bin_addr1[2*n+0]);
            max_add = MAX(max_add, rg->bin_addr1[2*n+1]);
	  }

      /* further restrict searches using bounding boxes */

      num_srch_cells = 0;
      for ( grid1_add = min_add; grid1_add <= max_add; grid1_add++ )
	{
          srch_mask[grid1_add] = (rg->grid1_bound_box[4*grid1_add+0] <= 
                                  rg->grid2_bound_box[4*grid2_add+1])  &&
                                 (rg->grid1_bound_box[4*grid1_add+1] >= 
                                  rg->grid2_bound_box[4*grid2_add+0])  &&
                                 (rg->grid1_bound_box[4*grid1_add+2] <= 
                                  rg->grid2_bound_box[4*grid2_add+3])  &&
                                 (rg->grid1_bound_box[4*grid1_add+3] >= 
                                  rg->grid2_bound_box[4*grid2_add+2]);

          if ( srch_mask[grid1_add] ) num_srch_cells++;
        }

      if ( num_srch_cells == 0 ) continue;

      /* create search arrays */
      
      srch_corners = grid1_corners;
      srch_add = (int *) malloc(num_srch_cells*sizeof(int));
      srch_corner_lat = (double *) malloc(srch_corners*num_srch_cells*sizeof(double));
      srch_corner_lon = (double *) malloc(srch_corners*num_srch_cells*sizeof(double));

      n = 0;
      /* gather2 */
      for ( grid1_add = min_add; grid1_add <= max_add; grid1_add++ )
	if ( srch_mask[grid1_add] )
	  {
            srch_add[n] = grid1_add;
	    for ( k = 0; k < srch_corners; k++ )
	      {
		srch_corner_lat[n*srch_corners+k] = rg->grid1_corner_lat[grid1_add*srch_corners+k];
		srch_corner_lon[n*srch_corners+k] = rg->grid1_corner_lon[grid1_add*srch_corners+k];
	      }
            n++;
	  }

      /* call timer_stop(5) */

      /* integrate around this cell */

      for ( corner = 0; corner < grid2_corners; corner++ )
	{
          next_corn = (corner+1)%grid2_corners;

          beglat = rg->grid2_corner_lat[grid2_add*grid2_corners+corner];
          beglon = rg->grid2_corner_lon[grid2_add*grid2_corners+corner];
          endlat = rg->grid2_corner_lat[grid2_add*grid2_corners+next_corn];
          endlon = rg->grid2_corner_lon[grid2_add*grid2_corners+next_corn];
          lrevers = FALSE;

	  /*
             to ensure exact path taken during both
             sweeps, always integrate in the same direction
          */

          if ( (endlat < beglat) || (endlat == beglat  && endlon < beglon) )
	    {
	      beglat = rg->grid2_corner_lat[grid2_add*grid2_corners+next_corn];
	      beglon = rg->grid2_corner_lon[grid2_add*grid2_corners+next_corn];
	      endlat = rg->grid2_corner_lat[grid2_add*grid2_corners+corner];
	      endlon = rg->grid2_corner_lon[grid2_add*grid2_corners+corner];
	      lrevers = TRUE;
	    }

          begseg[0] = beglat;
          begseg[1] = beglon;
          lbegin = TRUE;

          /*
	    if this is a constant-longitude segment, skip the rest 
	    since the line integral contribution will be zero.
          */
          if ( endlon != beglon )
	    {
	      num_subseg = 0;
	      /*
		integrate along this segment, detecting intersections 
		and computing the line integral for each sub-segment
	      */
	      while ( beglat != endlat ||  beglon != endlon )
		{
		  /*
		    prevent infinite loops if integration gets stuck
		    near cell or threshold boundary
		  */
		  num_subseg++;

		  if ( num_subseg >= max_subseg )
 		    cdoAbort("integration stalled: num_subseg exceeded limit");

		  /* find next intersection of this segment with a gridline on grid 2. */

		  /* call timer_start(6) */

		  intersection(&grid1_add, &intrsct_lat, &intrsct_lon, &lcoinc,
			       beglat, beglon, endlat, endlon, begseg,
			       lbegin, lrevers,
                               num_srch_cells, srch_corners, srch_add,
			       srch_corner_lat, srch_corner_lon,
			       &last_loc, &lthresh, &intrsct_lat_off, &intrsct_lon_off,
			       &luse_last, &intrsct_x, &intrsct_y,
			       &avoid_pole_count, &avoid_pole_offset);

		  /* call timer_stop(6) */

		  lbegin = FALSE;

		  /* compute line integral for this subsegment. */

		  /* call timer_start(7) */

		  if ( grid1_add != -1 )
		    line_integral(weights, rv->num_wts,
				  beglon, intrsct_lon, beglat, intrsct_lat,
				  rg->grid1_center_lat[grid1_add], 
				  rg->grid1_center_lon[grid1_add],
				  rg->grid2_center_lat[grid2_add], 
				  rg->grid2_center_lon[grid2_add]);
		  else
		    line_integral(weights, rv->num_wts,
				  beglon, intrsct_lon, beglat, intrsct_lat,
				  rg->grid2_center_lat[grid2_add], 
				  rg->grid2_center_lon[grid2_add],
				  rg->grid2_center_lat[grid2_add], 
				  rg->grid2_center_lon[grid2_add]);

		  /* call timer_stop(7) */

		  /* if integrating in reverse order, change sign of weights */

		  if ( lrevers ) for ( k = 0; k < 6; k++ ) weights[k] = -weights[k];

		  /*
		    store the appropriate addresses and weights. 
		    also add contributions to cell areas and centroids.
		    if there is a coincidence, do not store weights
		    because they have been captured in the previous loop.
		    the grid1 mask is the master mask
		  */
		  if ( ! lcoinc && grid1_add != -1 )
		    if ( rg->grid1_mask[grid1_add] )
		      {
			/* call timer_start(8) */
			store_link_cnsrv(rg, rv, grid1_add, grid2_add, weights, link_add1, link_add2);
			/* call timer_stop(8) */
			rg->grid1_frac[grid1_add] += weights[0];
			rg->grid2_frac[grid2_add] += weights[rv->num_wts];
		      }

		  rg->grid2_area[grid2_add]     += weights[rv->num_wts+0];
		  grid2_centroid_lat[grid2_add] += weights[rv->num_wts+1];
		  grid2_centroid_lon[grid2_add] += weights[rv->num_wts+2];

		  /* reset beglat and beglon for next subsegment. */

		  beglat = intrsct_lat;
		  beglon = intrsct_lon;
		}
	    }
          /* end of segment */
	}

      /* finished with this cell: deallocate search array and start on next cell */

      free(srch_corner_lon);
      free(srch_corner_lat);
      free(srch_add);
    }

  free(srch_mask);

  /*
     correct for situations where N/S pole not explicitly included in
     grid (i.e. as a grid corner point). if pole is missing from only
     one grid, need to correct only the area and centroid of that 
     grid.  if missing from both, do complete weight calculation.
  */

  /* North Pole */
  weights[0] =  pi2;
  weights[1] =  pi*pi;
  weights[2] =  zero;
  weights[3] =  pi2;
  weights[4] =  pi*pi;
  weights[5] =  zero;

  grid1_add = -1;
  /* pole_loop1 */
  for ( n = 0; n < grid1_size; n++ )
    if ( rg->grid1_area[n] < -three*pih  && rg->grid1_center_lat[n] > zero )
      {
	grid1_add = n;
	break;
      }

  grid2_add = -1;
  /* pole_loop2 */
  for ( n = 0; n < grid2_size; n++ )
    if ( rg->grid2_area[n] < -three*pih && rg->grid2_center_lat[n] > zero )
      {
	grid2_add = n;
	break;
      }

  if ( grid1_add != -1 )
    {
      rg->grid1_area[grid1_add]     += weights[0];
      grid1_centroid_lat[grid1_add] += weights[1];
      grid1_centroid_lon[grid1_add] += weights[2];
    }

  if ( grid2_add != -1 )
    {
      rg->grid2_area[grid2_add]     += weights[rv->num_wts+0];
      grid2_centroid_lat[grid2_add] += weights[rv->num_wts+1];
      grid2_centroid_lon[grid2_add] += weights[rv->num_wts+2];
    }

  if ( grid1_add != -1 && grid2_add != -1 )
    {
      store_link_cnsrv(rg, rv, grid1_add, grid2_add, weights, link_add1, link_add2);

      rg->grid1_frac[grid1_add] += weights[0];
      rg->grid2_frac[grid2_add] += weights[rv->num_wts+0];
    }

  /* South Pole */
  weights[0] =  pi2;
  weights[1] = -pi*pi;
  weights[2] =  zero;
  weights[3] =  pi2;
  weights[4] = -pi*pi;
  weights[5] =  zero;

  grid1_add = -1;
  /* pole_loop3 */
  /* Unvectorized loop: break */
  for ( n = 0; n < grid1_size; n++ )
    if ( rg->grid1_area[n] < -three*pih && rg->grid1_center_lat[n] < zero )
      {
	grid1_add = n;
	break;
      }

  grid2_add = -1;
  /* pole_loop4 */
  /* Unvectorized loop: break */
  for ( n = 0; n < grid2_size; n++ )
    if ( rg->grid2_area[n] < -three*pih && rg->grid2_center_lat[n] < zero )
      {
	grid2_add = n;
	break;
      }

  if ( grid1_add != -1 )
    {
      rg->grid1_area[grid1_add]     += weights[0];
      grid1_centroid_lat[grid1_add] += weights[1];
      grid1_centroid_lon[grid1_add] += weights[2];
    }

  if ( grid2_add != -1 )
    {
      rg->grid2_area[grid2_add]     += weights[rv->num_wts+0];
      grid2_centroid_lat[grid2_add] += weights[rv->num_wts+1];
      grid2_centroid_lon[grid2_add] += weights[rv->num_wts+2];
    }

  if ( grid1_add != -1 && grid2_add != -1 )
    {
      store_link_cnsrv(rg, rv, grid1_add, grid2_add, weights, link_add1, link_add2);

      rg->grid1_frac[grid1_add] += weights[0];
      rg->grid2_frac[grid2_add] += weights[rv->num_wts];
    }

  /* finish centroid computation */

  for ( n = 0; n < grid1_size; n++ )
    if ( rg->grid1_area[n] != zero )
      {
        grid1_centroid_lat[n] = grid1_centroid_lat[n]/rg->grid1_area[n];
        grid1_centroid_lon[n] = grid1_centroid_lon[n]/rg->grid1_area[n];
      }

  for ( n = 0; n < grid2_size; n++ )
    if ( rg->grid2_area[n] != zero )
      {
        grid2_centroid_lat[n] = grid2_centroid_lat[n]/rg->grid2_area[n];
        grid2_centroid_lon[n] = grid2_centroid_lon[n]/rg->grid2_area[n];
      }

  /* include centroids in weights and normalize using destination area if requested */

  /* Unvectorized loop */
  for ( n = 0; n < rv->num_links; n++ )
    {
      grid1_add = rv->grid1_add[n];
      grid2_add = rv->grid2_add[n];
      for ( nwgt = 0; nwgt < rv->num_wts; nwgt++ )
	weights[nwgt] = rv->wts[nwgt][n];

      if ( rv->norm_opt == norm_opt_destarea )
	{
          if ( rg->grid2_area[grid2_add] != zero )
	    norm_factor = one/rg->grid2_area[grid2_add];
          else
            norm_factor = zero;
	}
      else if ( rv->norm_opt == norm_opt_fracarea )
	{
          if ( rg->grid2_frac[grid2_add] != zero )
	    norm_factor = one/rg->grid2_frac[grid2_add];
          else
            norm_factor = zero;
	}
      else if ( rv->norm_opt == norm_opt_none )
	{
          norm_factor = one;
	}

      rv->wts[0][n] =  weights[0]*norm_factor;
      rv->wts[1][n] = (weights[1] - weights[0]*grid1_centroid_lat[grid1_add])*norm_factor;
      rv->wts[2][n] = (weights[2] - weights[0]*grid1_centroid_lon[grid1_add])*norm_factor;
    }

  if ( cdoVerbose )
    cdoPrint("Total number of links = %d", rv->num_links);

  for ( n = 0; n < grid1_size; n++ )
    if ( rg->grid1_area[n] != zero ) rg->grid1_frac[n] = rg->grid1_frac[n]/rg->grid1_area[n];

  for ( n = 0; n < grid2_size; n++ )
    if ( rg->grid2_area[n] != zero ) rg->grid2_frac[n] = rg->grid2_frac[n]/rg->grid2_area[n];

  /*  perform some error checking on final weights  */

  for ( n = 0; n < grid2_size; n++ )
    {
      grid2_centroid_lat[n] = 0;
      grid2_centroid_lon[n] = 0;
    }

  /* Unvectorized loop: print */
  for ( n = 0; n < grid1_size; n++ )
    {
      if ( rg->grid1_area[n] < -.01 )
	cdoPrint("Grid 1 area error: %d %g", n, rg->grid1_area[n]);

      if ( grid1_centroid_lat[n] < -pih-.01 || grid1_centroid_lat[n] > pih+.01 )
	cdoPrint("Grid 1 centroid lat error: %d %g", n, grid1_centroid_lat[n]);

      grid1_centroid_lat[n] = zero;
      grid1_centroid_lon[n] = zero;
    }

  /* Unvectorized loop: print */
  for ( n = 0; n < grid2_size; n++ )
    {
      if ( rg->grid2_area[n] < -.01 )
	cdoPrint("Grid 2 area error: %d %g", n, rg->grid2_area[n]);
      if (grid2_centroid_lat[n] < -pih-.01 || grid2_centroid_lat[n] >  pih+.01 )
	cdoPrint("Grid 2 centroid lat error: %d %g", n, grid2_centroid_lat[n]);

      grid2_centroid_lat[n] = zero;
      grid2_centroid_lon[n] = zero;
    }

  /* Unvectorized loop: print */
  for ( n = 0; n < rv->num_links; n++ )
    {
      grid1_add = rv->grid1_add[n];
      grid2_add = rv->grid2_add[n];

      if ( rv->wts[0][n] < -0.01 )
	cdoPrint("Map 1 weight < 0 %d %d %g", grid1_add, grid2_add, rv->wts[0][n]);

      if ( rv->norm_opt != norm_opt_none && rv->wts[0][n] > 1.01 )
	cdoPrint("Map 1 weight > 1 %d %d %g", grid1_add, grid2_add, rv->wts[0][n]);

      grid2_centroid_lat[grid2_add] += rv->wts[0][n];
    }

  /* Uwe Schulzweida: check if grid2_add is valid first */
  if ( grid2_add != -1 )
    /* Unvectorized loop: print */
    for ( n = 0; n < grid2_size; n++ )
      {
        if ( rv->norm_opt == norm_opt_destarea )
          norm_factor = rg->grid2_frac[grid2_add];
        else if ( rv->norm_opt == norm_opt_fracarea )
          norm_factor = one;
        else if ( rv->norm_opt == norm_opt_none )
	  norm_factor = rg->grid2_area[grid2_add];

        if ( fabs(grid2_centroid_lat[grid2_add] - norm_factor) > .01 )
          cdoPrint("Error: sum of wts for map1 %d %g %g",
		   grid2_add, grid2_centroid_lat[grid2_add], norm_factor);
      }

  free(grid1_centroid_lat);
  free(grid1_centroid_lon);
  free(grid2_centroid_lat);
  free(grid2_centroid_lon);

  free(link_add1[0]);
  free(link_add1[1]);
  free(link_add2[0]);
  free(link_add2[1]);

} /* remap_conserv */


void remap_stat(REMAPGRID rg, REMAPVARS rv, double *array1, double *array2, double missval)
{
  static char func[] = "remap_stat";
  int n, ns, i;
  int idiff, imax, imin, icount;
  int *grid2_count;
  double minval, maxval, sum;
	  
  cdoPrint("First order mapping from grid1 to grid2:");
  cdoPrint("----------------------------------------");

  ns = 0;
  sum = 0;
  minval = DBL_MAX;
  maxval = DBL_MIN;
  for ( n = 0; n < rg.grid1_size; n++ )
    {
      if ( array1[n] != missval )
	{
	  if ( array1[n] < minval ) minval = array1[n];
	  if ( array1[n] > maxval ) maxval = array1[n];
	  sum += array1[n];
	  ns++;
	}
    }
  if ( ns > 0 ) sum /= ns;
  cdoPrint("Grid1 min,mean,max: %g %g %g", minval, sum, maxval);

  ns = 0;
  sum = 0;
  minval = DBL_MAX;
  maxval = DBL_MIN;
  for ( n = 0; n < rg.grid2_size; n++ )
    {
      if ( array2[n] != missval )
	{
	  if ( array2[n] < minval ) minval = array2[n];
	  if ( array2[n] > maxval ) maxval = array2[n];
	  sum += array2[n];
	  ns++;
	}
    }
  if ( ns > 0 ) sum /= ns;
  cdoPrint("Grid2 min,mean,max: %g %g %g", minval, sum, maxval);

  /* Conservation Test */

  if ( rg.grid1_area )
    {
      cdoPrint("Conservation:");
      sum = 0;
      for ( n = 0; n < rg.grid1_size; n++ )
	if ( array1[n] != missval )
	  sum += array1[n]*rg.grid1_area[n]*rg.grid1_frac[n];
      cdoPrint("Grid1 Integral = %g", sum);

      sum = 0;
      for ( n = 0; n < rg.grid2_size; n++ )
	if ( array2[n] != missval )
	  sum += array2[n]*rg.grid2_area[n]*rg.grid2_frac[n];
      cdoPrint("Grid2 Integral = %g", sum);
    }

  cdoPrint("number of sparse matrix entries %d", rv.num_links);
  cdoPrint("total number of dest cells %d", rg.grid2_size);

  grid2_count = (int *) malloc(rg.grid2_size*sizeof(int));

  for ( n = 0; n < rg.grid2_size; n++ ) grid2_count[n] = 0;

  for ( n = 0; n < rv.num_links; n++ ) grid2_count[rv.grid2_add[n]]++;

  imin = INT_MAX;
  imax = INT_MIN;
  for ( n = 0; n < rg.grid2_size; n++ )
    {
      if ( grid2_count[n] > 0 )
	if ( grid2_count[n] < imin ) imin = grid2_count[n];
      if ( grid2_count[n] > imax ) imax = grid2_count[n];
    }

  idiff =  (imax - imin)/10 + 1;
  icount = 0;
  for ( i = 0; i < rg.grid2_size; i++ )
    if ( grid2_count[i] > 0 ) icount++;

  cdoPrint("number of cells participating in remap %d", icount);

  if ( icount )
    {
      cdoPrint("min no of entries/row = %d", imin);
      cdoPrint("max no of entries/row = %d", imax);

      imax = imin + idiff;
      for ( n = 0; n < 10; n++ )
	{
	  icount = 0;
	  for ( i = 0; i < rg.grid2_size; i++ )
	    if ( grid2_count[i] >= imin && grid2_count[i] < imax ) icount++;
	  cdoPrint("num of rows with entries between %d - %d  %d", imin, imax-1, icount);
	  imin = imin + idiff;
	  imax = imax + idiff;
	}
    }

  free(grid2_count);

} /* remap_stat */


void remap_gradients(REMAPGRID rg, double *array, double *grad1_lat,
		     double *grad1_lon, double *grad1_latlon)
{
  static char func[] = "remap_gradients";
  int n, nx, ny, grid1_size;
  int i, j, ip1, im1, jp1, jm1, in, is, ie, iw, ine, inw, ise, isw;
  double delew, delns;
  double *grad1_lat_zero, *grad1_lon_zero;

  grid1_size = rg.grid1_size;
  nx = rg.grid1_dims[0];
  ny = rg.grid1_dims[1];

  grad1_lat_zero = (double *) malloc(grid1_size*sizeof(double));
  grad1_lon_zero = (double *) malloc(grid1_size*sizeof(double));

  for ( n = 0; n < grid1_size; n++ )
    {
      grad1_lat[n] = zero;
      grad1_lon[n] = zero;
      grad1_latlon[n] = zero;

      if ( rg.grid1_mask[n] )
	{
	  delew = half;
	  delns = half;

	  j = n/nx + 1;
	  i = n - (j-1)*nx + 1;

	  ip1 = i + 1;
	  im1 = i - 1;
	  jp1 = j + 1;
	  jm1 = j - 1;

	  if ( ip1 > nx ) ip1 = ip1 - nx;
	  if ( im1 <  1 ) im1 = nx;
	  if ( jp1 > ny )
	    {
              jp1 = j;
              delns = one;
            }
	  if ( jm1 < 1 )
	    {
              jm1 = j;
              delns = one;
            }

	  in  = (jp1-1)*nx + i - 1;
	  is  = (jm1-1)*nx + i - 1;
	  ie  = (j  -1)*nx + ip1 - 1;
	  iw  = (j  -1)*nx + im1 - 1;

	  ine = (jp1-1)*nx + ip1 - 1;
	  inw = (jp1-1)*nx + im1 - 1;
	  ise = (jm1-1)*nx + ip1 - 1;
	  isw = (jm1-1)*nx + im1 - 1;

	  /* compute i-gradient */

	  if ( ! rg.grid1_mask[ie] )
	    {
              ie = n;
              delew = one;
            }
	  if ( ! rg.grid1_mask[iw] )
	    {
              iw = n;
              delew = one;
            }
 
	  grad1_lat[n] = delew*(array[ie] - array[iw]);

	  /* compute j-gradient */

	  if ( ! rg.grid1_mask[in] )
	    {
              in = n;
              delns = one;
            }
	  if ( ! rg.grid1_mask[is] )
	    {
              is = n;
              delns = one;
            }
 
	  grad1_lon[n] = delns*(array[in] - array[is]);

	  /* compute ij-gradient */

	  delew = half;
	  if ( jp1 == j || jm1 == j )
	    delns = one;
	  else 
	    delns = half;

	  if ( ! rg.grid1_mask[ine] )
	    {
              if ( in != n )
		{
		  ine = in;
		  delew = one;
		}
              else if ( ie != n )
		{
		  ine = ie;
		  inw = iw;
		  if ( inw == n ) delew = one;
		  delns = one;
		}
              else
		{
		  ine = n;
		  inw = iw;
		  delew = one;
		  delns = one;
		}
	    }

	  if ( ! rg.grid1_mask[inw] )
	    {
              if ( in != n )
		{
		  inw = in;
		  delew = one;
		}
              else if ( iw != n )
		{
		  inw = iw;
		  ine = ie;
		  if ( ie == n ) delew = one;
		  delns = one;
		}
              else
		{
		  inw = n;
		  ine = ie;
		  delew = one;
		  delns = one;
		}
	    }

	  grad1_lat_zero[n] = delew*(array[ine] - array[inw]);

	  if ( ! rg.grid1_mask[ise] )
	    {
              if ( is != n )
		{
		  ise = is;
		  delew = one;
		}
              else if ( ie != n )
		{
		  ise = ie;
		  isw = iw;
		  if ( isw == n ) delew = one;
		  delns = one;
		}
              else
		{
		  ise = n;
		  isw = iw;
		  delew = one;
		  delns = one;
		}
	    }

	  if ( ! rg.grid1_mask[isw] )
	    {
              if ( is != n )
		{
		  isw = is;
		  delew = one;
		}
              else if ( iw != n )
		{
		  isw = iw;
		  ise = ie;
		  if ( ie == n ) delew = one;
		  delns = one;
		}
              else
		{
		  isw = n;
		  ise = ie;
		  delew = one;
		  delns = one;
		}
	    }

	  grad1_lon_zero[n] = delew*(array[ise] - array[isw]);

	  grad1_latlon[n] = delns*(grad1_lat_zero[n] - grad1_lon_zero[n]);
	}
    }

  free(grad1_lat_zero);
  free(grad1_lon_zero);

} /* remap_gradients */
